import org.apache.commons.math.complex.*;
import org.apache.commons.math.linear.*;

public class DLAC
{
	static final int NUM_ROOTS = 5;
	
	private final RealMatrix [] M;
	private final double [][] C;
	private final double [][] NN;
	private final Complex [][] root;
	private final double [] dlac;
	
	private DLAC(RealMatrix [] M, double [][] C, double [][] NN,
		Complex [][] root, double [] dlac)
	{
		this.M = M;
		this.C = C;
		this.NN = NN;
		this.root = root;
		this.dlac = dlac;
	}
	
	public RealMatrix [] getM()
	{
		return M;
	}

	public double [][] getC()
	{
		return C;
	}

	public double [][] getNN()
	{
		return NN;
	}

	public Complex [][] getRoot()
	{
		return root;
	}

	public double [] getDLAC()
	{
		return dlac;
	}

	/**
	 * @param W_EH Experimental healthy frequencies
	 * @param W_AH Analytical healthy frequencies
	 * @param W_AD Analytical damaged frequencies
	 */
	public static DLAC shm(float [] fm, double [] W_EH, double [] W_AH,
		double [][] W_AD, int [] start_index, int [] end_index,
		float sampleFreq, int numNodes, int numSamples)
	{
		double [] W = new double[numSamples];
		for(int i = 0; i < (numSamples / 2); ++i)
			W[i] = ((i * sampleFreq)/numSamples) * 2 * Math.PI;
		
		RealMatrix [] M = new RealMatrix[NUM_ROOTS];
		double [][] C = new double[NUM_ROOTS][NUM_ROOTS];
		double [][] NN = new double[NUM_ROOTS][NUM_ROOTS];
		Complex [][] root = new Complex[NUM_ROOTS][2];
		
		double [] d_w_e = new double[NUM_ROOTS];
		for(int i = 0; i < NUM_ROOTS; ++i)
		{
			float y0 = 0, y2 = 0, y4 = 0;
			float t1 = 0, t3 = 0;
			float s0 = fm[i * 5 + 0], s2 = fm[i * 5 + 1], s4 = fm[i * 5 + 2];
			float u2 = fm[i * 5 + 3], u4 = fm[i * 5 + 4];
			
			for(int x = start_index[i]; x <= end_index[i]; ++x)
			{
				y0 += 1;
				y2 += W[x] * W[x];
				y4 += W[x] * W[x] * W[x] * W[x];
			}

			double [][] M_dat = new double[5][5];
			M_dat[0][0] = y0;	M_dat[0][1] = 0;	M_dat[0][2] = -y2;	M_dat[0][3] = t1;	M_dat[0][4] = s2;
			M_dat[1][0] = 0;	M_dat[1][1] = y2;	M_dat[1][2] = 0; 	M_dat[1][3] = -s2; 	M_dat[1][4] = t3;
			M_dat[2][0] = y2;	M_dat[2][1] = 0;	M_dat[2][2] = -y4;	M_dat[2][3] = t3; 	M_dat[2][4] = s4;
			M_dat[3][0] = t1;	M_dat[3][1] = -s2;	M_dat[3][2] = -t3;	M_dat[3][3] = u2; 	M_dat[3][4] = 0;
			M_dat[4][0] = s2;	M_dat[4][1] = t3;	M_dat[4][2] = -s4;	M_dat[4][3] = 0; 	M_dat[4][4] = u4;
			M[i] = new RealMatrixImpl(M_dat);
			
			C[i][0] = s0;
			C[i][1] = t1;
			C[i][2] = s2;
			C[i][3] = 0;
			C[i][4] = u2;

			NN[i] = M[i].solve(C[i]);

			Complex a = new Complex(NN[i][4], 0.0);
			Complex b = new Complex(NN[i][3], 0.0);
			Complex c = new Complex(1.0, 0.0);
			
			// Use quadratic equation to find roots of B
			// (-b +/- sqrt(b^2 - 4ac))/2a
			
			Complex temp_sqrt = (b.multiply(b).subtract((new Complex(4.0, 0.0).multiply(a).multiply(c)))).sqrt();
			root[i][0] = b.negate().add(temp_sqrt).divide(new Complex(2.0, 0.0).multiply(a));
			root[i][1] = b.negate().subtract(temp_sqrt).divide(new Complex(2.0, 0.0).multiply(a));

			// Divide roots by 2*pi
			root[i][0] = root[i][0].divide(new Complex(Math.PI * 2.0, 0.0));
			root[i][1] = root[i][1].divide(new Complex(Math.PI * 2.0, 0.0));
			
			d_w_e[i] = (W_EH[i] - root[i][0].getImaginary()) / W_EH[i];
		}

		//
		// 5. Damage Localization Assurance Criterion (DLAC)
		//

		// Normalize experimental delta vector
		// We only need the first (positive) imaginary root
		
		double [] dlac = new double[numNodes];
		for(int i = 0; i < numNodes; ++i)
		{
			// Normalize analytical delta vector
			double [] d_w_a = new double[NUM_ROOTS];
			
			for (int j = 0; j < NUM_ROOTS; ++j)
				d_w_a[j] = (W_AH[j] - W_AD[i][j])/(W_AH[j]);

			// DLAC correlation calculation
			// ((d_w_e'*d_w_a)^2)/((norm(d_w_e)^2*norm(d_w_a)^2))
			double temp_num = 0.0;
			for (int j = 0; j < NUM_ROOTS; ++j)
				temp_num += d_w_e[j] * d_w_a[j];
			temp_num *= temp_num;

			double temp_den1 = 0.0;
			for (int j = 0; j < NUM_ROOTS; ++j)
				temp_den1 += d_w_e[j] * d_w_e[j];

			double temp_den2 = 0.0;
			for (int j = 0; j < NUM_ROOTS; ++j)
				temp_den2 += d_w_a[j] * d_w_a[j];

			dlac[i] = temp_num/(temp_den1*temp_den2);
		}
		
		return new DLAC(M, C, NN, root, dlac);
	}
}
