import java.io.*;
import java.text.*;
import java.util.*;
import java.util.logging.*;

import javax.swing.*;

import net.tinyos.message.*;
import net.tinyos.packet.*;

import servilla.network.*;

public class WS2 implements RMsgListener
{
	private static Logger logger = Logger.getLogger("WS2");

	private final MoteIF moteIF;
	private final RSNI rsni;
	private final WS2GUI gui;
	private WS2Configuration config;
	
	private String resultDir;
	
	final Map<Integer, float []> powerSpectrumData = new HashMap<Integer, float []>();
	final Map<Integer, float []> rootData = new HashMap<Integer, float []>();
	final Map<Integer, byte []> rawData = new HashMap<Integer, byte []>();

	static String intToHexString(int val) {
		String result = Integer.toHexString(val);
		if(result.length() == 1)
			result = "0" + result;
		return result;
	}

	public static final void main(String [] args)
	{
		String source = "sf@localhost:9001";
		if(args.length != 0)
		{
			if(args.length == 1)
			{
				if(args[0].equals("-h"))
					printHelp();
				else
					source = args[0];
			}
			else
				printHelp();
		}
		new WS2(source);
	}

	private static void printHelp()
	{
		System.err.println("Usage: java WS2 [source]\n\t"
			+ "[source] may be \"sf@localhost:[port]\", \"serial@[port]:[baud]\", or \"tossim-serial\"");
		System.exit(1);
	}

	public WS2(String source) {
		PhoenixSource psource = BuildSource.makePhoenix(BuildSource.makePacketSource(),
			net.tinyos.util.PrintStreamMessenger.err);
		moteIF = new MoteIF(psource);
		new ErrorListener(moteIF);
		rsni = new RSNI(new SNI(moteIF, true));

		rsni.registerListener(RawMsg.AM_TYPE, this);
		rsni.registerListener(RootMsg.AM_TYPE, this);
		rsni.registerListener(PSMsg.AM_TYPE, this);
		
		try
		{
			config = new WS2Configuration();
			if(config.debug)
				logger.setLevel(Level.FINE);
			else
				logger.setLevel(Level.WARNING);
		}
		catch(IOException e)
		{
			logger.log(Level.SEVERE, "Could not load configuration file", e);
			JOptionPane.showMessageDialog(null,
				"Could not load configuration file: " + e.getMessage(),
				"Error", JOptionPane.ERROR_MESSAGE);
			System.exit(1);
		}

		gui = new WS2GUI(this, config);
		gui.setVisible(true);
		makeResultDir();
	}
	
	private void makeResultDir()
	{
		SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd_HH-mm-ss");
		resultDir = "results" + File.separator + format.format(new Date())
			+ File.separator;
		if(new File(resultDir).mkdirs())
			logger.info("Results will be saved in " + resultDir);
		else
		{
			logger.severe("Failed to create directory for storing results.");
			JOptionPane.showMessageDialog(gui,
				"Failed to create directory for storing results.", "Error",
				JOptionPane.ERROR_MESSAGE);
			gui.setVisible(false);
		}
	}
	
	public void rcvdMsg(int amType, byte [] data) {
		switch(amType) {
		case RawMsg.AM_TYPE:
			handleRawMsg(new RawMsg(data));
			break;
		case RootMsg.AM_TYPE:
			handleRootMsg(new RootMsg(data));
			break;
		case PSMsg.AM_TYPE:
			handlePSMsg(new PSMsg(data));
			break;
		}
	}

	private void handleRawMsg(RawMsg rm)
	{
		byte [] data = new byte[RawMsg.totalSize_data()];
		for(int i = 0; i < data.length; i++)
			data[i] = (byte)rm.getElement_data(i);
		
		rawData.put(rm.get_id(), data);
		gui.makeRawDataButtonGreen(rm.get_id());
		logger.info("Received RawMsg from node " + rm.get_id());
	}
	
	private void handleRootMsg(RootMsg rm)
	{
		rootData.put(rm.get_id(), rm.get_data());
		gui.makeResultButtonGreen(rm.get_id());
		logger.info("Received RootMsg from node " + rm.get_id());
		logger.fine(rm.toString());
	}

	private void handlePSMsg(PSMsg psm)
	{
		powerSpectrumData.put(psm.get_id(), psm.get_data());
		gui.makePSDataButtonGreen(psm.get_id());
		logger.info("Received PSMsg from node " + psm.get_id());
		logger.fine(psm.toString());
	}

	boolean savePowerSpectrum(int nodeID)
	{
		String fileName = resultDir + "Node" + intToHexString(nodeID)
			+ "PS.txt";
		logger.info("Saving Power Spectrum of node " + intToHexString(nodeID)
			+ " into " + fileName);

		if(!powerSpectrumData.containsKey(nodeID))
		{
			logger.severe("Some Power Spectrum data is missing");
			return false;
		}

		float [] data = powerSpectrumData.get(nodeID);
		StringBuffer sb = new StringBuffer();
		for(int i = 0; i < data.length; i++)
			sb.append(i + "\t" + data[i] + "\n");

		try
		{
			FileOutputStream out = new FileOutputStream(fileName, false);
			out.write(sb.toString().getBytes());
			out.flush();
			out.close();
		}
		catch(IOException e)
		{
			logger.log(Level.SEVERE, "Error writing to file " + fileName, e);
		}

		return true;
	}
	
	private static void writeVector(double [] vector, StringBuffer sb)
	{
		for(double entry : vector)
			sb.append("\t" + entry);
		sb.append("\n");
	}
	
	private int [] toIndex(double [] search, float sampleFreq)
	{
		int [] indices = new int[search.length];
		for(int i = 0; i < indices.length; i++)
			indices[i] = (int)(search[i] * config.numSamples / sampleFreq) + 1;
		
		return indices;
	}
	
	boolean saveRoot(int nodeID, float sampleFreq)
	{
		if(!rootData.containsKey(nodeID))
			return false;
		
		int [] start_index = toIndex(config.start, sampleFreq);
		int [] end_index = toIndex(config.end, sampleFreq);

		DLAC dlac = DLAC.shm(rootData.get(nodeID), config.W_EH, config.W_AH,
			config.W_AD, start_index, end_index, sampleFreq,
			config.numNodes, config.numSamples);

		StringBuffer sb = new StringBuffer();

		for(int i = 0; i < DLAC.NUM_ROOTS; i++)
		{
			sb.append("\n --- --- --- Matrix " + i + " --- --- ---\n");
			for(double [] row : dlac.getM()[i].getData())
				writeVector(row, sb);

			sb.append("\n --- --- --- Vector C " + i + " --- --- ---\n");
			writeVector(dlac.getC()[i], sb);

			sb.append("\n --- --- --- Vector NN " + i + " --- --- ---\n");
			writeVector(dlac.getNN()[i], sb);
		}

		for(int i = 0; i < DLAC.NUM_ROOTS; i++)
		{
			sb.append("Roots " + (i + 1) + " = ");
			sb.append(dlac.getRoot()[i][0].getReal() + " + i ("
				+ dlac.getRoot()[i][0].getImaginary() + "), ");
			sb.append(dlac.getRoot()[i][1].getReal() + " + i ("
				+ dlac.getRoot()[i][1].getImaginary() + ")\n");
		}
		sb.append("\n");

		sb.append("\n");
		for(int i = 0; i < config.numNodes; i++)
			sb.append("DLAC " + (i + 1) + " = " + dlac.getDLAC()[i] + "\n");
		sb.append("\n");

		String fileName = resultDir + "Node" + intToHexString(nodeID)
			+ "Roots.txt";
		logger.info("Saving Roots of node " + intToHexString(nodeID) + " into "
			+ fileName);

		try
		{
			FileOutputStream out = new FileOutputStream(fileName, false);
			// false == do not append
			out.write(sb.toString().getBytes());
			out.flush();
			out.close();
		}
		catch(IOException e)
		{
			logger.log(Level.SEVERE, "Error writing to file " + fileName, e);
		}

		logger.info("Sampling Frequency:  " + sampleFreq);
		logger.info("Number of Samples:  " + config.numSamples);
		for(int i = 0; i < 5; i++)
			logger.info("Search Range:  " + start_index[i] + " - "
				+ end_index[i]);

		return true;
	}
	
	boolean saveRaw(int nodeID)
	{
		if(!rawData.containsKey(nodeID))
			return false;

		byte [] data = rawData.get(nodeID);
		String fileName = resultDir + "Node" + intToHexString(nodeID) + ".raw";
		logger.info("Saving raw data for node " + intToHexString(nodeID)
			+ " in " + fileName + ", size=" + data.length + "B");
		try
		{
			FileOutputStream out = new FileOutputStream(fileName);
			out.write(data);
			out.flush();
			out.close();
		}
		catch(IOException e)
		{
			logger.log(Level.SEVERE, "Error writing to file", e);
		}

		return true;
	}
	
	void setInitValues(int nodeID, float sampleFreq)
	{
		SetSamplingFreqMsg ssfm = new SetSamplingFreqMsg();

		ssfm.set_id(nodeID);
		ssfm.set_samplingFreq(sampleFreq);
		ssfm.set_st(toIndex(config.start, sampleFreq));
		ssfm.set_sp(toIndex(config.end, sampleFreq));

		logger.info("Setting sampling frequency of node "
			+ intToHexString(nodeID) + " to " + sampleFreq + "Hz");
		send(0xffff, ssfm); // broadcast
	}

	void requestRootData(int nodeID)
	{
		GetRootMsg grfm = new GetRootMsg();
		grfm.set_id(nodeID);

		logger.info("Getting root from node " + intToHexString(nodeID));
		send(0xffff, grfm);
	}

	void requestRawData(int nodeID)
	{
		// initially just get all of the raw messages
		GetRawMsg m = new GetRawMsg();
		m.set_id(nodeID);

		logger.info("Getting raw data from node " + intToHexString(nodeID));
		send(0xffff, m);
	}

	void requestPowerSpectrumData(int nodeID)
	{
		GetPSMsg m = new GetPSMsg();
		m.set_id(nodeID);
		m.set_index(0);

		logger.info("Getting PS from node " + intToHexString(nodeID));
		send(0xffff, m);
	}
	
	void startDLAC()
	{
		logger.info("Starting DLAC");
		send(0xffff, new StartMsg());
	}
	
	private void send(int nodeID, Message msg)
	{
		try
		{
			logger.finer("Sending message to node " + intToHexString(nodeID));
			logger.finest(msg.toString());
			moteIF.send(nodeID, msg);
		}
		catch(IOException e)
		{
			logger.log(Level.SEVERE, "Could not send message", e);
			JOptionPane.showMessageDialog(gui,
				"Could not send message to node " + intToHexString(nodeID) + 
				": " + e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
		}
	}
	
	@Override
	public String toString()
	{
		return "WS2";
	}

	private class ErrorListener implements MessageListener
	{
		ErrorListener(MoteIF moteIF)
		{
			moteIF.registerListener(new ErrorMsg(), this);
		}

		public void messageReceived(int to, Message m)
		{
			if(m instanceof ErrorMsg)
			{
				ErrorMsg em = (ErrorMsg)m;
				String msg = "Accelerometer driver on mote "
					+ intToHexString(em.get_id()) + " crashed";
				logger.severe(msg);
				JOptionPane.showMessageDialog(gui, msg, "Error",
					JOptionPane.ERROR_MESSAGE);
			}
		}
	}
}
