// $Id: AgentInjector.java,v 1.10 2005/12/13 00:18:42 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2004, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */
/*									tab:4
 * "Copyright (c) 2000-2003 The Regents of the University  of California.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Copyright (c) 2002-2003 Intel Corporation
 * All rights reserved.
 *
 * This file is distributed under the terms in the attached INTEL-LICENSE
 * file. If you do not find these files, copies can be found by writing to
 * Intel Research Berkeley, 2150 Shattuck Avenue, Suite 1300, Berkeley, CA,
 * 94704.  Attention:  Intel License Inquiry.
 */
/* Authors:	Phil Levis <pal@cs.berkeley.edu>,
 Chien-Liang Fok <liang@cse.wustl.edu>
 * Date:        May 2 2005
 * Desc:        Main class for the Servilla Agent Injector.
 *
 */

package servilla;

import servilla.compiler.ServillaScriptC;
import servilla.compiler.tree.*;
import servilla.service.ServiceSpecC;
import servilla.service.tree.*;
import servilla.util.ArrayToHexString;
import servilla.util.Debugger;

import java.util.*;
import java.io.*;

import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;

//import servilla.rmi.agentInjector.*;
import servilla.msgs.*;
import servilla.network.*;

import org.apache.commons.lang.ArrayUtils;

/**
 * This is the main class that launches the Java application that interacts
 * with a Servilla sensor network.
 * 
 * @author Chien-Liang Fok <liangfok@wustl.edu>
 */
public class Main implements ServillaConstants {
	private MoteIF moteIF = null;
	private SNI sni;
	private RSNI rsni;
	private ScriptSndr scriptSender;
	private String source;
	private PhoenixSource psource;
	private ScriptInjectorGUI gui = null;
	private NeighborListQuerier nlq;
	private ServillaScriptC scriptCompiler;
	
	/**
	 * Creates the main Servilla component.
	 *
	 * @param source The MoteIF source, e.g., COM4:mica2 or sf@localhost:9001
	 * @param debug Whether to be in debug mode
	 * @param optimize Whether code optimization should be used.
	 */
	public Main(String source, boolean createGUI, ServillaScriptC scriptCompiler) throws Exception {
		this.scriptCompiler = scriptCompiler;
		sni = new SNI(false); // not promiscuous
		rsni = new RSNI(sni);
		connect(source);
		new HeartBeat(sni);
		new ErrorDisplayer(sni);
		new SendDataRcvr(rsni);
		nlq = new NeighborListQuerier(sni);
		scriptSender = new ScriptSndr(this, rsni);
		if (createGUI) {
			gui = createSwingGUI(this);
			gui.setSFStatusConnected(source);
		}
	}
	
	/**
	 * Connects to the MoteIF.
	 * 
	 * @return true if successful.
	 */
	public boolean connect(String source) throws Exception {
		if (moteIF == null) {
			if (source.equals("tossim-serial"))
				moteIF = new MoteIF(PrintStreamMessenger.err);
			else if (source.startsWith("sf")) {
				psource = BuildSource.makePhoenix(BuildSource.makePacketSource(source),
						net.tinyos.util.PrintStreamMessenger.err);
				moteIF = new MoteIF(psource);
			}
			else {
				psource = BuildSource.makePhoenix(
					BuildSource.makeArgsSerial(source),
					net.tinyos.util.PrintStreamMessenger.err);
				moteIF = new MoteIF(psource);
			}
			
			if (gui != null) gui.setSFStatusConnected(source);
			this.source = source;
			log("Created MoteIF: " + source);
			
			sni.setMoteIF(moteIF);
			return true;
		} else return false;
	}
	
	/**
	 * Disconnects servilla.Main from the MoteIF.
	 */
	public void disconnect() {
		if (moteIF != null) {
			psource.shutdown();
			sni.disconnect();
			moteIF = null;
			if (gui != null) gui.setSFStatusDisconnected();
		}
	}
	
	/**
	 * Returns true if the gui is being shown.
	 * @return true if the gui is being shown.
	 */
	public boolean showGUI() { return gui != null; }
	
	/**
	 * Use reflection to launch a SwingScriptInjectorGUI.  This is necessary to allow
	 * Servilla to run on mobile java.
	 * 
	 * @param parent  The main (parent) component that is creating the GUI.
	 * @return The swing-based GUI, or null if an error occured.
	 */
	private static ScriptInjectorGUI createSwingGUI(Main parent) {
		try {
			Class guiClass = Class.forName("servilla.SwingScriptInjectorGUI");
			java.lang.reflect.Constructor constructor = guiClass.getConstructor(new Class [] { Main.class });
			return (ScriptInjectorGUI)constructor.newInstance(new Object [] { parent });
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}
	
	/**
	 * Returns the name of the connection to the sensor network.
	 * 
	 * @return the name of the connection to the sensor network.
	 */
	public String source() { return source; }
	
	/**
	 * Returns the reliable sensor network interface.
	 * 
	 * @return the reliable sensor network interface, or null if servilla.Main is currently
	 * disconnected from the sensor network.
	 */
//	public ReliableSNInterface rsni() { return rsni; }
	
	/**
	 * Returns the sensor network interface.
	 * 
	 * @return the sensor network interface, or null if servilla.Main is currently
	 * disconnected from the sensor network.
	 */
//	public SNInterface sni() { return sni; }
	
	/**
	 * Determines whether servilla.Main is connected to the sensor network.
	 * 
	 * @return true if connected.
	 */
	public boolean isConnected() { return moteIF == null; }
	
	/**
	 * Resets all motes in the network.  After reseting a mote, the user
	 * must wait 3 seconds before the mote's network interface is running again.
	 */
	public void reset() { reset(TOS_BCAST_ADDRESS); }
	
	/**
	 * Resets a mote at the specified address.  This method forces the
	 * calling thread to sleep for 3 seconds to allow the resetting mote
	 * to come back online.
	 *
	 * @param addr The original address of the mote to reset.
	 */
	public void reset(int addr) {
		if (addr == TOS_BCAST_ADDRESS)
			Debugger.dbg("Main", "Resetting the entire network reachable from base station.");
		else
			Debugger.dbg("Main", "Resetting mote " + addr + ".");
		try {
			sni.send(addr, new ServillaResetMsgJ(TOS_UART_ADDRESS, addr));
			Thread.sleep(3000);  // pause for 3 seconds to allow mote to boot
		} catch(Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Inject a script into the network.
	 *
	 * @param script The script.
	 * @param dest The destination.
	 */
	public void inject(int dest, Script script) { scriptSender.send(dest, script); }
	
	/**
	 * Injects a script into the network.  This is called by the GUI when the user hits the
	 * inject button.
	 * 
	 * @param dest The destination address on which to inject the script.
	 * @param fileName The name of the file that contained the text.
	 * @param programText The actual text of the file.
	 * @throws Exception
	 */
	public void inject(int dest, String fileName, String programText) throws Exception {
		CodeWriter cw = scriptCompiler.compile(fileName, programText);
		inject(dest, new Script(fileName, cw));
	}
	
	/**
	 * Returns the neighbor list querier.
	 * 
	 * @return
	 */
	public NeighborListQuerier nbrListQuerier() { return nlq; }
	
	
//	private CodeWriter assembleSpec(String code) {
//		CodeWriter cw = new CodeWriter();
//		BufferedReader br = new BufferedReader(new StringReader(code));
//		int lineCnt = 1;
//		
//		try {
//			String line = br.readLine();
//			while (line != null) {
//				Instruction i = new Instruction(line);
//				cw.add(i);
//				lineCnt++;
//				line = br.readLine();
//			}
//		} catch(IOException ioe) {
//			System.err.println("Problems with line " + lineCnt);
//			ioe.printStackTrace();
//			System.exit(1);
//		} catch(InstructionException e) {
//			System.err.println("Problems creating instruction on line " + lineCnt);
//			e.printStackTrace();
//			System.exit(1);
//		}
//		
//		//System.out.println("Assembled program: \n" + cw.toString());
//		return cw;
//	}
	
	/**
	 * Injects a script into the network.
	 */
//	public void inject(int dest, Script s) throws Exception {
//		log("Inject called, dest="+dest+" script=" + s.fileName());
//		
//		ServillaScriptC scriptCompiler = new ServillaScriptC();	 // create the script compiler
//		ServiceSpecC serviceCompiler = new ServiceSpecC();
//		
//		CodeWriter cw = null;
//		String programText = null;
//		try {
//			programText = ProgramReader.readProgram(fileName);
//		} catch(IOException e) {
//			Debugger.log("Main", "ERROR: Unable to open " + fileName);
//			return;
//		}
//		if (programText.contains("main()"))
//			cw = scriptCompiler.compile(programText, optimize);
//		else
//			cw = assembleSpec(programText); // assemble a service spec.
//		Vector<ServiceSpec> requisiteServ = new Vector<ServiceSpec>();
//		Enumeration<ServiceDecl> e = cw.reqServices().elements();
//		while (e.hasMoreElements()) {
//			ServiceDecl s = e.nextElement();
//			if (s.hasHandle()) {	// only add services that have been referred to in the script's code
//				ServiceSpec ss = serviceCompiler.compile(s.name() + ".service");
//				ss.setHandle(s.handle());
//				requisiteServ.add(ss);
//			}
//		}
//		
//		Script script = new Script(cw.byteCode(), requisiteServ);
//		//script.heap().add(new servilla.variables.Int16Var(123));
//		
//		log("Compiled script: \n"+script);
//		scriptSender.send(dest, script);
//	}
	
	public static void main(String[] args) {
		try {
			String fileName = null, source = "sf@localhost:9001";
			boolean compileOnly = false, scanOnly = false, launchGUI = false, optimize = true;
			int addr = 0; // The destination address on which to inject the script
			
			if (args.length == 0) {
				System.err.println("ERROR: No script specified.");
				usage();
				System.exit(1);
			}
			
			int index = 0;
			while (index < args.length) {
				String arg = args[index];
				if (arg.equals("-h") || arg.equals("--help")) {
					usage();
					System.exit(0);
				} else if (arg.equals("-comm"))
					source = args[++index];
				else if (arg.equals("-compile") || arg.equals("-c"))
					compileOnly = true;
				else if (arg.equals("-scan") || arg.equals("-s"))
					scanOnly = true;
				else if (arg.equals("-d"))
					Debugger.debug = true;
				else if (arg.equals("-no"))
					optimize = false;
				else if (arg.equals("-addr"))
					addr = Integer.valueOf(args[++index]);
				else if (arg.equals("-gui"))
					launchGUI = true;
				else {
					if (index == args.length-1)
						fileName = args[args.length-1];
					else {
						usage();
						System.exit(1);
					}
				}
				index++;
			}
			
			
			if (fileName != null && fileName.endsWith(".service")) {	// recognize services
				if (scanOnly) {
					Debugger.log(null, "Scanning service description " + fileName + "\n");
					ServiceSpecC specC = new ServiceSpecC();
					specC.scan(fileName);
				} else {
					Debugger.log(null, "Compiling service description " + fileName + "\n");
					try {
						ServiceSpecC ssc = new ServiceSpecC();
						ServiceSpec ss = ssc.compile(fileName);
						printSpec(ss, "_desc");
					} catch(Exception e) {
						Debugger.log(null, "Error while compiling service spec " + fileName + ": " + e.getMessage());
						e.printStackTrace();
						System.exit(1);
					}
				}
				System.exit(0);
			}
			
			ServillaScriptC scriptCompiler = new ServillaScriptC(optimize);	 // create the script compiler
			
			if (scanOnly) {
				scanScript(fileName);
			} else {
				Script script = null;
				if (fileName != null) {
					System.out.println("// Compiling script " + fileName + ", optimize="+optimize);
					CodeWriter cw = scriptCompiler.compileFile(fileName);
					script = new Script(fileName, cw);
				}
				if (compileOnly) {
					printScript(script);
				} else {
					Main main = new Main(source, launchGUI, scriptCompiler);
					if (script != null) main.inject(addr, script);
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
			//usage();
		}
	}
	
	private static void printScript(Script script) throws Exception {
		try { System.out.println(script.parseTree()); } catch(AssemblerException e) {e.printStackTrace(); }
		System.out.println("// Script id=" + script.id() + " pc=" + script.pc() + " cond=" + script.cond() 
				+ " fp=" + script.fp() + " sp=" + script.stack().sp() + " init_sp=" + script.init_sp());
		System.out.println("// ------------------------- Assembled Program -------------------------");
		System.out.println(script.asm());
		System.out.println("// ---------------------------------------------------------------------");
		System.out.println("// Bytecode in NesC:");
		System.out.println("// " + script.byteCode().length + " Bytes");
		System.out.println(NesCWriter.byteArray(ArrayToHexString.toString(script.byteCode()), "scriptCode"));
		
		System.out.println("\n// ---------------------------------------------------------------------");
		System.out.println("#define DEFAULT_SCRIPT_ID "+script.id());
		//System.out.println("#define DEFAULT_SCRIPT_SP "+script.stack().sp());
		System.out.println("#define DEFAULT_SCRIPT_FP " + script.fp());
		System.out.println("#define DEFAULT_SCRIPT_INIT_SP " + script.init_sp());
		System.out.println("// Bytecode in NesC with Human-readable Instructions:");
		System.out.println("// " + script.byteCode().length + " Bytes");
		System.out.println(NesCWriter.byteArray(script.nesCCode(), "scriptCode"));
		
		System.out.println("\n// ---------------------------------------------------------------------");
		
		Vector<ServiceSpec> specs = script.specs();  // fetch the rquired services
		
		if (specs.size() == 0) {
			System.out.println("// Script required no services.");
		} else {
			// TODO: Print all specifications in one big array
			Enumeration<ServiceSpec> e = specs.elements();
			int numUsedServices = 0;
			while (e.hasMoreElements()) {
				ServiceSpec ss = e.nextElement();
				printSpec(ss, "specBytes");
				numUsedServices++;
			}
			if (numUsedServices == 0)
				System.out.println("// Script declared but did not use services.");
		}
		System.out.println("\n//Here is the initial state of the stack.");
		byte[] stackBytes = ArrayUtils.toPrimitive(script.stack().toByteVector().toArray(new Byte[]{}));
		System.out.println(NesCWriter.byteArray(ArrayToHexString.toString(stackBytes), "stackBytes"));
		System.out.println("// Done compiling " + script.fileName() + ".");
	}
	
	private static void printSpec(ServiceSpec ss, String name) throws SpecCompilerException {
		byte[] ba = ss.toByteArray();
		System.out.println("// Copy this into your service NesC component.");
		System.out.println("// The following specification was automatically generated by servilla.Main.");
		System.out.println("// Size: " + ba.length + " bytes");
		System.out.println(("// " + ss.toString()).replaceAll("\n", "\n//"));
		//System.out.println("uint16_t _servSize = " + ba.length + ";");
		System.out.println(NesCWriter.byteArray(ArrayToHexString.toString(ba), name));
	}

	/**
	 * Compile a service specification.  The second parameter specifies the service handle
	 * (or number) that the script will use to refer to the service.  If the service description
	 * is written by the service developer, the service number should be 0xff.
	 * 
	 * @param fileName The file that stores the service specification.
	 * @param isSpec Whether the service is a *specification* provided by the script, or a *description*
	 * provided by the service.  If is a spec provided by the script, it includes the script id and
	 * handle, otherwise it only includes the description.
	 * @param handle The handle to the service.
	 * @throws Exception
	 */
//	private static ServiceSpec compileSpec(String fileName) throws Exception {
//		ServiceSpecC serviceCompiler = new ServiceSpecC();
//		ServiceSpec ss = serviceCompiler.compile(fileName);
//		return ss;
//	}
	

	/**
	 * Scan a Servilla Script.
	 * 
	 * @param fileName The name of the Servilla Script.
	 * @throws IOException
	 */
	private static void scanScript(String fileName) throws IOException {
		System.out.println("Scanning script " + fileName + "...");
		Vector<ServiceDecl> services = new ServillaScriptC().scanFile(fileName);
		
		if (services.size() == 0) {
			System.out.println("Script required no services.");
		} else {
			ServiceSpecC serviceCompiler = new ServiceSpecC();
			Enumeration<ServiceDecl> e = services.elements();
			while (e.hasMoreElements()) {
				ServiceDecl s = e.nextElement();
				System.out.println("\nScanning required service " + s.name() + "...");
				serviceCompiler.scan(s.name());
				
			}
		}
		System.out.println("Done scanning " + fileName + ".");
	}
	
	private static void usage() {
		System.err.println("servilla.Main compiles and injects an script into the network.");
		System.err.println("usage: servilla.Main [options] [script file name]");
		System.err.println("\tValid [options]:");
		System.err.println("\t\t-h\t\tPrint this help message");
		System.err.println("\t\t-addr [addr]\tSpecify the destination address (default 0).");
		System.err.println("\t\t-comm <source>\t<source> is COMx:[platform] or tossim-serial (default: sf@localhost:9001)");
		System.err.println("\t\t-compile or -c\tCompile only, do not inject script");
		System.err.println("\t\t-gui\tDisplay the GUI (default false).");
		System.err.println("\t\t-scan or -s\t\tScan only, do not compile");
		System.err.println("\t\t-no\t\tDisable compiler optimizations.");
		System.err.println("\t\t-d\t\tEnable debug printouts.");
	}
	
	private void log(String msg) {
		Debugger.dbg("servilla.Main", msg);
	}
	
}

