// $Id: Agent.java,v 1.4 2005/12/09 07:24:05 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2004, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */


package servilla;

import org.apache.commons.lang.ArrayUtils;
import servilla.compiler.tree.*;
import servilla.service.tree.*;
import servilla.service.ServiceSpecC;
import servilla.util.*;

import java.io.*;
import java.util.*;

/**
 * This is a Java-implementation of an Servilla script.
 * The Servilla Script conists of instruction byte code, 
 * service specification byte code, and state (a stack).
 *
 * @author Chien-Liang Fok
 */
public class Script implements ServillaConstants, Serializable {
	static final long serialVersionUID = 8825485772863901849L;
	private static int idcount = 10; // the Agent ID counter
	private int id, cond = 0;
	private long pc = 0, init_sp = 0, fp = 0;
	private CodeWriter cw;
	private Vector<ServiceSpec> services; // required services
	private Stack stack;
	private String fileName;
	
	/**
	 * Create a script with a specific id.
	 * 
	 * @param id The id of the agent
	 * @param cw The CodeWriter that is able to generate the script's code
	 * @param services The script's services
	 */
	public Script(String fileName, int id, CodeWriter cw) throws Exception {
		this.fileName = fileName;
		this.id = id;
		this.cw = cw;
		this.services = new Vector<ServiceSpec>();
		this.init_sp = cw.initSP();
		this.stack = new Stack((int)cw.initSP());
		
		ServiceSpecC ssc = new ServiceSpecC();
		Enumeration<ServiceDecl> e = cw.getDecls().elements();
		while (e.hasMoreElements()) {
			ServiceDecl sd = e.nextElement();
			if (sd.hasHandle()) {							// only compile declarations that have handles
				ServiceSpec ss = ssc.compile(sd.name());
				ss.setHandle(sd.handle());					// sets the spec's handle
				this.services.add(ss);
			}
		}
	}
	
	/**
	 * Create a script with a unique ID.
	 * 
	 * @param cw The script's code
	 * @param services The script's services
	 */
	public Script(String fileName, CodeWriter cw) throws Exception { this(fileName, idcount++, cw); }
	
	/**
	 * Returns the name of the file containing the script.
	 * @return the name of the file containing the script.
	 */
	public String fileName() { return fileName; }
	
	/**
	 * Returns the parse tree.  The parse tree is the hierarchy of all of the statements and expressions
	 * in the script.
	 * @return The parse tree.
	 * @throws AssemblerException
	 */
	public String parseTree() throws AssemblerException { return cw.getMethod("main", 0, 0).toString(); }
	
	/**
	 * Returns the service declarations within the script.
	 * @return the service declarations within the script.
	 */
	public Vector<ServiceDecl> specDecls() { return cw.getDecls(); }
	
	/**
	 * Returns a list of service specifications required by the script.
	 * 
	 * @return A list of services required by the script.
	 */
	public Vector<ServiceSpec> specs() { return services; }
	
	/**
	 * Returns the compiled service specifications.
	 * 
	 * @return The compiled service specifications.
	 * @throws SpecCompilerException When the one of the script's specifications is not valid
	 */
	public byte[] specBytes() throws SpecCompilerException {
		Vector<Byte> result = new Vector<Byte>();
		Enumeration<ServiceSpec> e = services.elements();
		while (e.hasMoreElements()) {
			result.addAll(e.nextElement().toByteVector());
		}
		return ArrayUtils.toPrimitive(result.toArray(new Byte[]{}));
	}
	
	/**
	 * Returns the number of bytes the services consumes.
	 * 
	 * @return The number of bytes the services consumes.
	 */
	public int specSize() throws SpecCompilerException { return  specBytes().length; }
	
	/**
	 * Returns the script's ID.
	 * 
	 * @return The script ID.
	 */
	public int id() { return id; }
	
	/**
	 * Accessor to the program counter.
	 * 
	 * @return The program counter.
	 */
	public long pc() { return pc; }
	
	/**
	 * Set the program counter.
	 * 
	 * @param pc The program counter
	 */
	public void setPC(long pc) { this.pc = pc; }
	
	/**
	 * Returns the initial stack pointer.  This makes space for the 
	 * local variables declared in main.
	 * @return The initial stack pointer.
	 */
	public long init_sp() { return init_sp; }
	
	/**
	 * Returns the frame pointer.  Uneless the script is in mid-execution,
	 * this value should be 0.
	 * 
	 * @return the frame pointer.
	 */
	public long fp() { return fp; }
	
	/**
	 * Set the frame pointer.
	 * 
	 * @param fp The frame pointer
	 */
	public void set_fp(long fp) { this.fp = fp; }
	
	/**
	 * Accessor to the condition register.
	 * 
	 * @return The condition code.
	 */
	public int cond() { return cond; }
	
	/**
	 * Sets the condition code.
	 * 
	 * @param cond The condition code.
	 */
	public void setCond(int cond) { this.cond = cond; }
	
	/**
	 * Returns the size of the code.
	 * 
	 * @return the size of the code.
	 */
	public int codeSize() { return byteCode().length; }
	
	/**
	 * Returns a string of the byte code with the instructions
	 * in string format.
	 * 
	 * @return
	 */
	public String nesCCode() { return cw.nesCCode(); }
	
	/**
	 * Returns the script's byte code.
	 * 
	 * @return the script's byte code.
	 */
	public byte[] byteCode() { return cw.byteCode(); }

	/**
	 * Returns the script's stack.
	 * 
	 * @return the script's stack
	 */
	public Stack stack() { return stack; }
	
	/**
	 * Sets the script's stack.
	 * 
	 * @param s The stack.
	 */
	public void setStack(Stack s) { this.stack = s; }
	
	/**
	 * Generates the script's assembly code.
	 * @return the script's assembly code.
	 */
	public String asm() {
		return cw.toString();
	}
	
	public boolean equals(Object o) {
		if (o instanceof Script) {
			Script a = (Script)o;
			return a.id() == id;
		} else return false;
	}
	
	
	public String toString() {
		String result = "Script(" + id + "): pc=" + pc + " cond=" + cond + " code:\n";
		result += ArrayToHexString.toString(byteCode()) + "\nservices:\n";
		for (int i = 0; i < services.size(); i++) {
			ServiceSpec ss = services.get(i);
			result += ss.toString() + "\n";
		}
		result += stack.toString();
		return result;
	}
}

