// $Id: AgentSender.java,v 1.9 2005/12/09 07:24:05 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2004, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */
package servilla;

import servilla.msgs.*;
import servilla.msgs.reliable.*;
import servilla.network.*;
import servilla.util.*;
import servilla.compiler.tree.*;
import servilla.service.tree.SpecCompilerException;
import java.util.*;

/**
 * Sends a Servilla script into the wireless sensor network.
 *
 * @author Chien-Liang Fok
 */
public class ScriptSndr implements ServillaConstants, Runnable {
	private Vector<PendingScript> queue = new Vector<PendingScript>();  // a queue of outgoing scripts
	private TimeoutTimer timer = null;	// used for timing out the ACK for the state message.
	private Object lock = new Object(); // lock used when waiting for ACK for state message.
	private ServillaStateAckMsgJ ackMsg; // The ACK received.
	private RSNI rsni; // The network interface.
	private Main main; // The parent component.
	
	/**
	 *  Creates a ScriptSender.
	 */
	public ScriptSndr(Main main, RSNI rsni) {
		this.main = main;
		this.rsni = rsni;
		new Thread(this).start(); // start the thread that sends scripts
	}
	
	/**
	 * Sends an script to the specified destination.  Scripts are sent one at a time.
	 * 
	 * @param dest The TinyOS address of the destination.
	 * @param Script The script to send.
	 */
	public void send(int dest, Script script) {
		queue.add(new PendingScript(dest, script));
		synchronized(queue) {
			try {
				queue.notifyAll();
			} catch(IllegalMonitorStateException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 *  The sender thread sits in a loop sending scripts placed into the queue.
	 * */
	public void run() {
		while(true) {
			synchronized(queue) {  // wait for next script to send
				if (queue.size() == 0) {
					try { queue.wait(); } catch(Exception e) { e.printStackTrace(); }
				}
			}
			if (queue.size() > 0) {
				PendingScript currScript = queue.remove(0);
				if (sendState(currScript) && sendCode(currScript) && sendSpecs(currScript) && sendStack(currScript)) {
					log("Done sending script " + currScript.script.id()); continue;
				}
				log("ERROR: Failed to send script " + currScript.script.id() + "!");
				if (main.showGUI())
					DialogFactory.messageDialog("Error", "Failed to send script " + currScript.script.id() + "!");
			}
		}
	}
	
	private boolean sendState(PendingScript ps) {
		//public ServillaStateMsgJ(int id, long codeSize, long servSize, long sp, short op, 
		//	long pc, long init_sp, long fp, int cond)
		ServillaStateMsgJ msg;
		try {
			msg = new ServillaStateMsgJ(ps.script.id(), ps.script.codeSize(), ps.script.specSize(),
					ps.script.stack().sp(), (short)ISA.smove.byteCode(), ps.script.pc(), 
					ps.script.init_sp(), ps.script.fp(), ps.script.cond()); 
		} catch(SpecCompilerException e) {
			e.printStackTrace();
			return false;
		}
		
		logSendState("StateMsg dest=" + ps.dest+ ":\n" + msg);
		rsni.send(ps.dest, msg.amType(), msg.toByteArray());
		timer = new TimeoutTimer(lock, 2*RS_RETRY_PERIOD*RS_MAX_RXMIT_COUNT*RS_MAX_RETRY_COUNT);
		boolean done = false;
		while (!done) {
			logSendState("Awaiting ACK for state message from " + ArrayToHexString.intToHex(ps.dest) + "...");
			waitForAck();
			if (ackMsg == null) {
				done = true;  // break out of inner while loop
				logSendState("Timed out while waiting for ack from " + ArrayToHexString.intToHex(ps.dest) + "...");
			} else {
				if (ackMsg.id() == ps.script.id()) {
					if (ackMsg.accept()) {
						logSendState("ACK ACCEPTED state message.");
						return true;
					} else {
						logSendState("Received an ACK State but it REJECTED the message.");
						return false;
					}
				} else {
					logSendState("Ack was for the wrong script " + ackMsg.id());
				}
			}
		} // while...
		logSendState("ERROR: Exceeded maximum number of timeouts when sending state.");
		return false;
	}
	
	/**
	 * Sends a script's code.
	 * 
	 * @param ps The script that is being sent.
	 * @return true if the send was successful.
	 */
	private boolean sendCode(PendingScript ps) {
		ServillaCodeMsgJ codeMsg = new ServillaCodeMsgJ(ps.script);
		Debugger.dbg("ScriptSender.sendCode: ", codeMsg.toString());
		return rsni.send(ps.dest, codeMsg.amType(), codeMsg.toByteArray());
	}
	
	/**
	 * Sends a script's service specifications.
	 * 
	 * @param ps The script that is being sent.
	 * @return true if the send was successful, or if the script has no specifications to send.
	 */
	private boolean sendSpecs(PendingScript ps) {
		try {
			if (ps.script.specSize() == 0) return true; // no specs to send
		} catch(SpecCompilerException e) {
			e.printStackTrace();
			return false;
		}
		ServillaSpecMsgJ specMsg = new ServillaSpecMsgJ(ps.script);
		Debugger.dbg("ScriptSender.sendSpecs: ", specMsg.toString());
		byte[] byteArray = specMsg.toByteArray();
		if (byteArray != null)
			return rsni.send(ps.dest, specMsg.amType(), byteArray);
		else
			return false;
	}
	
	/**
	 * Sends a script's stack.
	 * 
	 * @param ps The script that is being sent.
	 * @return true if the send was successful, or if the script has no stack to send.
	 */
	private boolean sendStack(PendingScript ps) {
		if (ps.script.stack().sp() == 0) return true; // no stack to send
		ServillaStackMsgJ stackMsg = new ServillaStackMsgJ(ps.script);
		Debugger.dbg("ScriptSender.sendStack: ", stackMsg.toString());
		return rsni.send(ps.dest, stackMsg.amType(), stackMsg.toByteArray());
		
	}
	
	/**
	 *  Encapsulates an script context and the destination it wants migrate to.
	 */
	private class PendingScript {
		Script script;
		int dest;
		
		/**
		 *  Creates a PendingScript object.  This object encapsulates an
		 *  script that is being sent.
		 *  
		 * @param dest The destination address.
		 * @param script The script to send.
		 */
		public PendingScript(int dest, Script script) {
			this.script = script;
			this.dest = dest;
		}
	}
	
	public String toString() { return "servilla.ScriptSender"; }
	
	private void logSendState(String msg) { Debugger.dbg("ScriptSender.sendState: ", msg);}
	private void log(String msg) {Debugger.dbg("ScriptSender", msg);}
}


