package servilla.compiler;

import java_cup.runtime.*;
import servilla.service.tree.DataType;
import servilla.compiler.tree.*;

%%

%implements java_cup.runtime.Scanner
%function next_token
%type java_cup.runtime.Symbol

%eofval{
  return new Symbol(sym.EOF, null);
%eofval}

%{
  public String sourceFilename;
  private int prevState;
%}

%class ServillaScriptYylex

%line
%char
%state LINECOMMENT, PARCOMMENT

ALPHA=[A-Za-z_]
DIGIT=[0-9]
FILECHAR=[./:]
ALPHA_NUMERIC={ALPHA}|{DIGIT}|"."
STRING={ALPHA}({ALPHA_NUMERIC})*
INTEGER=({DIGIT})+
HEXINT=0x([0-9a-fA-F])+
FLOAT=({DIGIT})+(.{DIGIT}+)
WHITE_SPACE=([\ \n\r\t\f])
EQEQ="=="
EQ="="
LT="<"
GT=">"
LTE="<="
GTE=">="
OR="|"
AND="&"
LOGICAL_OR="||"
LOGICAL_AND="&&"
QUOTE=\"
PLUS="+"
MINUS="-"
MULT="*"
DIV="/"
INCREMENT="++"
DECREMENT="--"
COMMA=","
SEMICOLON=";"
BANG="!"
SEND="send"

RETURN="return"
SERVICE="service"
REACTION="reaction"
VAR_TYPE="void"|"int"|"float"|"string"
LBRACE="{"
RBRACE="}"
LPAREN="("
RPAREN=")"
LBRACKET="["
RBRACKET="]"
INVOKE="invoke"

WHILE="while"
FOR="for"
IF="if"
ELSE="else"

BIND="bind"

%%

<YYINITIAL> {SERVICE} {
  return new Symbol(sym.SERVICE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {REACTION} {
  return new Symbol(sym.REACTION, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {RETURN} {
  return new Symbol(sym.RETURN, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {VAR_TYPE} {
	String typeName = yytext().toUpperCase();
	if (typeName.endsWith("_T")) typeName = typeName.substring(0, typeName.length()-2);
	return new Symbol(sym.VAR_TYPE, new TypeToken(DataType.valueOf(typeName), yyline, yychar));
}

<YYINITIAL> {BIND} {
	return new Symbol(sym.BIND, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {INVOKE} {
	return new Symbol(sym.INVOKE, new StringToken(yytext(), yyline, yychar)); 
}

<YYINITIAL> {FOR} {
	return new Symbol(sym.FOR, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {WHILE} {
	return new Symbol(sym.WHILE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {IF} {
	return new Symbol(sym.IF, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {ELSE} {
	return new Symbol(sym.ELSE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {SEMICOLON} {
	return new Symbol(sym.SEMICOLON, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {COMMA} { 
	return new Symbol(sym.COMMA, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LBRACE} {
	return new Symbol(sym.LBRACE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {RBRACE} {
	return new Symbol(sym.RBRACE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LBRACKET} {
	return new Symbol(sym.LBRACKET, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {RBRACKET} {
	return new Symbol(sym.RBRACKET, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LPAREN} {
	return new Symbol(sym.LPAREN, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {RPAREN} {
	return new Symbol(sym.RPAREN, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {INTEGER}|{HEXINT} {
	return new Symbol(sym.INTEGER, new LongToken(yytext(), yyline, yychar));
}

<YYINITIAL> {FLOAT} {
	return new Symbol(sym.FLOAT32, new FloatToken(yytext(), yyline, yychar));
}

<YYINITIAL> {QUOTE}[^\"]*{QUOTE} {
	String s = yytext();
	s = s.substring(1, s.length());		// remove quote at beginning
	s = s.substring(0,s.length()-1);	// remove quote at end
  	return new Symbol(sym.STRING_VALUE, new StringToken(s, yyline, yychar));
}

<YYINITIAL> {SEND} {
	return new Symbol(sym.SEND, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {PLUS} {
	return new Symbol(sym.PLUS, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {MINUS} {
	return new Symbol(sym.MINUS, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {MULT} {
	return new Symbol(sym.MULT, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {DIV} {
	return new Symbol(sym.DIV, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LT} {
	return new Symbol(sym.LT, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {GT} {
	return new Symbol(sym.GT, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LTE} {
	return new Symbol(sym.LTE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {GTE} {
	return new Symbol(sym.GTE, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {EQ} {
	return new Symbol(sym.EQ, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {BANG}{EQ} {
	return new Symbol(sym.NEQ, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {BANG} {
	return new Symbol(sym.NOT, new StringToken(yytext(), yyline, yychar));
}


<YYINITIAL> {EQEQ} {
	return new Symbol(sym.EQEQ, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {AND} {
	return new Symbol(sym.AND, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {OR} {
	return new Symbol(sym.OR, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LOGICAL_AND} {
	return new Symbol(sym.LOGICAL_AND, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {LOGICAL_OR} {
	return new Symbol(sym.LOGICAL_OR, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {INCREMENT} {
	return new Symbol(sym.INCREMENT, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {DECREMENT} {
	return new Symbol(sym.DECREMENT, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {STRING} {
  	return new Symbol(sym.STRING, new StringToken(yytext(), yyline, yychar));
}

<YYINITIAL> {WHITE_SPACE} { }

<YYINITIAL> "//" {
	yybegin(LINECOMMENT);
}
<LINECOMMENT> [^\n] { }
<LINECOMMENT> [\n] {
	yybegin(YYINITIAL);			// am I allowed to do this??
}

<YYINITIAL> "/*" {
	//System.out.println("Entering PARCOMMENT");
  	yybegin(PARCOMMENT);
}
<PARCOMMENT> [^"*/"]|"*"|"/" { }
<PARCOMMENT> "*/" {
	//System.out.println("Entering YYINITIAL");
	yybegin(YYINITIAL);
}

. {
	System.out.println("Unmatched input: " + yytext());
  	return new Symbol(sym.error, new StringToken(yytext(), yyline, yychar));
}

