package servilla.compiler;

import java.io.*;
import java.util.*;

import servilla.ProgramReader;
import servilla.compiler.tree.*;
import servilla.util.Debugger;
/**
 * The top-level class for compiling a servilla script.
 * 
 * @author Chien-Liang Fok
 */
public class ServillaScriptC {
	boolean optimize;
	
	/**
	 * Creates a servilla script compiler.
	 * @param optimize Whether to optimize the compiled program.
	 */
	public ServillaScriptC(boolean optimize) { this.optimize = optimize; }
	
	/**
	 * Creates a servilla script compiler with all optimizations enabled.
	 */
	public ServillaScriptC() { this(true); }
	
	/**
	 * Tests the scanner generated by JLex.
	 * Prints each token produced by the scanner.
	 * 
	 * @param fileName The name of the script.
	 * @throws IOException
	 * @return A vector of services that the script requires.
	 */
	public Vector<ServiceDecl> scanFile(String fileName) throws IOException {
		Reader programReader = new StringReader(servilla.ProgramReader.readProgram(fileName));
		ServillaScriptYylex scanner = new ServillaScriptYylex(programReader);
		boolean service = false;
		
		Vector<ServiceDecl> result = new Vector<ServiceDecl>();
		
		java_cup.runtime.Symbol nextToken = scanner.next_token();
		do {
			if (nextToken.value instanceof Token) {				
				log("Type: " + name(nextToken.sym) + ",\tToken: " + nextToken.value);
				if (nextToken.sym == sym.REQUIRE) 
					service = true;						// record all required services
				else if (service) {
					service = false;
					if (nextToken.sym == sym.STRING) {
						StringToken token = (StringToken)nextToken.value;
						result.add(new ServiceDecl(token.text(), token.lineno(), token.charno()));				
					}
				}
			} else
				log("Type: " + name(nextToken.sym) + ",\t" + nextToken);
			nextToken = scanner.next_token();
		} while (nextToken.sym != sym.EOF);	
		
		return result;
	}

	/**
	 * Compiles a Servilla script.
	 * 
	 * @param fileName The name of the file that contains the script.
	 * @throws IOException
	 * @throws AssemblerException
	 */
	public CodeWriter compileFile(String fileName) throws IOException, AssemblerException {
		String programText = ProgramReader.readProgram(fileName);
		return compile(fileName, programText);
	}
	
	/**
	 * Compile a script's text.
	 * 
	 * @param programText
	 * @return
	 * @throws IOException
	 * @throws AssemblerException
	 */
	public CodeWriter compile(String fileName, String programText) throws IOException, AssemblerException {
		ServillaScriptParser parser = new ServillaScriptParser(fileName, new ServillaScriptYylex(new StringReader(programText)));
		parser.setProgramText(programText);		// this is to allow the parser to generate a meaningful error message.
		CodeWriter cw = null;
		try {
			parser.parse();
			ScriptRoot ps = parser.getProgramScript();
			if (ps != null) {
				cw  = ps.assemble();
				if (optimize) cw.optimize();
				if (Debugger.debug)
					System.out.println("/*\n * Here's the entire program tree:\n * "+parser.getProgramScript().toString() + "\n */");
			}
		} catch (IOException e) {
			throw e;
		} catch(AssemblerException e) {
			logError("Error on line " + e.lineno() + ": " + e.getMessage() + " near:");
			logError(servilla.util.CompileErrorUtils.getErrorLine(programText, e.lineno(), e.charno()));
			throw e;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return cw;
	}
	
	/**
	 * Uses reflection to find the name of a constant defined in sym.java.
	 * 
	 * @param type the integer constant
	 * @return The name of the constant as defined in sym.java.
	 */
	private String name(int type) {
		try {
			Class c = Class.forName("servilla.compiler.sym");
			java.lang.reflect.Field[] fields = c.getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
				if (fields[i].getInt(c) == type)
					return fields[i].getName();
			}
		} catch(Exception e) {
			e.printStackTrace();
		}
		return "UNKNOWN";
	}
	
	private void log(String msg) {
		System.out.println(msg);
	}
	
	private void logError(String msg) {
		System.err.println(msg);
	}


	/**
	 * @param args
	 */
	public static void main(String[] args) {
		boolean scan = false;

		for (int i = 0; i < args.length; i++) {
			if(args[i].equals("-scan"))
				scan = true;
			else if (args[i].equals("-d"))
				Debugger.debug = true;
		}

		try {
				String fileName = args[args.length-1];
				ServillaScriptC compiler = new ServillaScriptC();
				if (scan)
					compiler.scanFile(fileName);
				else {
					CodeWriter cw = compiler.compileFile(fileName);
					System.out.println(cw.toString());
					System.out.println(servilla.util.ArrayToHexString.toString(cw.byteCode()));
				}
			} catch(Exception e) {
				e.printStackTrace();
			}
		
	}
}
