// $Id: AgentInjector.java,v 1.10 2005/12/13 00:18:42 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2004, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */
package servilla.compiler.tree;

import servilla.service.tree.DataType;

/**
 * An assignment expression assigns the results of an expression
 * to a variable.
 * 
 * @author Chien-Liang Fok
 *
 */
public class AssignmentExpr extends Expr {
	/**
	 * The variable being written to.
	 */
	private VarRef varRef = null;
	
	/**
	 * The expression generating the value.
	 */
	private Expr expr;
	
	/**
	 * Create an assignment expression in which the left-hand side is a variable reference.
	 * 
	 * @param varRef The left-hand side of an assignment expression
	 * @param expr The right-hand side of an assignment expression
	 * @param lineno
	 * @param charno
	 */
	public AssignmentExpr(VarRef varRef, Expr expr, int lineno, int charno) {
		super(lineno, charno);
		this.varRef = varRef;
		this.expr = expr;
	}
	
	/**
	 * Generates assembly code for the assignment statement.
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		DataType destType = varRef.retType(cw), retType = expr.retType(cw);
		if (destType == DataType.STRING) {
			
		} else {
			
		}
		
		if (!destType.equals(retType)) expr = new CastExpr(expr, destType); // This will throw an exception if the cast is impossible.
		expr.assemble(cw); // compile the expression (this leaves the results on top of the stack)
		
		// duplicate the results so a copy can be left on the stack.
		cw.add(new Instruction(ISA.duplicate, lineno, charno).setParam4(destType.stackSize())
				.setComment("AssignmentExpr: Leave a copy of " + destType.stackSize() + "byte variable on the stack"));
		
		saveResults(cw, varRef, lineno, charno); // save results
	}
	
	/**
	 * Saves the results on top of the stack into the local variable specified by
	 * varRef.
	 * 
	 * @param cw The code writer.
	 * @param varRef The variable reference where the results should be saved.
	 * @throws AssemblerException
	 */
	public static void saveResults(CodeWriter cw, VarRef varRef, int lineno, int charno) throws AssemblerException {
		Type destType = varRef.retType(cw);
		varRef.pushAddress(cw); // do not assemble b/c that would fetch the value of the referenced variable
		if (destType.isArray()) {
			switch(destType.type().size()) {
			case 2: cw.add(new Instruction(ISA.slv2, lineno, charno).setComment("AssignmentExpr: Saving 2-byte result into local memory.")); break;
			case 4: cw.add(new Instruction(ISA.slv4, lineno, charno).setComment("AssignmentExpr: Saving 4-byte result into local memory.")); break;
			default: 
				if (destType.type() != DataType.STRING) throw new AssemblerException("Invalid destination size: " + destType.type().size(), lineno, charno);
				cw.add(new Instruction(ISA.slvString, lineno, charno).setComment("AssignmentExpr: Saving result into local memory."));
			}
		} else {
			int numElements = destType.numElements();
			switch(destType.type().size()) {
			case 2: cw.add(new Instruction(ISA.slv2a, lineno, charno).setParam4(numElements).setComment("AssignmentExpr: Saving " + numElements + "-element array of 2-byte values into local memory.")); break;
			case 4: cw.add(new Instruction(ISA.slv4a, lineno, charno).setParam4(numElements).setComment("AssignmentExpr: Saving " + numElements + "-element array of 4-byte values into local memory.")); break;
			default: 
				if (destType.type() != DataType.STRING) throw new AssemblerException("Invalid destination size: " + destType.type().size(), lineno, charno);
				cw.add(new Instruction(ISA.slvStringa, lineno, charno).setParam4(numElements).setComment("AssignmentExpr: Saving " + numElements + "-element array of strings into local memory."));
			}
		}
	}
	
	public String toString() { return "(AssignmentExpr " + varRef + " " + expr + ")"; }

	@Override
	public DataType retType(CodeWriter cw) throws AssemblerException { return varRef.retType(cw); }

	@Override
	public int retSize(CodeWriter cw) throws AssemblerException { return varRef.retHeapSize(cw); }
	
	@Override
	public long initAddr(long startAddr) {
		long result = startAddr;
		if (expr != null) result = expr.initAddr(result);
		return result;
	}

	@Override
	public long localVarSize() { return expr.localVarSize(); }
}
