package servilla.compiler.tree;

/**
 * Encapsulates a binary expression. 
 * 
 * Binary expressions are those that involve an operation performed
 * between two expressions.
 * 
 * Ex: i + k;
 * 
 * @author liang
 *
 */
public class BinaryExpr extends Expr {
	public enum BinaryExprType {PLUS, MINUS, MULT, DIV, LOGICAL_OR, LOGICAL_AND, AND, OR};
	
	private BinaryExprType type;
	private Expr leftExpr, rightExpr;
	
	public BinaryExpr(BinaryExprType type, Expr leftExpr, Expr rightExpr, int lineno, int charno) {
		super(lineno, charno);
		this.type = type;
		this.leftExpr = leftExpr;
		this.rightExpr = rightExpr;
	}
	
	/**
	 * Generates assembly code for the assignment statement.
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		Type leftType = retType(cw); // this will throw an exception if the left and right types do not match
		leftExpr.assemble(cw); 	// ends w/ result of right exp on top of stack
		rightExpr.assemble(cw); // ends w/ result of left exp on top of stack
		
		ISA instruction = null;
		if (type == BinaryExprType.PLUS) {
			if (leftType.isInt()) {
				instruction = ISA.addInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.addFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.MINUS) {
			if (leftType.isInt()) {
				instruction = ISA.subInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.subFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.MULT) {
			if (leftType.isInt()) {
				instruction = ISA.multInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.multFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.DIV) {
			if (leftType.isInt()) {
				instruction = ISA.divInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.divFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.LOGICAL_OR) {
			if (leftType.isInt()) {
				instruction = ISA.lorInt;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
			//cw.addLine("lor");
		} else if (type == BinaryExprType.LOGICAL_AND) {
			if (leftType.isInt()) {
				instruction = ISA.landInt;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.AND) {
			if (leftType.isInt()) {
				instruction = ISA.andInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.andFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else if (type == BinaryExprType.OR) {
			if (leftType.isInt()) {
				instruction = ISA.orInt;
			} else if (leftType.isFloat()) {
				instruction = ISA.orFloat;
			} else throw new AssemblerException("Cannot add variables of type: " + leftType, lineno, charno);
		} else {
			throw new AssemblerException("Unknown binary expression type: " + type, lineno, charno);
		}
		cw.add(new Instruction(instruction, lineno, charno));
	}
	
	public String toString() { return "(BinaryExpr: " + leftExpr + " " + type.toString() + " " + rightExpr + ")"; }

	@Override
	public long initAddr(long startAddr) {
		startAddr = leftExpr.initAddr(startAddr);
		startAddr = rightExpr.initAddr(startAddr);
		return startAddr;
	}
	
	/**
	 * Checks wether the types are compatible.  If they are, adds the necessary
	 * cast expressions to make them identical.
	 * 
	 * @param cw The CodeWriter.
	 * @throws AssemblerException
	 */
	private void checkTypes(CodeWriter cw) throws AssemblerException {
		Type leftType = leftExpr.retType(cw), rightType = rightExpr.retType(cw);
		
		if (!(leftType.isInt() || leftType.isFloat())) throw new AssemblerException("BinaryExpr Invalid left-operand type: " + leftType, leftExpr.lineno(), leftExpr.charno());
		if (!(rightType.isInt() || rightType.isFloat())) throw new AssemblerException("BinaryExpr Invalid right-operand type: " + rightType, rightExpr.lineno(), rightExpr.charno());
		
		if (!leftType.equals(rightType)) {
			// If one side is a float, cast the other side into a float
			if (leftType.isFloat())			rightExpr	= new CastExpr(rightExpr, Type.FLOAT32);
			else if (rightType.isFloat())	leftExpr	= new CastExpr(leftExpr, Type.FLOAT32);
			
			// If one side is INT32, cast the other side into an int32_t
			else if (leftType.equals(Type.INT32)) rightExpr	= new CastExpr(rightExpr, Type.INT32);
			else if (rightType.equals(Type.INT32)) leftExpr	= new CastExpr(leftExpr, Type.INT32);
			
			// If one side is an int16, cast the other side into an int16_t
			else if (leftType.equals(Type.INT16)) rightExpr = new CastExpr(rightExpr, Type.INT16);
			else if (rightType.equals(Type.INT16)) leftExpr = new CastExpr(leftExpr, Type.INT16);
		}
		
		leftType = leftExpr.retType(cw); rightType = rightExpr.retType(cw);
		if (!leftType.equals(rightType)) throw new AssemblerException("BinaryExpr Type conflict: " + leftType + " vs. " + rightType, lineno, charno);
	}

	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { checkTypes(cw); return leftExpr.retType(cw); }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return retType(cw).heapSize(); }

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { checkTypes(cw); return leftExpr.retStackSize(cw); }
	
	@Override
	public long localVarSize() { return leftExpr.localVarSize() + rightExpr.localVarSize(); }
}
