// $Id: AgentInjector.java,v 1.10 2005/12/13 00:18:42 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2004, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */
package servilla.compiler.tree;

import java.util.Vector;

/**
 * A bind expression initiates the process of binding a service specification to a service.
 * It has an uint16_t return value that indicates whether the bind operation was successful.
 * If it is successful, the return value is 1, otherwise it is 0.
 * 
 * @author Chien-Liang Fok
 */
public class BindExpr extends Expr {
	private Expr semantics, hops;
	private String servSpecName;
	
	/**
	 * Creates a BindExpr.  
	 * 
	 * @param servSpecName The name of the specification being bound.
	 * @param semantics The binding semantics.  This expression should evaluate to a 16-bit integer
	 * whose bits match those defined in servilla.compier.tree.BindType.java.
	 * @param hops An expression that evaluates to the number of hops to search for a matching
	 * service.
	 * @param lineno The line number of the program code that this expression maps to.
	 * @param charno The character number of the program code that this expression maps to.
	 */
	public BindExpr(String servSpecName, Expr semantics, Expr hops, int lineno, int charno) {
		super(lineno, charno);
		this.servSpecName = servSpecName;
		this.semantics = semantics;
		this.hops = hops;
	}
	
	/**
	 * The binding assembly code is:
	 * 
	 * <pre>
	 * [expression that returns the hops]
	 * [expression that returns the semantics]
	 * push2 handle
	 * bind
	 * </pre>
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		ServiceDecl s = cw.getServiceDecl(servSpecName);
		if (s == null) {
			String errMsg = "Undefined service spec \"" + servSpecName + "\" on line " + lineno + "\n";
			errMsg += "\tKnown service specs:\n";
			Vector<ServiceDecl> services = cw.getServiceDecls();
			for (int i = 0; i < services.size(); i++) {
				errMsg+= "\t\t" + services.get(i).name() + "\n";
			}
			throw new AssemblerException(errMsg, lineno, charno);
		}
		if (!s.hasHandle()) s.setHandle(cw.nonce()); // set a unique number to the service so it can be referred to
			
		// Check the return types of the semantics and hops expressions.
		if (!hops.retType(cw).equals(Type.UINT16))			hops = new CastExpr(hops, Type.UINT16);
		if (!semantics.retType(cw).equals(Type.UINT16))		hops = new CastExpr(semantics, Type.UINT16);
			
		hops.assemble(cw); semantics.assemble(cw);
		cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(s.handle()));
		cw.add(new Instruction(ISA.bind, lineno, charno));
	}
	
	public String toString() { return "(BindExpr: "+servSpecName+", "+semantics+", "+hops+")"; }

	@Override
	public long initAddr(long startAddr) {
		startAddr = hops.initAddr(startAddr);
		startAddr = semantics.initAddr(startAddr);
		return startAddr;
	}

	/**
	 * The bind operation does not return anything on the stack.
	 */
	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return Type.VOID; }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return 0; }

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return 0; }

	@Override
	public long localVarSize() { return semantics.localVarSize() + hops.localVarSize(); }
}
