package servilla.compiler.tree;

import java.util.*;

import servilla.service.tree.DataType;

/**
 * A call expression may execute a system-defined method, or make a function call.
 * 
 * @author Chien-Liang Fok
 *
 */
public class CallExpr extends Expr {
	
	/**
	 * The name of the method being called.
	 */
	private String methodName;
	
	/**
	 * The input parameters of the method.
	 */
	private Vector<Expr> params;
	
	/**
	 * Maps system-defined methods to their assembly implementation.
	 * The key is the name of the method, and the value is the instruction(s) that should be executed. when
	 * the function is called.
	 */
	private Hashtable<String, SysMethod> sysMethodTable = new Hashtable<String, SysMethod>();
	
	/**
	 * The constructor.
	 * 
	 * @param methodName The name of the method.
	 * @param params The input parameters.
	 * @param lineno The line number.
	 * @param charno The character number.
	 */
	public CallExpr(String methodName, Vector<Expr> params, int lineno, int charno) {
		super(lineno, charno);
		
		this.methodName = methodName;
		this.params = params;

		Vector<DataType> noparam = new Vector<DataType>();
		Vector<DataType> intparam = new Vector<DataType>(); intparam.add(DataType.INT);
		
		// Define the system calls.  These are instructions that scripts can directly access.
		// The name of the instruction is mapped to a SysMethod object, which defines the
		// system call's input and output parameters.

		sysMethodTable.put("sleep", 		new SysMethod(ISA.sleep, 		intparam, 		DataType.INT));
			
		sysMethodTable.put("exit", 			new SysMethod(ISA.halt,			noparam,		DataType.VOID));	// terminate an script
		sysMethodTable.put("address", 		new SysMethod(ISA.addr,			noparam,		DataType.INT));	// fetch the host address
		sysMethodTable.put("id", 			new SysMethod(ISA.id,			noparam,		DataType.INT));	// fetch the script ID
		sysMethodTable.put("rand", 			new SysMethod(ISA.rand,			noparam,		DataType.INT));	// generate a signed random number
//		sysMethodTable.put("condition", 	new SysMethod(ISA.cond,			noparam,		DataType.INT));	// fetch the condition code -- do we need this?
		sysMethodTable.put("numNbrs", 		new SysMethod(ISA.numnbrs, 		noparam,		DataType.INT));	// fetch the number of neighbors
		sysMethodTable.put("randNbr", 		new SysMethod(ISA.randnbr_c,	noparam,		DataType.INT));	// fetch the address of a random neighbor
		
		sysMethodTable.put("getNbr", 		new SysMethod(ISA.getnbr_c, 	intparam,		DataType.INT));	// fetch a specific neighbor, param = neighbor number (0-numNeighbors-1)
		sysMethodTable.put("isNbr", 		new SysMethod(ISA.isnbr_c, 		intparam,		DataType.VOID));
		
		// migration operations
		sysMethodTable.put("strongMove", 	new SysMethod(ISA.smove,		intparam,		DataType.VOID));	
		sysMethodTable.put("weakMove", 		new SysMethod(ISA.wmove, 		intparam,		DataType.VOID));			
		sysMethodTable.put("strongClone", 	new SysMethod(ISA.sclone, 		intparam,		DataType.VOID));
		sysMethodTable.put("weakClone", 	new SysMethod(ISA.wclone, 		intparam,		DataType.VOID));
		
		sysMethodTable.put("toggleLED", 	new SysMethod(ISA.toggleled, 	intparam,		DataType.VOID));	// TODO: make LEDs accessible via instructions
		sysMethodTable.put("putLED", 		new SysMethod(ISA.putled, 		intparam,		DataType.INT));  
		
		//sysMethodTable.put("unbind", 		new SysMethod(ISA.unbind,		intparam,	DataType.VOID)); // TODO modify compiler to check whether parameter is valid
	}

	/**
	 * Assembles a call expression.
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		if (sysMethodTable.containsKey(methodName)) {  // If this is a system-defined method
			SysMethod meth = sysMethodTable.get(methodName);
			assembleParams(cw, meth.inputParams);
			cw.add(new Instruction(meth.op, lineno, charno));
		} else {
			String retLabel = "RETURN" + cw.nonce(); // generate the return location
			
			// Get the method declaration to check parameter type compatability
			MethodDecl methodDecl = cw.getRoot().getMethod(methodName, lineno, charno);
			assembleParams(cw, methodDecl.paramTypes());	// push the parameters on the stack
			cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(methodName).setComment("Push address of destination method " + methodName));
			cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(cw.getMethod(methodName, lineno, charno).frameSize()));
			cw.add(new Instruction(ISA.call, lineno, charno).setComment("Call method " + methodName));	// jump to the method
			cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Return location of method " + methodName).setLabel(retLabel)); // return location
		}

	}
	
	/**
	 * Assemble each of the parameters.
	 * 
	 * @param cw The code writer.
	 * @throws AssemblerException
	 */
	private void assembleParams(CodeWriter cw, Vector<DataType> inputParams) throws AssemblerException {
		// assemble each of the parameters
		Enumeration<Expr> e = params.elements();
		int index = 0;
		while(e.hasMoreElements()) {
			Expr currParam = e.nextElement();
			Type retType = currParam.retType(cw), specType = inputParams.get(index++);
			if (!retType.equals(specType))
				throw new AssemblerException("Parameter type mismatch, got " + retType + " expected " + specType, lineno, charno);
			currParam.assemble(cw);
		}
	}
	
	public String toString() {
		StringBuffer pList = new StringBuffer();
		Enumeration e = params.elements();
		while (e.hasMoreElements()) {
			pList.append(e.nextElement());
			if (e.hasMoreElements())
				pList.append(", ");
		}
		return "(CallExpr: " + methodName + "(" + pList + "))";
	}

	@Override
	public long initAddr(long startAddr) {
		Enumeration<Expr> e = params.elements();
		while (e.hasMoreElements()) {
			Expr s = e.nextElement();
			startAddr = s.initAddr(startAddr);
		}
		return startAddr;
	}
	
	@Override
	public long localVarSize() {
		long result = 0;
		Enumeration<Expr> pv = params.elements();
		while (pv.hasMoreElements()) {
			result += pv.nextElement().localVarSize();
		}
		return result;
	}

	@Override
	public int retSize(CodeWriter cw) throws AssemblerException {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public DataType retType(CodeWriter cw) throws AssemblerException {
		if (sysMethodTable.containsKey(methodName)) {  // check whether it is a system-defined method
			SysMethod meth = sysMethodTable.get(methodName);
			return meth.retVal;
		} else {
			return cw.getRoot().getMethod(methodName, params, lineno, charno).returnType();
		}
	}
	
	/**
	 * Maps an instruction to its input and output parameters.
	 * 
	 * @author Chien-Liang Fok
	 *
	 */
	private class SysMethod {
		ISA op;
		DataType retVal;
		Vector<DataType> inputParams;
		
		public SysMethod(ISA op, Vector<DataType> inputParams, DataType retVal) { 
			this.op = op; this.inputParams = inputParams; this.retVal = retVal; 
		}
	}
}
