package servilla.compiler.tree;

/**
 * Casts one type to another.  This can only be done on primitive data types, no arrays.
 * 
 * @author Chien-Liang Fok
 *
 */
public class CastExpr extends Expr {
	/**
	 * The expression whose return type is being cast.
	 */
	private Expr expr;
	
	/**
	 * The type to cast the return value into.
	 */
	private Type destType;
	
	/**
	 * Creates a CastExpr.
	 * 
	 * @param expr The expression whose output must be casted to another type
	 * @param destType The type to cast it to.
	 */
	public CastExpr(Expr expr, Type destType) throws AssemblerException {
		super(expr.lineno(), expr.charno());
		this.expr = expr;
		this.destType = destType;
		if (destType.isArray()) throw new AssemblerException("Attemping to cast as an array", lineno, charno);
	}
	
	@Override
	public long initAddr(long startAddr) { return expr.initAddr(startAddr); }

	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return destType; }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return retType(cw).heapSize(); }

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return retType(cw).stackSize(); }

	@Override
	public String toString() { return "(CastExpr " + destType + " " + expr + ")"; }

	/*
	 * All integers consume the same amount of space on the stack and are encoded in the
	 * same format, thus they can be cast into each other without changing the actual 
	 * bytes in the stack.
	 * 
	 * However, casting from integers to floating point and vice-versa requires changing 
	 * the bytes.
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		expr.assemble(cw); // first assemble the expression
		Type retType = expr.retType(cw);
		if (retType.isArray()) throw new AssemblerException("Attemping to cast an array", lineno, charno);
		if (!retType.equals(destType)) {
			if (retType.equals(Type.UINT16)) {
				if (destType.equals(Type.FLOAT32)) 	cw.add(new Instruction(ISA.castUint16toF, lineno, charno));
				else if (!destType.isInt())			throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
			} else if (retType.equals(Type.INT16)) {
				if (destType.equals(Type.FLOAT32))	cw.add(new Instruction(ISA.castInt16toF, lineno, charno));
				else if (!destType.isInt())			throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
			} else if (retType.equals(Type.INT32)) {
				if (destType.equals(Type.FLOAT32))	cw.add(new Instruction(ISA.castInt32toF, lineno, charno));
				else if (!destType.isInt())			throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
			} else if (retType.equals(Type.UINT32)) {
				if (destType.equals(Type.FLOAT32))	cw.add(new Instruction(ISA.castUint32toF, lineno, charno));
				else if (!destType.isInt())			throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
			} else if (retType.equals(Type.FLOAT32)) {
				if (destType.isInt())				cw.add(new Instruction(ISA.castFtoInt32, lineno, charno));
				else if (!destType.isInt())			throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
			} else throw new AssemblerException("Unable to cast " + retType + " to " + destType, lineno, charno);
		}
	}
	
	@Override
	public long localVarSize() { return expr.localVarSize(); }

}
