package servilla.compiler.tree;

import java.util.*;

/**
 * Serves as a shared object that all the other compiler
 * components use to write the assembly code.
 * 
 * @author Chien-Liang Fok
 *
 */
public class CodeWriter {
	//private SymbolTable varSymbolTable = new SymbolTable(); // maps variables (including services) to their number
	
	private int nonce = 0;
	
	/**
	 * These are the instructions.
	 */
	private Vector<Instruction> iv = new Vector<Instruction>();
	
	private ScriptRoot root;
	
	private Vector<Frame> frameStack = new Vector<Frame>();	// for holding the local variables in the current frame.
	
	private long initSP; // The initial stack pointer
	
	private Vector<MethodDecl> methodStack = new Vector<MethodDecl>();
	
	/**
	 * Creates an empty codeWriter.
	 */
	public CodeWriter() {}
	
	/**
	 * Pushes a method declaration onto the method stack.  This is used
	 * by the return statement to determine whether its return type matches
	 * the return type specified by the method declaration.
	 * 
	 * @param md The method declaration.
	 */
	public void pushMethod(MethodDecl md) { methodStack.add(md); }
	
	/**
	 * Returns the current method declaration.
	 * 
	 * @return The current method declaration.
	 */
	public MethodDecl peekMethod() { return methodStack.get(methodStack.size()-1); }
	
	/**
	 * Pops the current method declaration off the stack.
	 *
	 */
	public void popMethod() { methodStack.remove(methodStack.size()-1); }
	
	/**
	 * Returns a unique number for distinguishing return locations.
	 * 
	 * @return A unique number each time this method is called.
	 */
	public int nonce() { return nonce++; }
	
	/**
	 * Returns the initial stack pointer value.  This value accounts for the
	 * number of bytes necessary to hold the local variables within Main.
	 * 
	 * @return Returns the initial stack pointer value
	 */
	public long initSP() { return initSP; }
	
	/**
	 * Sets the initial stack pointer value.  This value accounts for the
	 * number of bytes necessary to hold the local variables within Main.
	 */
	public void setInitSP(long initSP) { this.initSP = initSP; }
	
	/**
	 * Look through the instructions and optimize it!
	 *
	 * Possible optimizations:
	 * 
	 * 1) [copy] [pushUint16] [setvar] [pop] can be reduced to [pushUint16] [setvar] if
	 * 		none of the instructions have labels.
	 * 2) [copy] [inc/dec] [pushUint16] [setvar] [pop] can be reduced to [inc/dec][pushUint16] [setvar] if
	 * 		none of the instructions have labels.
	 * 3) A noop with a label can be removed if it is followed by another instruction that does not have a label.
	 * 		A noop without a label can be removed entirely.
	 * 4) [pushUing16] [pop] can be eliminated.
	 */
	public CodeWriter optimize() throws AssemblerException {
		/*for (int i = 0; i < iv.size(); i++) {
			Instruction ci = iv.get(i);
			if (ci.opcode() == ISA.copy && !ci.hasLabel()) {
				Instruction ni = iv.get(i+1);
				if (ni.opcode() == ISA.pushUint16 && !ni.hasLabel()) {
					Instruction nni = iv.get(i+2);
					if (nni.opcode() == ISA.setvar && !nni.hasLabel()) {
						Instruction nnni = iv.get(i+3);
						if (nnni.opcode() == ISA.pop && !nnni.hasLabel()) {
							iv.remove(i+3);
							iv.remove(i);
							i--;
						}
					}
				}
			}
		}
		
		for (int i = 0; i < iv.size(); i++) {
			Instruction ci = iv.get(i);
			if (ci.opcode() == ISA.copy && !ci.hasLabel()) {
				Instruction ni = iv.get(i+1);
				if ((ni.opcode() == ISA.inc || ni.opcode() == ISA.dec) && !ni.hasLabel()) {
					Instruction nni = iv.get(i+2);
					if (nni.opcode() == ISA.pushUint16 && !nni.hasLabel()) {
						Instruction nnni = iv.get(i+3);
						if (nnni.opcode() == ISA.setvar && !nnni.hasLabel()) {
							Instruction nnnni = iv.get(i+4);
							if (nnnni.opcode() == ISA.pop && !nnnni.hasLabel()) {
								iv.remove(i+4);
								iv.remove(i);
								i--;
							}
						}
					}
				}
			}
		}*/
		
		for (int i = 0; i < iv.size(); i++) {
			Instruction ci = iv.get(i);
			if (ci.opcode() == ISA.noop) {
				if (ci.hasLabel()) {
					Instruction ni = iv.get(i+1);
					if (!ni.hasLabel()) {
						ni.setLabel(ci.label());
						iv.remove(i);
						i--;
					} else {
						// TODO: If the instruction after the noop has a label
						// search for all references to the noop label and change
						// it to be the following instruction's label, and remove
						// the noop
					}
				} else {
					iv.remove(i);
					i--;
				}
			}
		}
		
		/*for (int i = 0; i < iv.size(); i++) {
			Instruction ci = iv.get(i);
			if (ci.opcode() == ISA.pushUint16 && !ci.hasLabel()) {
				Instruction ni = iv.get(i+1);
				if (ni.opcode() == ISA.pop && !ni.hasLabel()) {				
					iv.remove(i+1);
					iv.remove(i);
					i--;								
				}
			}
		}*/
		
		expandLabelRefs();  // re-assign all of the label references
		return this;
	}
	
	/**
	 * Sets the script root.
	 * 
	 * @param s The script root
	 */
	public void setRoot(ScriptRoot root) { this.root = root; }
	
	public ScriptRoot getRoot() { return root; }
	
	/**
	 * Returns the service decl of the service with the specified name.
	 * @param name
	 * @return
	 */
	public ServiceDecl getServiceDecl(String name) {
		Enumeration<ServiceDecl> e = services.elements();
		while (e.hasMoreElements()) {
			ServiceDecl s = e.nextElement();
			if (s.name().equals(name))
				return s;
		}
		return null;
	}
	
	/**
	 * Adds an instruction.
	 * 
	 * @param i The instruction to add.
	 */
	public void add(Instruction i) { iv.add(i); }
	
	/**
	 * Adds a sequence of instructions.
	 * 
	 * @param i A sequence of instructions
	 */
	public void add(Vector<Instruction> i) {
		Enumeration<Instruction> e = i.elements();
		while (e.hasMoreElements()) {
			add(e.nextElement());
		}
	}
	
	/**
	 * Creates a new empty frame.
	 */
	public void newFrame() { frameStack.add(new Frame()); }
	
	/**
	 * Deletes the current frame.
	 */
	public void deleteFrame() { frameStack.remove(frameStack.size()-1); }
	
	/**
	 * Adds a local variable declaration to the current frame.
	 * 
	 * @param var The variable declaration to add.
	 * @throws AssemblerException If a variable of the same name 
	 * has already been added.
	 */
	public void addLocalVar(VarDeclExpr var, int lineno, int charno) throws AssemblerException {
		if (currFrame().get(var.name()) != null) 
			throw new AssemblerException("Duplicate variable " + var.name(), lineno, charno);
		currFrame().add(var); 
	}
	
	/**
	 * Returns the current frame.
	 * @return the current frame.
	 */
	private Frame currFrame() { return frameStack.get(frameStack.size()-1); }
	
	/**
	 * Returns the local variable's declaration.  It searches in the current
	 * frame. 
	 * 
	 * @param The name of the variable.
	 * @param lineno The line number of the program code that is being compiled.
	 * @param charno The character number of the program code that is being compiled.
	 * @return The variable's declaration.
	 * @throws AssemblerException if the variable is not found in the current frame.
	 */
	public VarDeclExpr getLocalVar(String ref, int lineno, int charno) throws AssemblerException {
		VarDeclExpr result = currFrame().get(ref);
		if (result == null)  //search through the list of specs for the name
			throw new AssemblerException("Unknown variable \"" + ref + "\"", lineno, charno);
		return result;
	}
	

	
	/**
	 * Translates all label refs into their corresonding addresses.
	 *
	 */
	public void expandLabelRefs() throws AssemblerException {
		// first traverse through the list assigning the correct program counter
		Enumeration<Instruction> e = iv.elements();
		int pc = 0;
		while (e.hasMoreElements()) {
			Instruction i = e.nextElement();
			i.setAddress(pc);
			pc += i.numBytes();
		}
		
		// now create a hashtable that maps labels to their addresses
		Hashtable<String, Long> labelTable = new Hashtable<String, Long>();
		e = iv.elements();
		while (e.hasMoreElements()) {
			Instruction i = e.nextElement();
			if (i.hasLabel()) {
				if (labelTable.containsKey(i.label()))
					throw new AssemblerException("ERROR: Duplicate method names \"" + i.label() + "\"", i.lineno(), i.charno());
				labelTable.put(i.label(), new Long(i.address()));
			}
		}
		
		// finally go through and replace the label references with their values
		e = iv.elements();
		while (e.hasMoreElements()) {
			Instruction i = e.nextElement();
			if (i.hasLabelRef()) {
				Long refpc = labelTable.get(i.labelRef());
				if (refpc == null)
					throw new AssemblerException("ERROR: Unknown label reference \"" + i.labelRef() + "\" on line " + i.lineno(), i.lineno(), i.charno());
				else {
					i.setParam4(refpc.longValue());
				}
			}
		}
	}
	
	/**
	 * Generates the bytecode that will be sent to the motes.
	 * 
	 * @return The byte code.
	 */
	public byte[] byteCode() {
		Vector<Byte> v = new Vector<Byte>();
		
		Enumeration<Instruction> e = iv.elements();
		while (e.hasMoreElements()) {
			Instruction i = e.nextElement();
			v.addAll(i.byteCode());
		}
		byte[] result = new byte[v.size()];
		for (int i = 0; i < result.length; i++) {
			result[i] = v.get(i).byteValue();
		}
		return result;
	}
	
	/**
	 * Generates NesC code that can be copied into the DefaultAgent.txt file
	 * for including a default agent.
	 * 
	 * @return NesC code that contains the agent's code.
	 */
	public String nesCCode() {
		Vector<String> v = new Vector<String>();
		Enumeration<Instruction> e = iv.elements();
		while (e.hasMoreElements()) {
			Instruction i = e.nextElement();
			v.addAll(i.nesCCode());
		}
		String result = "";
		for (int i = 0; i < v.size(); i++) {
			result +=v.get(i);
			if (i < v.size()-1)
				result += " ";
		}
		return result;
	}
	
	/**
	 * Prints an agent's assembly code.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration<Instruction> e = iv.elements();
		while (e.hasMoreElements()) {
			sb.append("// " + e.nextElement().toString() + "\n");
		}
		return sb.toString();
	}
	
	/**
	 * Holds the local variables used in a method.  Every time a method is called,
	 * a new frame is pushed onto the stack for holding the method's local variables.
	 * This class keeps track of what local variables are instantiated and used by 
	 * a method.
	 * 
	 * @author Chien-Liang Fok
	 */
	class Frame {
		private Hashtable<String, VarDeclExpr> table = new Hashtable<String, VarDeclExpr>();

		/**
		 * Creates an empty frame.
		 */
		public Frame() {}
		
		/**
		 * Adds a local variable declaration to the current frame.
		 * 
		 * @param var The variable to add.
		 */
		public void add(VarDeclExpr var) { table.put(var.name(), var); }
		
		/**
		 * Gets the variable with the specified name, or null if no such variable exists.
		 * 
		 * @param name The name of the variable.
		 * @return The variable's declaration, or null if it does not exist.
		 */
		public VarDeclExpr get(String name) { return table.get(name); }
		
		/**
		 * Returns the size of the local variables in the frame.
		 * 
		 * @return the size of the local variables in the frame.
		 */
		public int size() {
			int result = 0;
			Enumeration<VarDeclExpr> e = table.elements();
			while (e.hasMoreElements()) {
				result += e.nextElement().localVarSize();
			}
			return result;
		}
	}
}
