package servilla.compiler.tree;

/**
 * Encapsulates a conditional expression.
 * 
 * Ex: i + k;
 * 
 * @author Chien-Liang Fok
 *
 */
public class CondExpr extends Expr {
	public enum CondExprType {LT, GT, LTE, GTE, EQEQ, NEQ};
	
	private CondExprType type;
	private Expr leftExpr, rightExpr;
	
	public CondExpr(CondExprType type, Expr leftExpr, Expr rightExpr, int lineno, int charno) {
		super(lineno, charno);
		this.type = type;
		this.leftExpr = leftExpr;
		this.rightExpr = rightExpr;
	}
	
	/**
	 * A conditional expression has no return type.
	 */
	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return Type.VOID; }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return 0; }

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return 0; }
	
	@Override
	public long localVarSize() { return leftExpr.localVarSize() + rightExpr.localVarSize(); }
	
	@Override
	public long initAddr(long startAddr) {
		startAddr = leftExpr.initAddr(startAddr);
		startAddr = rightExpr.initAddr(startAddr);
		return startAddr;
	}
	
	public void assemble(CodeWriter cw) throws AssemblerException {
		Type compareType = null, leftType = leftExpr.retType(cw), rightType = rightExpr.retType(cw);
		
		if (leftType.isArray() || rightType.isArray()) throw new AssemblerException("Cannot compare arrays.", lineno, charno);
		
		// get the value of the left and right expressions to be on the stack.
		if (leftType.isFloat() || rightType.isFloat()) {
			compareType = Type.FLOAT32;
			if (!leftType.equals(Type.FLOAT32)) leftExpr = new CastExpr(leftExpr, Type.FLOAT32);
			if (!rightType.equals(Type.FLOAT32)) rightExpr = new CastExpr(rightExpr, Type.FLOAT32);
			leftExpr.assemble(cw);
			rightExpr.assemble(cw); // now there are two FLOAT32 values on top of the stack
		} else if (leftType.isInt() && rightType.isInt()) {
			compareType = Type.INT32;
			if (!leftType.equals(Type.INT32)) leftExpr = new CastExpr(leftExpr, Type.INT32);
			if (!rightType.equals(Type.INT32)) rightExpr = new CastExpr(rightExpr, Type.INT32);
			leftExpr.assemble(cw);
			rightExpr.assemble(cw); // now there are two INT32 values on top of the stack
		} else if (leftType.isString() && rightType.isString()) {
			compareType = Type.STRING;
			leftExpr.assemble(cw);
			rightExpr.assemble(cw); // now there are two STRING values on top of the stack
		} else throw new AssemblerException("Cannot compare types: " + leftType + " and " + rightType, lineno, charno);
		
		
		if (type == CondExprType.LT) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.ltFloat_c, lineno, charno).setComment("[stack 2nd pos.] < [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.ltInt_c, lineno, charno).setComment("[stack 2nd pos.] < [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.ltString_c, lineno, charno).setComment("[stack 2nd pos.] < [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else if (type == CondExprType.GT) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.gtFloat_c, lineno, charno).setComment("[stack 2nd pos.] > [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.gtInt_c, lineno, charno).setComment("[stack 2nd pos.] > [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.gtString_c, lineno, charno).setComment("[stack 2nd pos.] > [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else if (type == CondExprType.LTE) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.lteFloat_c, lineno, charno).setComment("[stack 2nd pos.] <= [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.lteInt_c, lineno, charno).setComment("[stack 2nd pos.] <= [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.lteString_c, lineno, charno).setComment("[stack 2nd pos.] <= [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else if (type == CondExprType.GTE) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.gteFloat_c, lineno, charno).setComment("[stack 2nd pos.] >= [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.gteInt_c, lineno, charno).setComment("[stack 2nd pos.] >= [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.gteString_c, lineno, charno).setComment("[stack 2nd pos.] >= [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else if (type == CondExprType.EQEQ) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.eqFloat_c, lineno, charno).setComment("[stack 2nd pos.] == [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.eqInt_c, lineno, charno).setComment("[stack 2nd pos.] == [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.eqString_c, lineno, charno).setComment("[stack 2nd pos.] == [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else if (type == CondExprType.NEQ) {
			if (compareType.equals(Type.FLOAT32))
				cw.add(new Instruction(ISA.neqFloat_c, lineno, charno).setComment("[stack 2nd pos.] != [stack 1st pos.]"));
			else if (compareType.equals(Type.INT32))
				cw.add(new Instruction(ISA.neqInt_c, lineno, charno).setComment("[stack 2nd pos.] != [stack 1st pos.]"));
			else if (compareType.equals(Type.STRING))
				cw.add(new Instruction(ISA.neqString_c, lineno, charno).setComment("[stack 2nd pos.] != [stack 1st pos.]"));
			else throw new AssemblerException("Unkown compare type: " + compareType, lineno, charno);
		} else {
			throw new AssemblerException("Unknown binary compare type: " + type, lineno, charno);
		}
	}
	
	public String toString() {
		return "(CondExpr: " + leftExpr + " " + type + " " + rightExpr + ")";
	}
}
