package servilla.compiler.tree;

/**
 * Encapsulates a constant expression. A constant expression can only be used
 * on the right-side of an expression.  When compiled, it assumes the user wants to
 * obtain a value.
 * 
 * Constant expressions include variable references and primitive values.
 * 
 * @author Chien-Liang Fok
 *
 */
public class ConstExpr extends Expr {
	private Var var = null;
	
	/**
	 * Creates a constant expression that is a variable.  A variable
	 * can either be a VarRef or a Value.
	 * 
	 * @param var The variable.
	 * @param lineno 
	 * @param charno
	 */
	public ConstExpr(Var var, int lineno, int charno) {
		super(lineno, charno);
		this.var = var;
	}
	
	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return var.retType(cw); }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return var.retHeapSize(cw); }
	
	@Override
	public long initAddr(long startAddr) {return startAddr;}

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return retType(cw).stackSize();}
	
	/**
	 * Returns the variable contained within this constant expression.
	 * @return  the variable contained within this constant expression.
	 */
	public Var var() { return var; }
	
	/**
	 * Assembles the constant expression.  Assumes that the user wants to
	 * obtain the value of the constant.
	 * 
	 */
	public void assemble(CodeWriter cw) throws AssemblerException { var.assemble(cw, Var.AssembleType.GET); }
	
	public String toString() { return "(ConstExpr "+var+")"; }
	
	@Override
	public long localVarSize() { return var.localVarSize(); }
}
