package servilla.compiler.tree;

import servilla.service.tree.*;

/**
 * The super class of all expressions.  An expression is the same as a statement except it does
 * not end with a semicolon and may result in a value being pushed onto the stack.  Like a statement,
 * an expression may modify the condition code.
 * 
 * @author Chien-Liang Fok
 */
public abstract class Expr implements Assembler {
	int lineno, charno;
	
	/**
	 * Creates an expression.
	 * 
	 * @param lineno The line number of the program code that this expression maps to.
	 * @param charno The character number of the program code this expression maps to.
	 */
	public Expr(int lineno, int charno) { this.lineno = lineno; this.charno = charno; }
	
	
	/**
	 * Returns the line number of the program code that this expression maps to.
	 * @return the line number of the program code that this expression maps to.
	 */
	public int lineno() { return lineno; }
	
	/**
	 * Returns the character number of the program code this expression maps to.
	 * @return the character number of the program code this expression maps to.
	 */
	public int charno() { return charno; }
	
	/**
	 * Returns the size of the return value in bytes.
	 * 
	 * @return the size of the return value in bytes.
	 */
	public abstract int retSize(CodeWriter cw) throws AssemblerException; // TODO: is this needed?
	
	/**
	 * Returns the type of the return value.
	 * 
	 * @return the type of the return value.
	 */
	public abstract DataType retType(CodeWriter cw) throws AssemblerException;
	
	/**
	 * Searches for local variable declarations and assigns addresses to them.
	 *
	 * @param startAddr The starting address from which addresses may be assigned
	 * @return The next starting address that may be assigned after all of the local 
	 * variables in this expression have been allocated space.
	 */
	public abstract long initAddr(long startAddr);
	
	/**
	 * Returns the number of bytes needed by local variables within the expression.
	 */
	public abstract long localVarSize();
	
	/**
	 * All expressions must include a toString for debugging purposes.
	 * The string must be surrounded by parenthesis to distinguish each expression.
	 */
	public abstract String toString();
}
