package servilla.compiler.tree;

import java.util.*;

/**
 * An for statement executes a list of statements a fixed number of times.
 * 
 * @author Chien-Liang Fok
 */
public class ForStatement extends Statement {
	private Expr initExpr, condExpr, postExpr;
	private Vector<Statement> body;
	
	/**
	 * A for statement consists of an initial expression, conditional expression, and post expression, followed
	 * by a list of statements.  The initial expression is executed before the start of the loop.  Before the loop
	 * executes, the conditional expression is evaluated.  If the conditional expression is true, the body of the
	 * for loop is executed.  At the end of the loop, the post expression is evaluated, and the loop returns to the 
	 * conditional expression.
	 * 
	 * @param initExpr The inital expression.
	 * @param condExpr The conditional expression.
	 * @param postExpr The post expression.
	 * @param body  The list of statements that consistute the body of the for loop.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public ForStatement(Expr initExpr, CondExpr condExpr, Expr postExpr, Vector<Statement> body, int lineno, int charno) {
		super(lineno, charno);
		this.initExpr = initExpr;
		this.condExpr = condExpr;
		this.postExpr = postExpr;
		this.body = body;
	}
	
	/**
	 * Assembles the for statement.
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		String loopLabel = "LOOP" + cw.nonce();
		String doneLabel = "DONE" + cw.nonce();
		
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Begin FOR loop (line="+lineno+")"));
		initExpr.assemble(cw);
		popRetVal(cw, initExpr, "pop result of FOR loop init expr");
		
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(loopLabel)); // mark beginning of loop
		condExpr.assemble(cw);
		popRetVal(cw, condExpr, "pop result of FOR loop cond expr");
		
		// jump to DONE if condition == 0
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(doneLabel)); // push address of done onto stack
		cw.add(new Instruction(ISA.jump_nc, lineno, charno).setComment("jump to FOR loop done if cond=0"));
		
		assembleStatements(cw, body);
		
		postExpr.assemble(cw); // assemble the post expression
		popRetVal(cw, postExpr, "pop result of FOR loop post expr");
		
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(loopLabel));
		cw.add(new Instruction(ISA.jump, lineno, charno).setComment("jump back to the beginning of the for loop"));
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("End of FOR loop (line="+lineno+")").setLabel(doneLabel));
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			sb.append(e.nextElement());
			if (e.hasMoreElements()) sb.append(" ");
		}
		return "(FOR " + initExpr + " " + condExpr + " " + postExpr + " (" + sb + "))";
	}
	
	@Override
	public long initAddr(long startAddr) {
		startAddr = initExpr.initAddr(startAddr);
		startAddr = condExpr.initAddr(startAddr);
		startAddr = postExpr.initAddr(startAddr);
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			Statement s = e.nextElement();
			startAddr = s.initAddr(startAddr);
		}
		return startAddr;
	}
	
	@Override
	public long localVarSize() {
		long result = condExpr.localVarSize() + initExpr.localVarSize() + postExpr.localVarSize();
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		return result;
	}
}
