package servilla.compiler.tree;
import java.util.*;

/**
 * This defines Servilla's instruction set architecture.
 * Each element in the enum corresponds to an instruction.
 * 
 * <p>This class contains a main method that generates NesC code that specifies Servilla's ISA.
 * To generate this, each element in the enum may take up to two parameters specifying (1) the name
 * of the NesC component that implements them, and (2) a comment that describes the instruction and
 * is included in the generated NesC code.  If no NesC component implements the instruction (some
 * instructions are implemented by the ServillaEngine itself), "N/A" is specified in the first parameter.
 * 
 * @author Chien-Liang Fok
 *
 */
public enum ISA {	
	// String operators
	strlen("string", "Pushes the length of a string onto the operand stack.  The abs. addr. of the string must be on top of the stack"),
	strcmp("string", "Compares two strings.  The abs. addr. of both strings must be on top of the stack."),
	setStrVar("string", "Sets one string variable equal to another, the abs. addr. of both are on top of the stack.  The destination address is on top followed by the source address."),
	setStrConst("string", "Sets the value of a string variable to be a constant, the abs. addr. of the string is on the stack"),
	
	// Push an absolute address (in code memory) onto the stack.
	pushRelAddr("addr", "param: relative address, action: push abs. address onto op stack"),
	pushIndAddr("addr", "param: indirect address, action: push abs. address onto op stack"),
	pushAbsAddr("addr", "param: absolute address, action: push abs. address onto op stack"),
	
	// Push a constant (in code memory) onto the operand stack.
	pushIntConst("pushConst", "Get an integer constant that is stored in code memory."),
	pushFltConst("pushConst", "Get a float constant that is stored in code memory."),
	
	// Fetch variable (in heap) given memory address (on operand stack).
	// Push variable on operand stack.
	getIntVarAbs("getVar", "Get an integer variable whose absolute address is on the stack"),
	getFltVarAbs("getVar", "Get a float variable whose absolute address is on the stack"),
	getIntVarInd("getVar", "Get an integer variable whose indirect address is on the stack"),
	getFltVarInd("getVar", "Get a float variable whose indirect address is on the stack"),
	getIntVarRel("getVar", "Get an integer variable whose relative address is on the stack"),
	getFltVarRel("getVar", "Get a float variable whose relative address is on the stack"),
	
	// Save variable (in heap) given address (on operand stack) and value (on operand stack)
	setIntAbs("setVar", "Op Stack: [abs. addr.][integer], Result: var at [abs. addr.] = [integer]"),
	setFltAbs("setVar", "Op Stack: [abs. addr.][float], Result: var at [abs. addr.] = [float]"),
	setIntInd("setVar", "Op Stack: [ind. addr.][integer], Result: var at [ind. addr.] = [integer]"),
	setFltInd("setVar", "Op Stack: [ind. addr.][float], Result: var at [ind. addr.] = [float]"),
	setIntRel("setVar", "Op Stack: [rel. addr.][integer], Result: var at [rel. addr.] = [integer]"),
	setFltRel("setVar", "Op Stack: [rel. addr.][float], Result: var at [rel. addr.] = [float]"),
	
	// Perform function calls
	call("call", "Call a function"),
	ret("call", "Return from a function"),
	
	// Jump instructions
	jump_c("jump", "conditional jump, jump when cond=1"), 		// conditional jump
	jump_nc("jump", "conditional jump, jump when cond=0"),		 // conditional jump
	jump("jump", "Unconditional Jump"), 		// unconditional jump
	
	// Unary math operations
	incInt("uMath"),
	decInt("uMath"),
	incFloat("uMath"),
	decFloat("uMath"),
	negInt("uMath"),			// negate a value
	negFloat("uMath"),
	notInt_c("uMath"),
	notFloat_c("uMath"),
	
	// Binary math operations
	addInt("bMath", "[var2] + [var1] where var2 is second var on stack"),
	subInt("bMath", "[var2] - [var1] where var2 is second var on stack"),
	multInt("bMath", "[var2] * [var1] where var2 is second var on stack"),
	divInt("bMath", "[var2] / [var1] where var2 is second var on stack"),
	lorInt("bMath", "[var2] | [var1] where var2 is second var on stack"),
	landInt("bMath", "[var2] & [var1] where var2 is second var on stack"),
	orInt("bMath", "[var2] || [var1] where var2 is second var on stack"),
	andInt("bMath", "[var2] && [var1] where var2 is second var on stack"),
	
	addFloat("bMath", "[var2] + [var1] where var2 is second var on stack"),
	subFloat("bMath", "[var2] - [var1] where var2 is second var on stack"),
	multFloat("bMath", "[var2] * [var1] where var2 is second var on stack"),
	divFloat("bMath", "[var2] / [var1] where var2 is second var on stack"),
	orFloat("bMath", "[var2] || [var1] where var2 is second var on stack"),
	andFloat("bMath", "[var2] && [var1] where var2 is second var on stack"),
	
	// Service provisioning instructions
	invoke("invoke", "Invoke a service.  param: service name (uint16_t)"),
	//bind("bind", "bind a servRef to a serv"),
	//unbind("bind", "unbind a servRef to a serv"),
	
	numnbrs("nbr", "pushes the number of neighbors onto the stack")	,
	randnbr_c("nbr", "pushes a random neighbor onto the stack, if no neighbors are present, -1 is pushed onto the stack"),
	getnbr_c("nbr", "retrieves a particular neighbor, the index of the neighbor is on top of the stack, if the neighbor index is invalid, -1 is pushed onto the stack"),
	isnbr_c	("nbr", "determines whether a neighbor has a particular address, the address of the neighbor is on top of the operand stack"),
	
	smove("migrate"),
	wmove("migrate"),
	sclone("migrate"),
	wclone("migrate"),
	
	ltInt_c("compare", "compare two integers, less than"),			// comparison operators
	gtInt_c("compare", "compare two integers, greater than"),
	lteInt_c("compare", "compare two integers, less than or equal to"),
	gteInt_c("compare", "compare two integers, greater than or equal to"),
	eqInt_c("compare", "compare two integers, equal to"),
	neqInt_c("compare", "compare two integers, not equal"),
	
	ltFlt_c("compare", "compare two floats, less than"),
	gtFlt_c("compare", "compare two floats, greater than"),
	lteFlt_c("compare", "compare two floats, less than or equal to"),
	gteFlt_c("compare", "compare two floats, greater than or equal to"),
	eqFlt_c("compare", "compare two floats, equal to"),
	neqFlt_c("compare", "compare two floats, not equal"),
	
	cond("cond", "push the condition code onto the stack"),
	pull_c("cond", "takes the integer value on the stack and save it in the condition code register"),
	neg_c("cond", "negate the condition code."),
	
	popInt("pop", "pops an integer value off the stack"),
	popFlt("pop", "pops a float value off the stack"),
	
	/*rout("rts"),
	rinp("rts"),
	rrdp("rts"),
	out("ts"),
	in("ts"),
	rd("ts"),
	inp("ts"),
	rdp("ts"),*/
	
	toggleled("toggleled", "Toggles the LED."),
	putled("putled", "controls the LEDs"),
	
	castItoF("cast", "Cast an integer into a float"),
	castFtoI("cast", "Cast a float into an integer"),
	
	// Misc. instructions
	noop("noop", "no op"),
	sleep("sleep", "Pause a script for a specified period of time."),
	halt("halt", "Kills the script"),
	addr("localAddr", "Pushes the address of the local node onto the stack."),
	id("id", "Push's the script's id onto the stack"),
	rand("rand", "push a random 32-bit value onto the stack")
	
	// Unused instructions
	//send("send", "Send some data to the base station"),
	;
	
	static byte counter = 0;
	
	static byte nextCount() {
		return (byte)counter++;
	}
	
	byte value;
	String component; // the name of the NesC component that implements the instruction
	String desc;
	
	private ISA(){ //byte value) {
		//this.value = value;
		this.value = ISA.nextCount();
	}
	
	private ISA(String comp) {
		this();
		this.component = comp;
	}
	
	private ISA(String comp, String desc) {
		this (comp);
		this.desc = desc;
	}
	
	/**
	 *  Returns the name of the NesC component that implements the instruction.
	 *  By default the name of the component is the name of the instruction.
	 *  
	 * @return The name of the component that implements the instruction.
	 */
	public String comp() {
		if (component == null)
			return this.toString() + "C";
		else
			return component + "C";
	}
	
	/**
	 * Returns the ISA with the specified instruction, or null if the instruction does not exist.
	 * 
	 * @param instr The instruction.
	 * @return the ISA containing the specified instruction.
	 */
	public static ISA getInstruction(String instr) {
		ISA[] v = ISA.values();
		for (int i = 0; i < v.length; i++) {
			if (v[i].toString().equals(instr)) {
				return v[i];
			}
		}
		return null;
	}
	
	/**
	 * Returns the description of the instruction.
	 * 
	 * @return the description of the instruction.
	 */
	public String desc() { return desc; }
	
	/**
	 * Returns the bytecode of the instruction.
	 * 
	 * @return the instruction's bytecode.
	 */
	public byte byteCode() { return value; }
	
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		/*try {
			Class c = Class.forName("servilla.compiler.tree.ISA");
			java.lang.reflect.Field[] fields = c.getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
					System.out.println(fields[i].getName());
			}
		} catch(Exception e) {
			e.printStackTrace();
		}*/
		ISA[] v = ISA.values();
		System.out.println("// This was automatically generated by servilla.compiler.tree.ISA");
		System.out.println("// Copy this into nesc/servilla/vm/opcodes/Opcodes.h");
		System.out.println("// Total # instructions: " + v.length);
		System.out.println("#ifndef SERVILLA_OPCODES_H_INCLUDED\n#define SERVILLA_OPCODES_H_INCLUDED");
		System.out.println("typedef enum {");
		for (int i = 0; i < v.length; i++) {
			System.out.println("\tIOP" + v[i] + " = "+ servilla.util.ArrayToHexString.byteToHexString(v[i].byteCode()) + ",");
		}
		System.out.println("} OpCode;\n#endif\n");	
		
		// Generate the OpToTextC.nc component
		System.out.println("\n\n\n// This was automatically generated by servilla.compiler.tree.ISA.");
		System.out.println("// Copy this into nesc/servilla/vm/opcodes/OpToTextP.nc");
		System.out.println("module OpToTextP {");
		System.out.println("	provides interface OpToText;");
		System.out.println("}");
		System.out.println("implementation {");
		System.out.println("	command char* OpToText.toString(uint8_t op) {");
		System.out.println("		switch(op) {");
		for (int i = 0; i < v.length; i++) {
			System.out.println("		case IOP" + v[i] +": return \"" + v[i] + "\";");
		}
		System.out.println("		default: return \"UNKNOWN\";");
		System.out.println("		}	");
		System.out.println("	}");
		System.out.println("}");
		
		StringBuffer sncBuff = new StringBuffer();
		sncBuff.append("\n\n\n// This was automatically generated by servilla.compiler.tree.ISA.\n");
		sncBuff.append("// Copy this into nesc/servilla/vm/VM.nc\n");
		sncBuff.append("includes Servilla;\n" + 
				"includes Opcodes;\n\n" + 
				"configuration VM {}\n" +
				"implementation {\n" +
				"\tcomponents Main, EngineC;\n" +
				"\tcomponents ");
		Vector<String> opComps = new Vector<String>();
		for (int i = 0; i < v.length; i++) {
			//System.out.println("Checking: " + v[i].comp());
			if (!opComps.contains(v[i].comp())) {
				opComps.add(v[i].comp());
			}// else System.out.println("REJECT");
		}
		for (int i = 0; i < opComps.size(); i++) {
			sncBuff.append("OP" + opComps.get(i));
			if (i < opComps.size()-1)
				sncBuff.append(", ");
		}
		sncBuff.append(";\n\n");
		
		for (int i = 0; i < v.length; i++) {
			sncBuff.append("\tEngineC.BasicISA[IOP" + v[i] +"] -> OP"+v[i].comp() + ";" + ((v[i].desc() != null) ? ("\t//" + v[i].desc()) : "") + "\n");
		}
		sncBuff.append("}\n");
		System.out.println(sncBuff.toString());
	}
}
