package servilla.compiler.tree;

import java.util.*;

/**
 * An if-else statement consists of two blocks of statements.  Only one block
 * is executed depending on a conditional expression.
 * 
 * @author Chien-Liang Fok
 */
public class IfElseStatement extends Statement {
	private Expr ifExpr; // the variable being written to
	private Vector<Statement> trueBody, falseBody;
	
	/**
	 * Creates an if-else statement.
	 * 
	 * @param ifExpr The condition that is evaluated to determine which block of statements to execute.
	 * @param trueBody The block of statements that is executed if ifExpr evaluates to TRUE.
	 * @param falseBody The block of statements that is executed if ifExpr evaluates to FALSE.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public IfElseStatement(Expr ifExpr, Vector<Statement> trueBody, Vector<Statement> falseBody, int lineno, int charno)
		throws AssemblerException {
		super(lineno, charno);
		//if (! (ifExpr instanceof CondExpr)) throw new AssemblerException ("If-Else statement expression not a CondExpr", lineno, charno);
		this.ifExpr = ifExpr;
		this.trueBody = trueBody;
		this.falseBody = falseBody;
	}
	
	public void assemble(CodeWriter cw) throws AssemblerException {
		String trueLabel = "TRUE" + cw.nonce();
		String continueLabel = "CONT" + cw.nonce();
		
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Begin IF-ELSE statement (line="+lineno+")"));
		ifExpr.assemble(cw);  // assemble the conditional value
		popRetVal(cw, ifExpr, "pop result of IF-ELSE cond expr");
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(trueLabel));
		cw.add(new Instruction(ISA.jump_c, lineno, charno).setComment("Jump to true block if condition is true")); // jump to true label
		assembleStatements(cw, falseBody);
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(continueLabel));
		cw.add(new Instruction(ISA.jump, lineno, charno).setComment("Jump past the true block")); // jump to continue label
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(trueLabel));
		assembleStatements(cw, trueBody);
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(continueLabel).setComment("End IF-ELSE statement (line="+lineno+")"));
	}
	
	public String toString() {
		StringBuffer truesb = new StringBuffer();
		Enumeration<Statement> e = trueBody.elements();
		while (e.hasMoreElements()) {
			truesb.append(e.nextElement());
			if (e.hasMoreElements()) 
				truesb.append(" ");
		}
		
		StringBuffer falsesb = new StringBuffer();
		e = falseBody.elements();
		while (e.hasMoreElements()) {
			falsesb.append(e.nextElement());
			if (e.hasMoreElements()) 
				falsesb.append(" ");
		}
		return "(IF " + ifExpr + " (" + truesb + ") (" + falsesb + "))";
	}
	
	@Override
	public long initAddr(long startAddr) {
		startAddr = ifExpr.initAddr(startAddr);
		Enumeration<Statement> e = trueBody.elements();
		while (e.hasMoreElements()) {
			Statement s = e.nextElement();
			startAddr = s.initAddr(startAddr);
		}
		e = falseBody.elements();
		while (e.hasMoreElements()) {
			Statement s = e.nextElement();
			startAddr = s.initAddr(startAddr);
		}
		return startAddr;
	}
	
	@Override
	public long localVarSize() {
		long result = ifExpr.localVarSize();
		Enumeration<Statement> e = trueBody.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		e = falseBody.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		return result;
	}
}
