package servilla.compiler.tree;

import java.util.*;

/**
 * An if statement consists of a condition and a block of statements that is only executed
 * if the condition is true. 
 * 
 * @author Chien-Liang Fok
 */
public class IfStatement extends Statement {
	private Expr ifExpr; // the variable being written to
	private Vector<Statement> body;
	
	/**
	 * Creates an if statement.
	 * 
	 * @param ifExpr The condition that is evaluated to determine which block of statements to execute.
	 * @param trueBody The block of statements that is executed if ifExpr evaluates to TRUE.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public IfStatement(Expr ifExpr, Vector<Statement> body, int lineno, int charno) {
		super(lineno, charno);
		/*
		 * The IF statement's ifExpr may be a ParenExpr in addition to a CondExpr.
		 */
		// TODO Add a new method to Expr:  setsCond() that tells whether the expression sets the condition code.
		//if (! (ifExpr instanceof CondExpr)) throw new AssemblerException ("If statement expression not a CondExpr", lineno, charno);
		this.ifExpr = ifExpr;
		this.body = body;
	}
	
	@Override
	public void assemble(CodeWriter cw) throws AssemblerException {
		String trueLabel = "TRUE" + cw.nonce();
		String continueLabel = "CONT" + cw.nonce();
		
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Begin IF statement (line="+lineno+")"));
		ifExpr.assemble(cw);  // assemble the conditional value
		popRetVal(cw, ifExpr, "pop result of IF cond expr");
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(trueLabel));
		cw.add(new Instruction(ISA.jump_c, lineno, charno).setComment("Jump to true block if condition is true")); // jump to true label
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(continueLabel));
		cw.add(new Instruction(ISA.jump, lineno, charno).setComment("Jump past the true block")); // jump to continue label
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(trueLabel));
		assembleStatements(cw, body);
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(continueLabel).setComment("End IF statement (line="+lineno+")"));
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			sb.append(e.nextElement());
			if (e.hasMoreElements()) 
				sb.append(" ");
		}
		return "(IF (" + ifExpr + ") (" + sb + "))";
	}

	@Override
	public long initAddr(long startAddr) {
		startAddr = ifExpr.initAddr(startAddr);
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			Statement s = e.nextElement();
			startAddr = s.initAddr(startAddr);
		}
		return startAddr;
	}

	@Override
	public long localVarSize() {
		long result = ifExpr.localVarSize();
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		return result;
	}
}
