package servilla.compiler.tree;

import java.io.Serializable;
import java.util.*;
import servilla.util.*;

/**
 * Encapsulates an instruction and program meta data.  This is
 * used by the ProgramTokenizer, which transforms 
 * Servilla code into a list of Instruction objects.
 * 
 * @author Chien-Liang Fok
 */
public class Instruction implements Serializable{
	private static final long serialVersionUID = 1457872207630899740L;
	
	private ISA opcode;
	private String label = null, labelParam = null, comment = null;
	private int lineno, charno;
	private long address;
	private Vector<Byte> params = null; // the parameters
	
	/**
	 * Creates an instruction.
	 * 
	 * @param opcode The instruction opcode.
	 * @param lineno The line number of the code that generated the instruction
	 * @param charno The character number within the code that generated the instruction.
	 */
	public Instruction(ISA opcode, int lineno, int charno) {
		this.opcode = opcode;
		this.lineno = lineno;
		this.charno = charno;
	}
	
	/**
	 * Sets a comment about this instruction.  It returns itself so it can be 
	 * strung together with the creation of the instruction.
	 * 
	 * For example:  Instruction i = new Instruction(...).setComment("...");
	 * 
	 * @param comment The comment to add.  Note that it should *NOT* start with the comment symbol "--"
	 * @return Itself.
	 */
	public Instruction setComment(String comment) { this.comment = comment; return this; }
	
	/**
	 * Sets a 2-byte parameter.  This may be a uint16_t or int16_t.
	 * @param param The 16-bit parameter.
	 * @return Itself.
	 */
	//public Instruction setParam2(int param) { params = new Vector<Byte>(); ByteUtils.saveBytes2(params, param); return this;}
	
	/**
	 * Sets a 4-byte parameter.  This may be a uint32_t or int32_t.
	 * @param param The 32-bit parameter.
	 * @return Itself.
	 */
	public Instruction setParam(long param) { params = new Vector<Byte>(); ByteUtils.saveBytes4(params, param); return this;}
	
	/**
	 * Sets a 4-byte floating point number parameter.
	 * @param param The 32-bit floating point parameter.
	 * @return Itself.
	 */
	public Instruction setParam(float param) { params = new Vector<Byte>(); ByteUtils.saveBytes4(params, param); return this;}
	
	/**
	 * Sets a string parameter.
	 * @param param The string parameter.
	 * @return Itself.
	 */
	public Instruction setParam(String param) {params = new Vector<Byte>(); ByteUtils.saveBytes(params, param); return this;}
	
	/**
	 * Creates an Instruction object based on a line of assembly code.
	 * 
	 * @param asm The line of assembly code.
	 */
	public Instruction(String asm) throws InstructionException {
		String code = asm;
		
		// remove the comment if there is one
		int startCommentIndex = asm.indexOf("--");
		if (startCommentIndex != -1) {
			this.comment = asm.substring(startCommentIndex+2);
			code = asm.substring(0, startCommentIndex);
		}
		
		String[] tokens = code.split("\\s+");
		if (tokens.length == 1) {
			opcode = ISA.getInstruction(tokens[0]);
			if (opcode == null) 
				throw new InstructionException("Invalid instruction: " + tokens[0]);
			
		} else if (tokens.length == 2) {
			// If the first block is an instruction, the second block must be a parameter
			opcode = ISA.getInstruction(tokens[0]);
			if (opcode != null) {
				saveParam(opcode, tokens[1]);
			}
			// Otherwise, the first block must be a label and the second must be an instruction
			else {
				label = tokens[0];
				opcode = ISA.getInstruction(tokens[1]);
				if (opcode == null) 
					throw new InstructionException("Invalid instruction: " + tokens[1]);
			}
		} else if (tokens.length == 3) {
			label = tokens[0];
			opcode = ISA.getInstruction(tokens[1]);
			if (opcode == null) 
				throw new InstructionException("Invalid instruction: " + tokens[1]);
			saveParam(opcode, tokens[2]);
		}
		
	}
	
	/**
	 * Saves the parameter for a specific instruction.
	 * 
	 * @param instr The instruction with the parameter.
	 * @param param The parameter to save.
	 * @throws InstructionException When the instruction is not supposed to have parameters.
	 */
	private void saveParam(ISA instr, String param) throws InstructionException {
		if (instr.equals(ISA.pushIntConst)) {
			try {
				setParam(Long.valueOf(param).longValue());
			} catch(NumberFormatException e) {
				setLabelParam(param); // if the parameter is not an integer, it must be a label
			}
		} else if (instr.equals(ISA.pushFltConst)) {
			try {
				setParam(Float.valueOf(param).floatValue());
			} catch(NumberFormatException e) {
				setLabelParam(param); // if the parameter is not an integer, it must be a label
			}
		} else if (instr.equals(ISA.setStrConst)) {
			setParam(param);
		} else throw new InstructionException("ERROR: Cannot save parameter for instruction "+ instr);
	}
	
	/**
	 * Some instructions like jump, jump_c, and call use a label instead of
	 * a destination address.  This method stores this label temporarily.  It
	 * will be resolved to an address during the final stages of compilation.
	 * 
	 * @param label The destination's label.
	 */
	public Instruction setLabelParam(String label) { this.labelParam = label; return this; }
	
	/**
	 * Returns true if the instruction has a label as a parameter.  This label must
	 * eventually be resolved to an address.
	 * 
	 * @return true if the instruction has a label as a parameter
	 */
	public boolean hasLabelRef() { return labelParam != null; }
	
	/**
	 * Returns the label reference. 
	 * @return the label reference
	 */
	public String labelRef() { return labelParam; }
	
	/**
	 * Determines the equality of this object and another object.
	 */
	public boolean equals(Object o) {
		if (o instanceof Instruction) {
			Instruction i = (Instruction)o;
			return i.opcode().equals(opcode());
		} else
			return false;
	}

	/**
	 * An accessor to the opcode.
	 * 
	 * @return The opcode.
	 */
	public ISA opcode() { return opcode; }
	
	/**
	 * An accessor to the line number.
	 * 
	 * @return The line number.
	 */
	public int lineno() { return lineno; }
	
	/**
	 * An accessor to the character number.
	 * 
	 * @return The character number.
	 */
	public int charno() { return charno; }
	
	/**
	 * Returns true if this instruction has a parameter.
	 * @return true if this instruction has a parameter.
	 */
	public boolean hasParam() { return params != null; }
	
	/**
	 * Returns the number of bytes this instruction consumes
	 * within the code memory.  This is the sum of the instruction
	 * plus its parameters.
	 * 
	 * @return The number of bytes consumed by this instruction.
	 */
	public int numBytes() {
		if(hasParam()) {
			return 1+params.size();
		} else if (labelParam != null)
			return 1+4;  // four bytes to store address that the label resolves to.
		else
			return 1;
	}
	
	/**
	 *  An accessor to the line's label.
	 *  
	 * @return The line's label.
	 */
	public String label() {
		return label;
	}
	
	/**
	 * Returns whether this instruction has a label.
	 * 
	 * @return whether this instruction has a label.
	 */
	public boolean hasLabel() { return label != null; }
	
	/**
	 * Sets this instruction's label.
	 * 
	 * @param label this instruction's label.
	 * @return Itself.
	 */
	public Instruction setLabel(String label) { this.label = label; return this; }
	
	/**
	 * Sets the address of the instruction within the code memory.
	 * 
	 * @param address The address.
	 */
	public Instruction setAddress(long address) { this.address = address; return this; }
	
	/**
	 * Returns the address of the instruction within the code memory.
	 *  
	 * @return the address of the instruction within the code memory.
	 */
	public long address() { return address; }
	
	/**
	 * Translates the instruction into byte code.  All label parameters must have
	 * been resolved into an address by the time this method is called.
	 * 
	 * @return The byte code representation of the instruction.
	 */
	public Vector<Byte> byteCode() {
		Vector<Byte> v = new Vector<Byte>();
		v.add(opcode.byteCode());
		if (hasParam()) {
			v.addAll(params);
		}
		return v;
	}
	
	/**
	 * Translates the instruction into NesC code.
	 * @return a vector of strings in which each string represents
	 * a byte of code.
	 */
	public Vector<String> nesCCode() {
		Vector<String> v = new Vector<String>();
		v.add("IOP" + opcode);
		if (hasParam())	v.addAll(VectorUtils.toStringVector(params));
		return v;
	}
	
	/**
	 * Returns a String representation of this class.
	 * 
	 * @param return a String representation of this class.
	 */
	public String toString() {
		String param = "";
		if(hasParam()) param = ArrayToHexString.toString(params);
		else if (labelParam != null) param = labelParam;
		return lineno + " (" + address + "): " + (label == null ? "\t" : label) + "\t" + opcode + "\t" + " " + param + (comment != null ? "\t--"+comment : "");
	}
}
