package servilla.compiler.tree;
import servilla.service.tree.DataType;
import servilla.util.*;
import java.util.*;

/**
 * Holds a primitive integer value.  This incldues INT16, UINT16, INT32, and UINT32.
 * 
 * @author Chien-Liang Fok
 */
public class IntValue extends Var {
	private Type type;
	private Vector<Long> intVal;		// multi-dimensional arrays are stored linearly
	
	
	/**
	 * Creates an integer value.  The type of value is the smallest type that
	 * can hold its value.
	 * 
	 * @param intVal  The integer value
	 * @param lineno The line number of the program code that this object maps to.
	 * @param charno The character number of the program code that this object maps to.
	 */
	public IntValue(long intVal, int lineno, int charno) {
		super(lineno, charno);
		this.intVal = new Vector<Long>();
		this.intVal.add(intVal);
		if (intVal < 0) {
			if (-1*intVal > 0xffff)
				this.type = Type.INT32;
			else
				this.type = Type.INT16;
		} else {
			if (intVal > 0xffff) 
				this.type = Type.UINT32;
			else
				this.type = Type.UINT16;
		}
	}
	
	/**
	 * Creates an array integer value.
	 * 
	 * @param arrayDim The array dimensions.
	 * @param intVal The values in the array.
	 * @param lineno The line number of the program code that this object maps to.
	 * @param charno The character number of the program code that this object maps to.
	 */
	public IntValue(Vector<Integer> arrayDim, Vector<Long> intVal, int lineno, int charno) {
		super(lineno, charno);
		this.intVal = intVal;
		this.type = new Type(detArrayType(intVal), arrayDim);
	}
	
	/**
	 * Determine the data type of the array.  The data type is the smallest type that can
	 * hold all of the values.
	 * 
	 * @param intVal The array values.
	 * @return The data type.
	 */
	private DataType detArrayType(Vector<Long>intVal) {
		boolean isSigned = false;
		for (int i = 0; !isSigned && i < intVal.size(); i++) { if (intVal.get(i) < 0) isSigned = true; }
		if (isSigned) {
			for (int i = 0; i < intVal.size(); i++) {
				long currVal = intVal.get(i);
				if (currVal > 0xffff) return DataType.UINT32;
			}
			return DataType.UINT16;
		} else { // It's either Type.INT32 or Type.INT16
			for (int i = 0; i < intVal.size(); i++) {
				long currVal = intVal.get(i);
				if (currVal < 0 && -1*currVal > 0xffff)
					return DataType.INT32;
				else if (currVal > 0 && currVal > 0xffff)
					return DataType.INT32;
			}
			return DataType.INT16;
		}
	}
	
	/**
	 * Manually sets the data type within the value.  This is used by the
	 * SendDataRcvr when it receives data from the network.
	 * @param type
	 */
	//public IntValue setType(Type type) { this.type = type; return this; }
	
	@Override
	public void assemble(CodeWriter cw, AssembleType aType) throws AssemblerException {
		if (aType == AssembleType.SET) throw new AssemblerException("Attempting to write to a value", lineno, charno);
		if (type == DataType.UINT16 || type == DataType.INT16) {
			cw.add(new Instruction(ISA.push2, lineno, charno).setParam2((int)intVal));
		} else if (type == DataType.UINT32 || type == DataType.INT32) {
			cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(intVal));
		} else if (type == DataType.FLOAT32) {
			cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(floatVal));
		} else if (type == DataType.STRING) {
			cw.add(new Instruction(ISA.pushString, lineno, charno).setParam(stringVal));
		} else {
			throw new AssemblerException("Unknown data type: " + type, lineno, charno);
		}
	}

	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return type; }
	
	@Override
	/**
	 * This returns the actual size of the value.  For a string,
	 * this is the <b>actual number of bytes</b> it consumes on the stack.
	 */
	public int retValSize(CodeWriter cw) throws AssemblerException {
		return retType(cw).size();
	}
	
	public Vector<Long> getValue() { return intVal; }
	
	public String toString() {
		StringBuffer result = new StringBuffer();
		result.append("(Value ");
		if (type == DataType.INT16 || type == DataType.UINT16) {
			result.append(type.toString());
			result.append(" ");
			result.append((int)intVal);
			result.append(" (0x" + Integer.toHexString((int)intVal) + ")");
		}
		else if (type == DataType.INT32 || type == DataType.UINT32) {
			result.append(type.toString());
			result.append(" ");
			result.append(intVal);
			result.append(" (0x" + Long.toHexString(intVal) + ")");
		}
		else if (type == DataType.FLOAT32) {
			result.append(DataType.FLOAT32.toString());
			result.append(" ");
			result.append(floatVal);
			Vector<Byte> bv = new Vector<Byte>();
			ByteUtils.saveBytes4(bv, floatVal);
			
			result.append(" (" + ByteUtils.toString(bv)+ ")");
		}
		else if (type == DataType.STRING) {
			result.append(DataType.STRING.toString());
			result.append(" ");
			result.append(stringVal);
		}
		else {
			result.append("ERROR: Unknown type: ");
			result.append(type.toString());
		}
		result.append(")");
		return result.toString();
	}

	@Override
	public long localVarSize() { return 0; }
}
