package servilla.compiler.tree;

import java.util.*;
import servilla.service.tree.Tuple;
import servilla.service.tree.Field;

/**
 * An invoke statement invokes a service.
 * 
 * @author Chien-Liang Fok
 *
 */
public class InvokeExpr extends Expr {
	private String specName;
	private Vector<Expr> inputParams;
	private Vector<Expr> outputParams;
	
	public InvokeExpr(String specName, Vector<Expr> inputParams, Vector<Expr> outputParams, int lineno, int charno) {
		super(lineno, charno);
		this.specName = specName;
		this.inputParams = inputParams;
		this.outputParams = outputParams;
	}

	@Override
	public Type retType(CodeWriter cw) throws AssemblerException { return Type.VOID; }

	@Override
	public int retHeapSize(CodeWriter cw) throws AssemblerException { return retType(cw).heapSize(); }
	
	@Override
	public long initAddr(long startAddr) {
		Enumeration<Expr> e = inputParams.elements();
		while (e.hasMoreElements()){
			Expr expr = e.nextElement();
			startAddr = expr.initAddr(startAddr);
		}
		e = outputParams.elements();
		while (e.hasMoreElements()){
			Expr expr = e.nextElement();
			startAddr = expr.initAddr(startAddr);
		}
		return startAddr;
	}

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return retType(cw).stackSize(); }
	
	@Override
	public long localVarSize() {
		long result = 0;
		Enumeration<Expr> e = inputParams.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		e = outputParams.elements();
		while (e.hasMoreElements()) {
			result += e.nextElement().localVarSize();
		}
		return result;
	}
	/**
	 * Generates the binary code for a service invocation.
	 * 
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		int nonce = cw.nonce();
		String retLabel = "INVOKE_RET" 		+ nonce; // generate the return location
		String startLabel = "INVOKE_STRT" 	+ nonce;
		String failLabel = "INVOKE_FAIL" 	+ nonce;
		String nosaveLabel = "INVOKE_NS" 	+ nonce;
		String doneLabel = "INVOKE_DONE" 	+ nonce;
		
		long inputParamSize = 0, outputParamSize = 0;
		Enumeration<Expr> ie = inputParams.elements();
		while (ie.hasMoreElements()) {
			inputParamSize += ie.nextElement().retStackSize(cw); 
		}
		Enumeration<Expr> oe = outputParams.elements();
		while (oe.hasMoreElements()) {
			outputParamSize += oe.nextElement().retStackSize(cw);
		}
		
		ServiceDecl spec = cw.getServiceDecl(specName);
		if (spec == null) throw new AssemblerException("Cannot find spec " + specName, lineno, charno);
		
		cw.add(new Instruction(ISA.set_nxt_fp, lineno, charno).setLabelParam(retLabel).setComment("mark new FP, retLabel="+retLabel+" begin invoke expression"));
		
		// push the input parameters onto the stack
		ServiceDecl sd = cw.getServiceDecl(specName);
		Tuple inputFields;
		try {
			inputFields = sd.getSpec().getInputProperty().value();
		} catch (Exception e1) {
			throw new AssemblerException("Error compiling InvokeExpr: Unable to get the input fields.", lineno, charno);
		}
		ie = inputParams.elements();
		int paramIndx = 0;
		while (ie.hasMoreElements()) {
			Expr e = ie.nextElement();
			Type retType = e.retType(cw); Field specType = inputFields.getField(paramIndx++);
			if (!specType.matches(retType)) throw new AssemblerException("Parameter type mismatch, got " + retType + " expected " + specType + ".", lineno, charno);
			e.assemble(cw); 
		}
		
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(startLabel));
		cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(0).setComment("No local variables within invoke method"));
		cw.add(new Instruction(ISA.call, lineno, charno).setComment("Set the new FP"));
		cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(inputParamSize).setLabel(startLabel));
		cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(spec.handle()));
		cw.add(new Instruction(ISA.invoke, lineno, charno).setComment("Invoke the service"));
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(failLabel).setComment("Push address of " + failLabel));
		cw.add(new Instruction(ISA.jump_nc, lineno, charno).setComment("Jump to " + failLabel + " if the invoke failed"));
		cw.add(new Instruction(ISA.ret, lineno, charno).setParam4(outputParamSize).setComment("Invoke succeeded, must save " + outputParamSize + " bytes"));
		cw.add(new Instruction(ISA.ret, lineno, charno).setParam4(0).setLabel(failLabel).setComment("Invoke failed, no bytes to save."));
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(nosaveLabel).setLabel(retLabel).setComment("Push address of " + nosaveLabel));
		cw.add(new Instruction(ISA.jump_nc, lineno, charno).setComment("Skip the saving of the results since invocation failed."));
		
		// save results from stack into local var
		Tuple outputFields;
		try {
			outputFields = sd.getSpec().getOutputProperty().value();
		} catch(Exception e) {
			throw new AssemblerException("Error compiling InvokeExpr: Unable to get the output fields.", lineno, charno);
		}
		paramIndx = 0;
		oe = outputParams.elements();
		while (oe.hasMoreElements()) {
			Expr currExpr = oe.nextElement();
			if (currExpr instanceof ConstExpr) {
				Var var = ((ConstExpr)currExpr).var();
				if (var instanceof VarRef) {
					VarRef ref = (VarRef)var;
					VarDeclExpr decl = cw.getLocalVar(ref.name(), lineno, charno);
					
					Type varType = decl.retType(cw);  // check that the spec type matches the variable type
					Field specType = outputFields.getField(paramIndx++);
					if (!specType.matches(varType)) throw new AssemblerException("Output parameter type mismatch, got " + varType + " expected " + specType +".", lineno, charno);
					
					AssignmentExpr.saveResults(cw, ref, lineno, charno); // save the results
				} else throw new AssemblerException("Attempting to save results in a value.", lineno, charno);
			} else throw new AssemblerException("Attempting to save results in non-constant expression.", lineno, charno);
		}
		cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(1).setComment("Push 1 onto stack indicating success"));
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(doneLabel).setComment("Push address of " + doneLabel));
		cw.add(new Instruction(ISA.jump, lineno, charno).setComment("Jump to " + doneLabel));
		cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(0).setLabel(nosaveLabel).setComment("Push 0 onto stack indicating failure"));
		cw.add(new Instruction(ISA.noop, lineno, charno).setLabel(doneLabel).setComment("End of invocation."));
	}
	
	public String toString() {
		StringBuffer inputsb = new StringBuffer();
		inputsb.append("(INPUT ");
		Enumeration<Expr> e = inputParams.elements();
		while (e.hasMoreElements()) {
			inputsb.append(e.nextElement());
			if (e.hasMoreElements()) inputsb.append(" ");
		}
		inputsb.append(")");
		
		StringBuffer outputsb = new StringBuffer();
		outputsb.append("(OUTPUT ");
		e = outputParams.elements();
		while (e.hasMoreElements()) {
			outputsb.append(e.nextElement());
			if (e.hasMoreElements()) outputsb.append(" ");
		}
		outputsb.append(")");
		
		return "(InvokeExpression " + specName + " " + inputsb + " " + outputsb + ")";
	}
}
