package servilla.compiler.tree;

import java.util.*;
import servilla.service.tree.*;

/**
 * Encapsulates a method declaration.
 * 
 * @author Chien-Liang Fok
 *
 */
public class MethodDecl implements Assembler {
	private DataType retType;
	private String name;  			// The name of this method
	private Vector<VarDeclExpr> params;
	private Vector<Statement> stmnts;
	private int lineno, charno;
	
	public MethodDecl(DataType retType, String name, Vector<VarDeclExpr> params, 
			Vector<Statement> stmnts, int lineno, int charno) {
		this.retType = retType;
		this.name = name;
		this.params = params;
		this.stmnts = stmnts;
		this.lineno = lineno;
		this.charno = charno;
	}
	
	/**
	 * Returns the return type of the method.
	 * @return the return type of the method.
	 */
	public DataType retType() {return retType;}
	
	/**
	 * Returns the number of bytes needed to hold the local variables
	 * used by the method.
	 * 
	 * @return the number of bytes needed to hold the local variables
	 * used by the method.
	 */
	public long frameSize() {
		long result = 0;
		Enumeration<VarDeclExpr> pv = params.elements();
		while (pv.hasMoreElements()) {
			result += pv.nextElement().localVarSize();
		}
		Enumeration<Statement> se = stmnts.elements();
		while (se.hasMoreElements()) {
			result += se.nextElement().localVarSize();
		}
		return result;
	}
	
	/**
	 * Initializes the relative addresses of all local variables used by this
	 * method.
	 */
	public void initAddr() {
		long currAddr = 0; 
		Enumeration<VarDeclExpr> e1 = params.elements();
		while (e1.hasMoreElements()) {
			VarDeclExpr vde = e1.nextElement();
			currAddr = vde.initAddr(currAddr);
		}
		Enumeration<Statement> e2 = stmnts.elements();
		while (e2.hasMoreElements()) {
			Statement stmt = e2.nextElement();
			currAddr = stmt.initAddr(currAddr);
		}
	}
	
	/**
	 * Assemble this method.
	 * 
	 */
	public void assemble(CodeWriter cw) throws AssemblerException {
		cw.newFrame();
		cw.pushMethod(this);
		
		// Add each parameter as a local variable
		Enumeration<VarDeclExpr> e1 = params.elements();
		while (e1.hasMoreElements()) {
			cw.addLocalVar(e1.nextElement(), lineno, charno);
		}
		
		// mark the beginning of the method
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Beginning of method " + name).setLabel(name));

		// assemble each statement
		Enumeration<Statement> e2 = stmnts.elements();
		while (e2.hasMoreElements()) {
			Statement stmt = e2.nextElement();
			stmt.assemble(cw);
		}
		
		if (name.equals("main")) {
			cw.setInitSP(frameSize());
			cw.add(new Instruction(ISA.halt, lineno, charno).setComment("End of script"));
		} else {
			cw.add(new Instruction(ISA.ret, lineno, charno).setParam(retType.size()).setComment("Return to calling function."));
		}
		cw.deleteFrame();
		cw.popMethod();
	}
	
	public int lineno() { return lineno; }
	
	public int charno() { return charno; }
	
	/**
	 * Returns the number of parameters this method requires.
	 * @return the number of parameters this method takes in.
	 */
	public int numParams() { return params.size(); }
	
	/**
	 * Returns the type of the specified parameter.  The parameter is specified
	 * by its order in the parameter list, starting from index 0.
	 * 
	 * @param indx The position of the return type.
	 * @return The data type of the return type
	 */
//	public DataType getParamType(int indx) {
//		return params.get(indx).type();
//	}
	
	public boolean equals(Object o) {
		if (o instanceof MethodDecl) {
			MethodDecl md = (MethodDecl)o;
			if (md.name().equals(name) && paramsMatch(md.getParamTypes())) {
				return true;
			} else return false;
		} else return false;
	}
	
	/**
	 * Determines whether this method has parameters of the
	 * specified types.
	 * 
	 * @param p The parameter types.
	 * @return whether the parameters match.
	 */
	public boolean paramsMatch(Vector<DataType> p) {
		if (p.size() == numParams()) {
			for (int i = 0; i < p.size(); i++) { 
				if (p.get(i) != params.get(i).type()) {
					return false;
				}
			}
			return true;
		} else return false;
	}
	
	/**
	 * Returns the parameter types.
	 * 
	 * @return the parameter types.
	 */
	public Vector<DataType> getParamTypes() {
		Vector<DataType> result = new Vector<DataType>();
		for (int i=0; i < numParams(); i++) {
			result.add(params.get(i).type());
		}
		return result;
	}
	
	/**
	 * Returns the name of the method.
	 * @return The name of the method.
	 */
	public String name() { return name; }

	public String toString() {
		StringBuffer paramString = new StringBuffer();
		paramString.append("(Parameters ");
		Enumeration<VarDeclExpr> pe = params.elements();
		while (pe.hasMoreElements()) {
			paramString.append(pe.nextElement());
			if (pe.hasMoreElements())
				paramString.append(" ");
		}
		paramString.append(")");
		
		StringBuffer statementString = new StringBuffer();
		statementString.append("(Body ");
		Enumeration<Statement> se = stmnts.elements();
		while (se.hasMoreElements()) {
			statementString.append(se.nextElement());
			if (se.hasMoreElements())
				statementString.append(" ");
		}
		statementString.append(")");
		
		return "(Method " + name + " " + retType + " " + paramString + " " + statementString + ")";
	}
}
