package servilla.compiler.tree;

/**
 * A return statement stops the execution of the current method and returns
 * control to the calling method. 
 * 
 * @author Chien-Liang Fok
 */
public class ReturnStatement extends Statement {
	/**
	 * Generates the value being returned.
	 */
	private Expr retExpr = null; 
	
	/**
	 * Creates a return statement.
	 * 
	 * @param retExpr The expression that generates the value being returned.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public ReturnStatement(Expr retExpr, int lineno, int charno) {
		super(lineno, charno);
		this.retExpr = retExpr;
	}
	
	/**
	 * Creates a return statement that does not return anything.
	 * 
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public ReturnStatement(int lineno, int charno) {
		this(null, lineno, charno);
	}
	
	@Override
	public void assemble(CodeWriter cw) throws AssemblerException {
		// TODO Check whether the return type matches the stated return type of the current method
		MethodDecl md = cw.peekMethod();
		if (retExpr != null) {
			if (!md.returnType().equals(retExpr.retType(cw))) throw new AssemblerException("Cannot return type " + retExpr.retType(cw) + ", expected " + md.returnType(), lineno, charno);
			retExpr.assemble(cw);
			// The return value is passed on the stack, thus we need to tell ret to save retExpr.retValStackSize(cw) bytes.
			cw.add(new Instruction(ISA.ret, lineno, charno).setParam4(retExpr.retStackSize(cw)).setComment("Return from method " + md.name()));
		} else {
			if (!md.returnType().equals(Type.VOID)) throw new AssemblerException("Must return a " + md.returnType(), lineno, charno);
			cw.add(new Instruction(ISA.ret, lineno, charno).setParam4(0).setComment("Return from method " + md.name()));
		}
	}
	
	public String toString() { return "(RETURN " + retExpr + ")"; }

	@Override
	public long initAddr(long startAddr) { return retExpr.initAddr(startAddr); }

	@Override
	public long localVarSize() { return retExpr.localVarSize(); }
}
