package servilla.compiler.tree;

import java.util.*;

import servilla.service.tree.*;

/**
 * This is the root of the tree that encodes a program's script.
 * 
 * @author Chien-Liang Fok
 */
public class ScriptRoot {
	private Vector<ServiceDecl> 	services	= new Vector<ServiceDecl>();
	private Vector<RxnDecl> 		rxns		= new Vector<RxnDecl>();
	private Vector<VarDeclExpr> 	globals		= new Vector<VarDeclExpr>();
	private Vector<MethodDecl> 		methods		= new Vector<MethodDecl>();
	
	/**
	 * Creates a new empty Servilla Script.
	 *
	 */
	public ScriptRoot() {}
	
	/**
	 * Add a service declaration.
	 * 
	 * @param sd the service declaration.
	 */
	public void addService(ServiceDecl sd) { services.add(sd); }
	
	/**
	 * Add a reaction declaration.
	 * 
	 * @param rd the reaction declaration.
	 */
	public void addRxn(RxnDecl rd) { rxns.add(rd); }
	
	/**
	 * Add global variables.
	 * 
	 * @param globals the global variables.
	 */
	public void addGlobal(Vector<VarDeclStmt> globals) { globals.addAll(globals); }
	
	/**
	 * Add a method.
	 * 
	 * @param md the method.
	 */
	public void addMethod(MethodDecl md, int lineno, int charno) throws AssemblerException { 
		methods.add(md); 
	}
	
	public void getMethod(String name, int lineno, int charno) {
	}
	
	/**
	 * Returns a specific method declaration.
	 * 
	 * @param name The name of the method.
	 * @param pTypes The types of the input parameters.
	 * @param lineno The line number.
	 * @param charno The character number.
	 * @return The method declaration.
	 * @throws AssemblerException When no method of the specified name exists
	 */
	public MethodDecl getMethod(String name, Vector<DataType> pTypes, int lineno, int charno) throws AssemblerException {
		Enumeration<MethodDecl> e = methods.elements();
		while (e.hasMoreElements()) {
			MethodDecl m = e.nextElement();
			if (m.name().equals(name) && paramsMatch(m, pTypes)) return m; // found match
		}
		throw new AssemblerException("Unknown method \""+name+"\"", lineno, charno);
	}
	
	/**
	 * Check whether the types of the input parameters to a match match the
	 * expected input methods to the method.
	 * 
	 * @param m
	 * @param pTypes
	 * @return
	 */
	private boolean paramsMatch(MethodDecl m, Vector<DataType> pTypes) {
		if (m.numParams() == pTypes.size()) {
			for (int i=0; i<m.numParams(); i++) {
				if (m.getParamType(i) != pTypes.get(i)) return false;
			}
			return true;
		} else return false;
	}
	
	/** 
	 * Move the main method to be the first method compiled.
	 * This is just to make sure its instructions start at PC=0.
	 * 
	 * @throws AssemblerException If no main method is found.
	 */
	private void sortMethods() throws AssemblerException {
		boolean found = false;
		for (int i = 0; !found && i < methods.size(); i++) {
			MethodDecl cMeth = methods.get(i);
			if (cMeth.name().equals("main")) {
				found = true;
				if (i != 0) {
					methods.remove(i);
					methods.add(0, cMeth);
				}
			}
		}
		if (!found) throw new AssemblerException("No main method!", -1, -1);
	}
	
	/**
	 * Assembles the script.
	 * 
	 * @return A CodeWriter that is able to generate several diferent types of code.
	 * @throws AssemblerException
	 */
	public CodeWriter assemble() throws AssemblerException {
		sortMethods(); // move main method to front
		
		CodeWriter cw = new CodeWriter(methods);
		cw.setRoot(this);
		
		// on the first pass, assign address to all local variables
		Enumeration<MethodDecl> e1 = methods.elements();
		while (e1.hasMoreElements()) {
			e1.nextElement().initAddr();
		}
		
		// generate assembly code for each successive code block
		Enumeration<MethodDecl> e = methods.elements();
		while (e.hasMoreElements()) {
			e.nextElement().assemble(cw);
		}
		
		servilla.util.Debugger.dbg("ScriptRoot Assembly:", cw.toString());
		
		// expand the label references to address values
		cw.expandLabelRefs();
		
		return cw;
	}
	
	/**
	 * Returns a string representation of the script root.
	 */
	public String toString() {
		StringBuffer s = new StringBuffer();
		s.append("(ScriptRoot ");
		
		Enumeration<ServiceDecl> e1 = services.elements();
		while (e1.hasMoreElements()) {
			s.append(e1.nextElement());
			if (e1.hasMoreElements()) s.append(" ");
		}
		
		Enumeration<RxnDecl> e2 = rxns.elements();
		while (e2.hasMoreElements()) {
			s.append(e2.nextElement());
			if (e2.hasMoreElements()) s.append(" ");
		}
		
		Enumeration<VarDeclExpr> e3 = globals.elements();
		while (e3.hasMoreElements()) {
			s.append(e3.nextElement());
			if (e3.hasMoreElements()) s.append(" ");
		}
		
		Enumeration<MethodDecl> e4 = methods.elements();
		while (e4.hasMoreElements()) {
			s.append(e4.nextElement());
			if (e4.hasMoreElements()) s.append(" ");
		}
		s.append(")");
		return s.toString();
	}
}
