package servilla.compiler.tree;

import servilla.service.tree.*;
import servilla.service.*;

/**
 * A ServiceDescl is a declaration of a service specification.  A service 
 * specification is identified by its file name, which contains the actual
 * specifications and is used to invoke the service.
 * 
 * The file name is mapped to an INT, which is its handle.  The handle
 * is generated by the compiler and is used by the code to refer to a particular
 * service specification.  It is set using the setHandle(int handle) method.
 * 
 * The CodeWriter is responsible for assigning the handle to each service,
 * and treating the service specifications as global variables.
 * 
 * @author Chien-Liang Fok
 *
 */
public class ServiceDecl  {
	private String fileName;
	private int handle = -1, lineno, charno;
	
	public ServiceDecl(String fileName, int lineno, int charno) {
		this.fileName = fileName;
		this.lineno = lineno;
		this.charno = charno;
	}
	
	/**
	 * Determines whether the service declaration's handle has been set.
	 * @return whether the service declaration's handle has been set.
	 */
	public boolean hasHandle() { return handle != -1; }
	
	/**
	 * Returns the service declaration's handle.
	 * @return the service declaration's handle.
	 */
	public int handle() { return handle; }
	
	/**
	 * Returns the name of the service specification.  This is also the name of the
	 * reference to the specification using in bind and invoke operations.
	 * @return the name of the service specification.
	 */
	public String name() { return fileName; }
	
	/**
	 * Converts the service decl into a service spec.
	 * @return The service spec named by this service decl.
	 * @throws Exception
	 */
	public ServiceSpec getSpec() throws Exception {
		return new ServiceSpecC().compile(fileName);  // TODO: Make use of lineno and charno here.
	}
	
	/**
	 * Sets the service declaration's handle.
	 * 
	 * @param handle the service declaration's handle.
	 * @throws AssemblerException If the handle is not valid.
	 */
	public void setHandle(int handle) { 
		this.handle = handle; 
	}
	
	/**
	 * Returns a string representation of this class.
	 * @return a string representation of this class.
	 */
	public String toString() { return "(ServiceDecl "+fileName+" ("+handle+"))"; }
	
}
