package servilla.compiler.tree;

import java.util.*;
import servilla.service.tree.*;

/**
 * This is the super class of all statements.  A statement is a block
 * of code that does not have a return value.  It can be an expression
 * followed by a semicolon, or a control structure like a for loop that
 * is enclosed in curly braces.
 * 
 * A function's body consists of a list of statements.
 * 
 * @author Chien-Liang Fok
 */
public abstract class Statement implements Assembler {
	
	int lineno, charno;
	
	/**
	 * Creates a statement.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code this statement maps to.
	 */
	public Statement(int lineno, int charno) {
		this.lineno = lineno;
		this.charno = charno;
	}
	
	/**
	 * Returns the line number of the program code that this statement maps to.
	 * @return the line number of the program code that this statement maps to.
	 */
	public int lineno() { return lineno; }
	
	/**
	 * Returns the character number of the program code this statement maps to.
	 * @return the character number of the program code this statement maps to.
	 */
	public int charno() { return charno; }
	
	/**
	 * Pops the return value of an expressions.
	 * 
	 * @param cw The code writer that instructions should be added to.
	 * @param e The expression whose value is being popped.
	 * @param c A comment about what return value is being popped.
	 * @throws AssemblerException 
	 */
	public void popRetVal(CodeWriter cw, Expr e, String c) throws AssemblerException {
		DataType retType = e.retType(cw);
		if (retType == DataType.INT)
			cw.add(new Instruction(ISA.popInt, lineno, charno).setComment(c));
		else if (retType == DataType.FLOAT)
			cw.add(new Instruction(ISA.popFlt, lineno, charno).setComment(c));
		else
			throw new AssemblerException("Unknown data type: " + retType, lineno, charno);
	}
	
	/**
	 * Assembles a list of statements.
	 * @param vs A list of statements.
	 */
	public void assembleStatements(CodeWriter cw, Vector<Statement> vs) throws AssemblerException {
		Enumeration<Statement> e = vs.elements();
		while(e.hasMoreElements()) {
			Statement s = e.nextElement();
			s.assemble(cw);
		}
	}
	
	/**
	 * Searches for local variable declarations and assigns addresses to them.
	 *
	 * @param startAddr The starting address from which addresses may be assigned
	 * @return The next starting address that may be assigned after all of the local 
	 * variables in this statement have been allocated space.
	 */
	public abstract long initAddr(long startAddr);
	
	/**
	 * Assembles the statement.
	 * 
	 * @param cw The code writer that the resulting instructions should be saved in.
	 */
	public abstract void assemble(CodeWriter cw) throws AssemblerException;
	
	/**
	 * Returns the number of bytes needed by local variables within the expression
	 * @return the number of bytes needed by local variables within the expression
	 */
	public abstract long localVarSize();
}
