package servilla.compiler.tree;

/**
 * Encapsulates a unary expression. 
 * 
 * Unary expressions are those that involve an operation performed
 * on a single expression.
 * 
 * Ex: i++;
 * 
 * @author Chien-Liang Fok
 *
 */
public class UnaryExpr extends Expr {
	public enum UnaryExprType {PRE_INC, POST_INC, PRE_DEC, POST_DEC, NEGATE, POSITIVE, NOT};
	
	private UnaryExprType type;
	private Expr expr;
	
	/**
	 * Creatse a UnaryExpr.
	 * 
	 * @param type The type of unary expression.
	 * @param expr The expression to modify.
	 * @param lineno
	 * @param charno
	 */
	public UnaryExpr(UnaryExprType type, Expr expr, int lineno, int charno) {
		super(lineno, charno);
		this.type = type;
		this.expr = expr;
	}
	
	@Override
	public Type retType(CodeWriter cw)  throws AssemblerException { return expr.retType(cw); }

	@Override
	public int retHeapSize(CodeWriter cw)  throws AssemblerException { return expr.retHeapSize(cw); }
	
	@Override
	public long initAddr(long startAddr) { return expr.initAddr(startAddr); }

	@Override
	public int retStackSize(CodeWriter cw) throws AssemblerException { return retType(cw).stackSize(); }
	
	/**
	 * Copy the value that is on top of the stack. This assumes that the value consumes 4 bytes on the stack.
	 * 
	 * @param cw The code writer.
	 * @param retType The return type of the expression.
	 * @throws AssemblerException
	 */
	private void copyValue(CodeWriter cw, Type retType) throws AssemblerException {
		cw.add(new Instruction(ISA.duplicate, lineno, charno).setParam4(4));  // make a copy of original value to leave on the stack
	}
	
	public void assemble(CodeWriter cw) throws AssemblerException {
		Type retType = expr.retType(cw);
		if (retType.isArray() || !(retType.isInt() || retType.isFloat()))
			throw new AssemblerException("Attempting to perform " + type + " on invalid type " + retType, lineno, charno);
		
		expr.assemble(cw);  // result of expression should be on top of stack
		
		if (type == UnaryExprType.NEGATE) {
			if (retType.isInt()) cw.add(new Instruction(ISA.negInt, lineno, charno));
			else if (retType.isFloat()) cw.add(new Instruction(ISA.negFloat, lineno, charno));
			else throw new AssemblerException("Attempting to negate invalid type " + retType, lineno, charno);
		} else if (type == UnaryExprType.POSITIVE) {
			// this operation has no affect (putting a plus sign in front of a value doesn't change its value
		} else if (type == UnaryExprType.NOT) {
			if (retType.isInt()) cw.add(new Instruction(ISA.notInt_c, lineno, charno));
			else if (retType.isFloat()) cw.add(new Instruction(ISA.notFloat_c, lineno, charno));
			else throw new AssemblerException("Attempting to perform NOT on invalid type " + retType, lineno, charno);
		} else { // The instruction is inc or dec.
			
			if (type == UnaryExprType.POST_INC || type == UnaryExprType.POST_DEC) copyValue(cw, retType); // post operation
			
			if (type == UnaryExprType.POST_INC || type == UnaryExprType.PRE_INC) {
				if (retType.isInt()) cw.add(new Instruction(ISA.incInt, lineno, charno));
				else if (retType.isFloat()) cw.add(new Instruction(ISA.incFloat, lineno, charno));
				else throw new AssemblerException("Attempting to perform " + type + " on invalid type " + retType, lineno, charno);
			} else {
				if (retType.isInt()) cw.add(new Instruction(ISA.decInt, lineno, charno));
				else if (retType.isFloat()) cw.add(new Instruction(ISA.decFloat, lineno, charno));
				else throw new AssemblerException("Attempting to perform " + type + " on invalid type " + retType, lineno, charno);
			}
		
			if (type == UnaryExprType.PRE_INC || type == UnaryExprType.PRE_DEC) copyValue(cw, retType); //pre operation
		
			/*
			 * If the expression being incremented is a reference to a local variable, save it.
			 * If it is not a local variable, pop the extra value on the stack.
			 * The result should be a single value on the stack.
			 * 
			 * At this time we know that the variable type is either an integer or a float.  However,
			 * we do not know if it is 32-bits or 16-bits.
			 */
			boolean popValue = true;
			if (expr instanceof ConstExpr) {
				ConstExpr ce = (ConstExpr)expr;
				Var var = ce.var();
				if (var instanceof VarRef) { // the constant is a reference, save the resulting value
					VarRef rf = (VarRef)var;
					rf.assemble(cw, Var.AssembleType.SET);
					popValue = false;
				}
			}
			if (popValue) cw.add(new Instruction(ISA.pop4, lineno, charno)); // discard the extra value on the stack
		}
	}
	
	public String toString() { return "(UnaryExpr: " + type + " " + expr + ")"; }

	@Override
	public long localVarSize() { return expr.localVarSize(); }
}
