package servilla.compiler.tree;

import servilla.service.tree.*;

/**
 * The parent class of a IntVal, FloatVal, StringVal, or VarRef.
 * 
 * @author Chien-Liang Fok
 */
public abstract class Var {
	public enum AssembleType {GET, SET};
	int lineno, charno;
	
	public Var(int lineno, int charno) {
		this.lineno = lineno;
		this.charno = charno;
	}
	
	public int lineno() { return lineno; }
	public int charno() { return charno; }
	
	/**
	 * Returns the type of data that this variable represents.
	 * 
	 * @param cw The Code Writer (provides the frame in which the variable is defined).
	 * @return the type of this variable.
	 */
	public abstract DataType retType(CodeWriter cw) throws AssemblerException;
	
	/**
	 * Returns the number of bytes that is pushed onto the stack when this var is accessed.
	 * 
	 * @param cw The Code Writer (provides the frame in which the variable is defined).
	 * @return The size of the returned data.
	 * @throws AssemblerException
	 */
	public abstract long retSize(CodeWriter cw) throws AssemblerException;
	
	/**
	 * Assembles the variable. 
	 * 
	 * @param cw The code writer.
	 * @param aType Whether to generate code to read or write to the variable.
	 * @throws AssemblerException
	 */
	public abstract void assemble(CodeWriter cw, AssembleType aType) throws AssemblerException;
	
	/**
	 * Returns the number of bytes needed by the variable on the heap.
	 * 
	 * Note that an integer or float constant does not consume any space
	 * on the heap since it's value is stored in code memory.
	 * 
	 * A String constant, however, does consume memory on the heap since
	 * the value of the string is stored in heap memory.
	 */
	public abstract long heapSize();
}
