package servilla.compiler.tree;

import servilla.service.tree.DataType;

/**
 * A VarDeclExpr declares a variable.  It consists of a type and name (identifier).
 * This is the super class of StringDeclExpr, IntDeclExpr, and FloatDeclExpr.
 * 
 * @author Chien-Liang Fok
 */
public abstract class VarDeclExpr extends Expr {
	/**
	 * The name of the variable. For example, in the declaration "int count", 
	 * the name of the variable is "count".
	 */
	private String name;
	
	/**
	 * The variable's data type.  For example, in the declaration "int count", 
	 * the type of the variable is 'int'.
	 */
	private DataType type;
	
	/**
	 * The variable's initial value, if applicable.
	 */
	private Expr initVal = null; // default no initial value
	
	/**
	 * The memory address of the variable.  This address may be relative, indirect, 
	 * or absolute depending on the context in which it is used.  It is set by the 
	 * compiler during compile-time.
	 */
	private long address;

	/** 
	 * The constructor for a variable without an initial value.
	 * 
	 * @param name The name of the variable.
	 * @param type The variable type
	 * @param lineno The line number.
	 * @param charno The character number.
	 */
	public VarDeclExpr(String name, DataType type, int lineno, int charno) {
		this(name, type, null, lineno, charno);
	}
	
	/** 
	 * The constructor for a variable with an initial value
	 * 
	 * @param name The name of the variable.
	 * @param type The variable type
	 * @param initVal An expression generating the initial value.
	 * @param lineno The line number.
	 * @param charno The character number.
	 */
	public VarDeclExpr(String name, DataType type, Expr initVal, int lineno, int charno) {
		super(lineno, charno);
		this.name = name;
		this.type = type;
		this.initVal = initVal;
	}
	
	/**
	 * Returns the variable's name.
	 * 
	 * @return the variable's name.
	 */
	public String name() { return name; }
	
	
	/**
	 * Returns the variable's primitive data type.  Note that the user can determine whether the
	 * variable is an array by calling numDim(), which returns the number of array dimensions.
	 *
	 * @return the variable's data type.
	 */
	public DataType type() { return type;}
	
	/**
	 * Returns the expression that generates the initial value of the variable.
	 * 
	 * @return  The initial value expression.
	 */
	public Expr initVal() { return initVal; }
	
	/**
	 * Returns the address of this variable.
	 * 
	 * @return the address of this variable.
	 */
	public long address() throws AssemblerException { 
		if (address < 0) 
			throw new AssemblerException("Address of \"" + name + "\" not set.", lineno, charno);
		else
			return address; 
	}

	/**
	 * Even though a VarDeclExpr is an expression, it has no return value.  
	 * Thus, it may be used within a for statement:
	 * 
	 * for (VarDeclExpr; ...; ...) {... }
	 * 
	 * But it may <b>not</b> be used in a while statement:
	 * 
	 * while (VarDeclExpr < 10) {...}
	 * 
	 * @param cw The code writer.
	 * @return DataType.VOID
	 */
	@Override
	public DataType retType(CodeWriter cw) throws AssemblerException { return DataType.VOID; }

	/**
	 * A VarDeclExpr has no return value; its return size is zero.
	 */
	@Override
	public int retSize(CodeWriter cw) throws AssemblerException { return 0; }
	
	/**
	 * Searches for local variable declarations and assigns addresses to them.
	 *
	 * @param startAddr The starting address from which addresses may be assigned
	 * @return The next starting address that may be assigned after all of the local 
	 * variables in this expression have been allocated space.
	 */
	@Override
	public long initAddr(long startAddr) { address = startAddr; return address + localVarSize(); }
	
	/**
	 * Generates an assignment statement that assigns an initial value to a variable.
	 * 
	 * @param cw The code writer.
	 * @param ref The variable reference.
	 * @param initValExpr The expression that generates the initial value.
	 * @throws AssemblerException If there is any problem compiling.
	 */
	protected void initVar(CodeWriter cw, VarRef ref, Expr initValExpr) throws AssemblerException {
		AssignmentExpr assignExpr = new AssignmentExpr(ref, initValExpr, lineno, charno);
		AssignmentStatement assignStmt = new AssignmentStatement(assignExpr, lineno, charno);
		assignStmt.assemble(cw);
	}
	
	/**
	 * Returns a string representation of this class.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append("(VarDeclExpr "+type+" "+name);
		if (initVal != null)
			sb.append(initVal.toString());
		sb.append(")");
		return sb.toString();
	}
}
