package servilla.compiler.tree;

import java.util.*;

import servilla.service.tree.DataType;

/**
 * A variable reference is a string that references a previously
 * defined variable or service.  By the time this is reached by the compiler,
 * a VarDeclExpr of the same name must already have been compiled, otherwise
 * there will be an error dereferencing the variable.
 * 
 * The VarRef class also defines global constants.  Global constants are mapped
 * to a sequence of assembly instructions using the "reservedWords" hashtable.
 * 
 * @author Chien-Liang Fok
 *
 */
public class VarRef extends Var {
	private String name;
	
	/**
	 * This defines global constants.  Some constant expressions have special meanings that
	 * need to be assembled in a special way.
	 * 
	 * Ex: "true", "false"
	 */
	public static final Hashtable<String,Vector<Instruction>> reservedWords = new Hashtable<String,Vector<Instruction>>();
	{
		Instruction i1t = new Instruction(ISA.pushIntConst, lineno, charno).setParam(1);
		Instruction i2t = new Instruction(ISA.pull_c, lineno, charno).setComment("Set condition code to 1");
		Instruction i3t = new Instruction(ISA.pushIntConst, lineno, charno).setParam(1).setComment("Ensure \"true\" expr leaves something on stack");
		Vector<Instruction> vtrue = new Vector<Instruction>();
		vtrue.add(i1t); vtrue.add(i2t); vtrue.add(i3t);
		reservedWords.put("true", vtrue);
		
		Instruction i1f = new Instruction(ISA.pushIntConst, lineno, charno).setParam(0);
		Instruction i2f = new Instruction(ISA.pull_c, lineno, charno).setComment("Set condition code to 0");
		Instruction i3f = new Instruction(ISA.pushIntConst, lineno, charno).setParam(0).setComment("Ensure \"false\" expr leaves something on stack");
		Vector<Instruction> vfalse = new Vector<Instruction>();
		vfalse.add(i1f); vfalse.add(i2f); vfalse.add(i3f);
		reservedWords.put("false", vfalse);
		
		// Define the macro replacements
		Vector<Instruction> vGreen = new Vector<Instruction>();
		vGreen.add(new Instruction(ISA.pushIntConst, lineno, charno).setParam(2));
		reservedWords.put("GREEN", vGreen);
		
		Vector<Instruction> vRed = new Vector<Instruction>();
		vRed.add(new Instruction(ISA.pushIntConst, lineno, charno).setParam(1));
		reservedWords.put("RED", vRed);
		
		Vector<Instruction> vBlue = new Vector<Instruction>();
		vBlue.add(new Instruction(ISA.pushIntConst, lineno, charno).setParam(4));
		reservedWords.put("BLUE", vBlue);
		
		BindType[] bindTypes = BindType.values();
		for (int i = 0; i < bindTypes.length; i++) {
			Vector<Instruction> v = new Vector<Instruction>();
			v.add(new Instruction(ISA.pushIntConst, lineno, charno).setParam(bindTypes[i].value()));
			reservedWords.put(bindTypes[i].toString(), v);
		}
	}
	
	/**
	 * Creates a VarRef with no array indices.
	 * 
	 * @param name The name of the variable being referenced.
	 * @param lineno
	 * @param charno
	 */
	public VarRef(String name, int lineno, int charno) {
		super(lineno, charno);
		this.name = name;
	}
	
	/**
	 * Returns the name of the variable being referenced.
	 * @return the name of the variable being referenced.
	 */
	public String name() { return name; }
	
	/**
	 * The return type depends on the type of variable being referenced.
	 * 
	 * <p>If the variable is a string or an array, the return type will be a 
	 * memory address, which is where the variable is being stored.  The 
	 * memory address's type is an INT.</p>
	 * 
	 * <p>If the variable is a value like an integer or float, it will return
	 * the integer and float directly.</p>
	 *
	 * <p>If the variable is a reserved word, the return type is an INT.</p>
	 * 
	 * @param cw The code writer.  This is necessary to find the original variable
	 * declaration.
	 */
	@Override
	public DataType retType(CodeWriter cw) throws AssemblerException {
		if (reservedWords.containsKey(name)) 
			return DataType.INT;
		else {
			VarDeclExpr varDecl = cw.getLocalVar(name, lineno, charno);
			return varDecl.type();
		}
	}
	
	/**
	 * If a global constant is being referenced, the return value size that
	 * of an integer.
	 * Otherwise, it will return the size of the data pushed onto the stack.  Note that this is the number of bytes
	 * pushed onto the stack, not heap.
	 */
	@Override
	public long retSize(CodeWriter cw) throws AssemblerException { 
		if (reservedWords.containsKey(name))
			return DataType.INT.size();
		else {
			VarDeclExpr varDecl = cw.getLocalVar(name, lineno, charno);
			return varDecl.retSize(cw);
		}
	}
	
	/**
	 * Variable references do not consume memory in the heap.
	 * Thus, this method returns zero.
	 */
	@Override
	public long heapSize() { return 0; }
	
	/**
	 * Pushes the absolute address of the variable being referenced.
	 * 
	 * @param cw The code writer.
	 * @throws AssemblerException When this is not referencing a local variable.
	 */
	public void pushAddress(CodeWriter cw) throws AssemblerException {
		VarDeclExpr varDecl = cw.getLocalVar(name, lineno, charno); 
		
		/*
		 * It is possible for the variable to be an array but the arrayIndxExpr is null.
		 * For example, when the user wants to send an array to the base station, he should
		 * be able to execute "send(myArray)".  Notice how there are no square brackets indexing
		 * the specific element within myArray to send.
		 */
		if (arrayIndxExpr != null) {
			long baseAddr = varDecl.address();
			for (int i = 0; i < arrayIndxExpr.size(); i++) {
	blah blah blah
				/*
				 * For now, assume that a signed integer is OK for use as an array index.
				 * Trust that the user will not insert a negative number into the array.
				 * TODO: See if we can check whether the arrayIndxExpr evaluates to a negative or out-of-bounds value.
				 */
				if (!arrayIndxExpr.get(i).retType(cw).isInt()) throw new AssemblerException("Array index not an integer ("+arrayIndxExpr.get(i).retType(cw)+")", lineno, charno);
				arrayIndxExpr.get(i).assemble(cw);
				cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(varDecl.type().size()));
				cw.add(new Instruction(ISA.multInt, lineno, charno).setComment("Calculating array index offset of variable"));	
			}
			cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(varDecl.address()));
			cw.add(new Instruction(ISA.addInt, lineno, charno).setComment("Calculating address of variable")); // variable address = indx*var_type_size+offset
		} else {
			cw.add(new Instruction(ISA.push4, lineno, charno).setParam4(varDecl.address())); // variable address = offset
		}
	}
	
	/**
	 * Generates the bytecode that fetches the value of the variable being referenced.
	 */
	public void assemble(CodeWriter cw, AssembleType aType) throws AssemblerException {
		// If the variable reference is a constant, push its value onto the stack.
		if (reservedWords.containsKey(name)) {
			if (arrayIndxExpr != null) throw new AssemblerException("Tried to access a global constant "+name+" as if it were an array", lineno, charno);
			if (aType == AssembleType.SET) throw new AssemblerException("Tried to write to global constant "+name+"", lineno, charno);
			cw.add(reservedWords.get(name));
			return;
		}
		
		try {
			// See if the variable being referenced is a local variable.
			pushAddress(cw);
			VarDeclExpr varDecl = cw.getLocalVar(name, lineno, charno); 
			
			/*
			 * If the type of the variable being referenced is an array, but the index into the array is not
			 * specified, only push the address of the array onto the stack.
			 */
			if (varDecl.type().isArray() && arrayIndxExpr == null) return;
			
			switch(varDecl.type().size()) {
			case 2:
				if (aType == AssembleType.GET) 
					cw.add(new Instruction(ISA.glv2, lineno, charno).setComment("Fetch 16-bit variable \"" + name + "\""));
				else
					cw.add(new Instruction(ISA.slv2, lineno, charno).setComment("Save 16-bit variable \"" + name + "\""));
				break;
			case 4:
				if (aType == AssembleType.GET)
					cw.add(new Instruction(ISA.glv4, lineno, charno).setComment("Fetch 32-bit variable \"" + name + "\""));
				else
					cw.add(new Instruction(ISA.slv4, lineno, charno).setComment("Save 32-bit variable \"" + name + "\""));
				break;
			default:
				if (aType == AssembleType.GET)
					cw.add(new Instruction(ISA.glvString, lineno, charno).setComment("Fetch String variable \"" + name + "\""));
				else
					cw.add(new Instruction(ISA.slvString, lineno, charno).setComment("Save String variable \"" + name + "\""));
			}
		} catch(AssemblerException ae) {
			// see if it is a reference to a service
			ServiceDecl spec = cw.getDecl(name);
			if (spec == null) throw new AssemblerException("Unknown variable reference \""+name+"\"", lineno, charno);
			cw.add(new Instruction(ISA.push2, lineno, charno).setParam2(spec.handle()).setComment("Fetch service spec "+spec.name()));
		}
	}
	
	public String toString() {
		if (arrayIndxExpr == null)
			return "(VarRef " + name +")";
		else
			return "(VarRef " + name + " " + arrayIndxExpr + ")";
	}
}