package servilla.compiler.tree;
import java.util.*;


/**
 * An while statement repeatedly executes a list of statements
 * so long as the while expression remains true.
 * 
 * @author Chien-Liang Fok
 */
public class WhileStatement extends Statement {
	private Expr whileExpr; // the variable being written to
	private Vector<Statement> body;
	
	/**
	 * Creates a while statement.
	 * 
	 * @param whileExpr The conditional expression that is evaluated to determine whether to continue to execute the while loop.
	 * @param body The block of statements that is executed each time the while loop loops.
	 * @param lineno The line number of the program code that this statement maps to.
	 * @param charno The character number of the program code that this statement maps to.
	 */
	public WhileStatement(Expr whileExpr, Vector<Statement> body, int lineno, int charno) {
		super(lineno, charno);
		/*
		 * WhileExpr may also be a constant expression, e.g., "true" or "false"
		 */
		//if (! (whileExpr instanceof CondExpr)) throw new AssemblerException ("While statement expression not a CondExpr", lineno, charno);
		this.whileExpr = whileExpr;
		this.body = body;
	}

	
	@Override
	public void assemble(CodeWriter cw) throws AssemblerException {
		String whileLoopLabel = "WHILE" + cw.nonce();
		String continueLabel = "CONTINUE" + cw.nonce();
		
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("Begin WHILE loop (line="+lineno+")").setLabel(whileLoopLabel));
		whileExpr.assemble(cw);
		popRetVal(cw, whileExpr, "Discard result of while expr");
		cw.add(new Instruction(ISA.neg_c, lineno, charno)); // negate the condition code set by the while expression
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(continueLabel));
		cw.add(new Instruction(ISA.jump_c, lineno, charno).setComment("Jump to end of WHILE loop if condition is false"));
		assembleStatements(cw, body);
		cw.add(new Instruction(ISA.push4, lineno, charno).setLabelParam(whileLoopLabel));
		cw.add(new Instruction(ISA.jump, lineno, charno).setComment("Jump back to beginning of WHILE loop"));
		cw.add(new Instruction(ISA.noop, lineno, charno).setComment("End of WHILE loop (line="+lineno+")").setLabel(continueLabel));
	}
	
	public String toString() {
		StringBuffer sb = new StringBuffer();
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			sb.append(e.nextElement());
			if (e.hasMoreElements())
				sb.append(" ");
		}
		return "(WHILE " + whileExpr + " (" + sb + "))";
	}


	@Override
	public long initAddr(long startAddr) {
		long result = startAddr;
		result = whileExpr.initAddr(result);
		Enumeration<Statement> e = body.elements();
		while (e.hasMoreElements()) {
			Statement s = e.nextElement();
			result = s.initAddr(result);
		}
		return result;
	}
	
	@Override
	public long localVarSize() {
		long result = whileExpr.localVarSize();
		Enumeration<Statement> bv = body.elements();
		while (bv.hasMoreElements()) {
			result += bv.nextElement().localVarSize();
		}
		return result;
	}
}
