package servilla.network;

import servilla.util.*;
import net.tinyos.message.*;

/**
 * This is an acknowledgement used for reliably sending messages in
 * a sensor network.
 * 
 * @author Chien-Liang Fok
 */
public class RAckMsgJ implements MessageJ {
	private int result;
	private long startAddr;
	short am;
	
	/**
	 * Default constructor.
	 */
	public RAckMsgJ() {}
	
	/**
	 * Creates a ReliableAckMsgJ.
	 * 
	 * @param result The result of the send.
	 * @param startAddr The start address of the sent message.
	 * @param am The AM type of the sent message.
	 */
	public RAckMsgJ(int result, long startAddr, short am) {
		this.result = result;
		this.startAddr = startAddr;
		this.am = am;
	}

	/**
	 * Creates an RAckMsgJ based on an RAckMsg.
	 *
	 * @param msg The MIG-generated message.
	 */
	public RAckMsgJ(RAckMsg msg) {
		result = msg.get_result();
		startAddr = msg.get_startAddr();
		am = msg.get_am();
	}
	
	/**
	 * Returns the result of the send.
	 * 
	 * @return The result of the send.
	 */
	public int result() { return result; }
	
	/**
	 * Returns the start address of the sent message.
	 * 
	 * @return The start address of the sent message.
	 */
	public long startAddr() { return startAddr;}
	
	/**
	 * Returns the AM type of the sent message.
	 * 
	 * @return the AM type of the sent message.
	 */
	public short am() { return am; }
	
	/**
	 * Returns the AM type of this message.
	 * 
	 * @return the AM type of this message.
	 */
	public int getType() { return RAckMsg.AM_TYPE;}
	
	/**
	 * Generates an equivalent MessageJ from a TinyOS message.
	 * 
	 * @param msg The TinyOS message.
	 * @return An equivalent Servilla message.
	 */
	public MessageJ clone(Message msg) {
		return new RAckMsgJ((RAckMsg)msg);	
	}
	
	/**
	 * Converted this Java object-based version of the ReliableAckMsg
	 * into one compatible with TinyOS.
	 */
	public Message toTOSMsg() {
		RAckMsg msg = new RAckMsg();
		msg.set_result(result);
		msg.set_startAddr(startAddr);
		msg.set_am(am);
		return msg;
	}
	
	/**
	 * Returns the string representation of this class.
	 * 
	 * @return the string representation of this class.
	 */
	public String toString() {
		return "RAckMsgJ: result=" + result + ", startAddr=" + startAddr + ", am=" + ArrayToHexString.shortToHex(am);
	}
}

