// $Id: ServillaAckStateMsgJ.java,v 1.2 2005/11/11 02:15:49 chien-liang Exp $

/* Servilla - A middleware for wireless sensor networks.
 * Copyright (C) 2007, Washington University in Saint Louis
 * By Chien-Liang Fok.
 *
 * Washington University states that Servilla is free software;
 * you can redistribute it and/or modify it under the terms of
 * the current version of the GNU Lesser General Public License
 * as published by the Free Software Foundation.
 *
 * Servilla is distributed in the hope that it will be useful, but
 * THERE ARE NO WARRANTIES, WHETHER ORAL OR WRITTEN, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.
 *
 * YOU UNDERSTAND THAT AGILLA IS PROVIDED "AS IS" FOR WHICH NO
 * WARRANTIES AS TO CAPABILITIES OR ACCURACY ARE MADE. THERE ARE NO
 * WARRANTIES AND NO REPRESENTATION THAT AGILLA IS FREE OF
 * INFRINGEMENT OF THIRD PARTY PATENT, COPYRIGHT, OR OTHER
 * PROPRIETARY RIGHTS.  THERE ARE NO WARRANTIES THAT SOFTWARE IS
 * FREE FROM "BUGS", "VIRUSES", "TROJAN HORSES", "TRAP DOORS", "WORMS",
 * OR OTHER HARMFUL CODE.
 *
 * YOU ASSUME THE ENTIRE RISK AS TO THE PERFORMANCE OF SOFTWARE AND/OR
 * ASSOCIATED MATERIALS, AND TO THE PERFORMANCE AND VALIDITY OF
 * INFORMATION GENERATED USING SOFTWARE. By using Servilla you agree to
 * indemnify, defend, and hold harmless WU, its employees, officers and
 * scripts from any and all claims, costs, or liabilities, including
 * attorneys fees and court costs at both the trial and appellate levels
 * for any loss, damage, or injury caused by your actions or actions of
 * your officers, servants, scripts or third parties acting on behalf or
 * under authorization from you, as a result of using Servilla.
 *
 * See the GNU Lesser General Public License for more details, which can
 * be found here: http://www.gnu.org/copyleft/lesser.html
 */
package servilla.network;

import net.tinyos.message.*;
import java.util.*;
import servilla.util.*;

/**
 * This is the message used for reliably sending messages in
 * a sensor network.
 * 
 * @author Chien-Liang Fok
 */
public class RMsgJ implements MessageJ {
	int replyAddr;
	long startAddr, size;
	short am;
	Vector<Short> data;
	
	/**
	 * Default constructor.
	 */
	public RMsgJ() { data = new Vector<Short>();}
	
	/**
	 * Creates a ReliableMsgJ.
	 * 
	 * @param result The result of the send.
	 * @param startAddr The start address of the sent message.
	 * @param am The AM type of the sent message.
	 */
	public RMsgJ(int replyAddr, long startAddr, long size, short am, Vector<Short> data) {
		this.replyAddr = replyAddr;
		this.startAddr = startAddr;
		this.size = size;
		this.am = am;
		this.data = data;
	}

	/**
	 * This constructor creates a Java object-based ReliableMsg out of
	 * the class generated by MIG.
	 *
	 * @param msg The MIG-generated message.
	 */
	public RMsgJ(RMsg msg) {
		replyAddr = msg.get_replyAddr();
		startAddr = msg.get_startAddr();
		size = msg.get_size();
		am = msg.get_am();
		data = new Vector<Short>();
		for (int i = 0; i < RMsg.numElements_data(); i++) {
			data.add(msg.getElement_data(i));
		}
	}
	
	/**
	 * Returns the result of the send.
	 * 
	 * @return The result of the send.
	 */
	public int replyAddr() { return replyAddr; }
	
	/**
	 * Returns the start address of the sent message.
	 * 
	 * @return The start address of the sent message.
	 */
	public long startAddr() { return startAddr;}
	
	/**
	 * Returns the total number of bytes being sent.
	 * 
	 * @return the total number of bytes being sent.
	 */
	public long size() { return size;}
	
	/**
	 * Returns the AM type of the sent message.
	 * 
	 * @return the AM type of the sent message.
	 */
	public short am() { return am; }
	
	/**
	 * Returns the data fragment included in this message.
	 * 
	 * @return the data fragment included in this message.
	 */
	public Vector<Short> data() { return data; }
	
	/**
	 * Returns the number of bytes of data inlcuded in this message.
	 * 
	 * @return the number of bytes of data inlcuded in this message.
	 */
	public int numDataBytes() { 
		if (size - startAddr < data().size())
			return (int)(size-startAddr);
		else
			return data().size();
	}
	
	public long numMsgs() {
		long numMsgs = size / RMsg.numElements_data(); 
		if (size % RMsg.numElements_data() != 0) numMsgs++;
		return numMsgs;
	}
	
	public long msgNum() {
		long msgNum = startAddr / RMsg.numElements_data(); 
		if (startAddr % RMsg.numElements_data() != 0) msgNum++;
		return msgNum+1;
	}
	
	/**
	 * Returns the AM type of this message.
	 * 
	 * @return the AM type of this message.
	 */
	public int getType() { return RMsg.AM_TYPE;}
	
	/**
	 * Generates an equivalent MessageJ from a TinyOS message.
	 * 
	 * @param msg The TinyOS message.
	 * @return An equivalent Servilla message.
	 */
	public MessageJ clone(Message msg) {
		return new RMsgJ((RMsg)msg);
	}
	
	/**
	 * Converted this Java object-based version of the RMsg
	 * into one compatible with TinyOS.
	 */
	public Message toTOSMsg() {
		RMsg msg = new RMsg();
		//msg.set_dest(dest);
		msg.set_replyAddr(replyAddr);
		msg.set_startAddr(startAddr);
		msg.set_size(size);
		msg.set_am(am);
		for (int i = 0; i < data.size(); i++) {
			msg.setElement_data(i, data.get(i));
		}
		return msg;
	}
	
	/**
	 * Returns the string representation of this class.
	 * 
	 * @return the string representation of this class.
	 */
	public String toString() {
		String result = "(ReliablMsgJ: replyAddr=" + replyAddr + ", startAddr=" + startAddr + ", size=" + size + " am=" + ArrayToHexString.shortToHex(am)+")";
		result += "\ndata=["+ArrayToHexString.toString(org.apache.commons.lang.ArrayUtils.toPrimitive(data.toArray(new Short[]{})))+"]";
		return result;
	}
}

