package servilla.network;

import servilla.*;
import servilla.util.*;

import java.util.*;

/**
 * Reliably receives data from the sensor network.
 * 
 * @author Chien-Liang Fok
 */
public class RMsgRcvr implements ServillaConstants, MessageListenerJ {
	private SNI sni;
	
	/**
	 *  A collection of DataMsg objects, one for each incoming reliable
	 *  data message. 
	 *  
	 *  Key: the reply address
	 *  Value: The DataMsg object
	 */
	private Hashtable<Integer, DataMsg> inBuff = new Hashtable<Integer, DataMsg>();
	// TODO: Remove stale DataMsg messages from the inBuff.
	
	/**
	 * A collection of listeners.
	 * Key: the active message number
	 * Value: A vector of RDataRcvr objects (listeners).
	 */
	private Hashtable<Integer, Vector<RMsgListener>> listeners = new Hashtable<Integer, Vector<RMsgListener>>();
	
	/**
	 * Creates a reliable receiver.
	 * 
	 * @param sni The sensor network interface.
	 */
	public RMsgRcvr(SNI sni) {
		this.sni = sni;
		sni.registerListener(new RMsgJ(), this);
	}
	
	/**
	 * Sends an acknowledgement to the sender.
	 * 
	 * @param result either ServillaConstants.SUCCESS or ServillaConstants.FAIL
	 * @param rmj The message that is being acknowledged.
	 */
	private void sendAck(int result, RMsgJ rm) {
		RAckMsgJ ack = new RAckMsgJ(result, rm.startAddr(), rm.am());
		try {
			sni.send(rm.replyAddr(), ack);
			Debugger.log("RMsgRcvr", "Sent ACK to " + rm.replyAddr());
			Debugger.dbg("RMsgRcvr", rm.toString());
		} catch (java.io.IOException e) {
			Debugger.log("RMsgRcvr", "ERROR: Unable to send ACK " + ack);
		}
	}
	
	/**
	 * When a RMsgJ is received, obtain the DataMsg that is responsible for aggregating
	 * the data and save the data contained within the RMsgJ into DataMsg.
	 * 
	 * @param dest the destination of the message
	 * @param msg the received message
	 */
	public void messageReceived(int dest, MessageJ msg) {
		RMsgJ rm = (RMsgJ)msg;
		DataMsg dm = inBuff.get(rm.replyAddr());
		boolean saved = false;
		if (dm == null) {
			if (rm.startAddr() == 0) {
				dm = new DataMsg(rm);	// create a new DataMsg for keeping track of incoming reliable messages
				inBuff.put(rm.replyAddr(), dm);
				saved = true;
			} else {
				Debugger.log("RMsgRcvr", "ERROR: Unexpected RMsgJ:\n" + rm);
			}
		} else {
			dm.saveMsg(rm);
			saved = true;
		}
		
		Debugger.log("RMsgRcvr", "Received msg " + rm.msgNum() + " of " + rm.numMsgs() + "\tTIME = " + System.nanoTime());
		Debugger.dbg("RMsgRcvr", "Received reliable message:\n" + rm);
		Debugger.dbg("RMsgRcvr", "Data message:\n" + dm);
		//Debugger.dbg("RMsgRcvr", "startAddr=" + rm.startAddr() + ", numDataBytes="+rm.numDataBytes() + ", numBytes=" + rm.size());
		sendAck(saved ? SUCCESS : FAIL, rm);
		if (saved && (rm.startAddr() + rm.numDataBytes()) == rm.size()) {
			inBuff.remove(rm.replyAddr()); // remove the message from the table
			Debugger.log("RMsgRcvr", "*** Received data message ***\n" + dm);
			notifyListeners(dm);
		}
	}
	
	/**
	 * Notifies the listeners that a data message has arrived.
	 * 
	 * @param dm The data message that was received.
	 */
	private void notifyListeners(DataMsg dm) {
		Vector<RMsgListener> toNotify = listeners.get(dm.am());
		if (toNotify != null) {
			for (int i = 0; i < toNotify.size(); i++) {
				RMsgListener rdr = toNotify.get(i);
				Debugger.log("RMsgRcvr", "Notifying listener \"" + rdr + "\" of message with type " + dm.am());
				rdr.rcvdMsg(dm.am(), dm.data());
			}
		} else 
			Debugger.log("RMsgRcvr", "No listeners for msg:\n" + dm);
	}
	
	/**
	 * Adds a data receiver that should be notified when a certain type of
	 * reliable message is received.
	 * 
	 * @param id The AM type of the message.
	 * @param rcvr The object that should be notified.
	 */
	public void addMsgListener(int id, RMsgListener rcvr) {
		Vector<RMsgListener> list = listeners.remove(id);
		if (list == null) {
			list = new Vector<RMsgListener>();
		}
		if (!list.contains(rcvr)) { 
			Debugger.log("RMsgRcvr", rcvr + " is listening to messages with ID=" + id);
			list.add(rcvr);
		} else
			Debugger.log("RMsgRcvr", "Attempting to add a listener twice.");
		
		listeners.put(id, list);
	}
	
	/**
	 * Deregisters a listener.
	 * 
	 * @param id The AM type of the message being listened for.
	 * @param rcvr The object that was listening to the message.
	 */
	public void removeMsgListener(int id, RMsgListener rcvr) {
		Vector<RMsgListener> list = listeners.get(id);
		if (list != null) list.remove(rcvr);
	}
}