package servilla.network;

import servilla.util.*;
import servilla.*;
import java.util.*;

/**
 * Reliably sends data into the sensor network.
 * 
 * @author Chien-Liang Fok
 *
 */
public class RMsgSndr implements ServillaConstants, MessageListenerJ {
	private SNI sni;
	private Object lock = new Object();   // a lock for threads to wait on while waiting for an ACK
	private RAckMsgJ ackMsg;
	private TimeoutTimer timeoutTimer;
	
	/**
	 * Creates a reliable sender.
	 * 
	 * @param sni The sensor network interface.
	 */
	public RMsgSndr(SNI sni) {
		this.sni = sni;
		sni.registerListener(new RAckMsgJ(), this);
	}
	
	/**
	 * Sends data to a node in the sensor network.  This method is synchronized
	 * to prevent multiple threads from sending a reliable message.
	 * 
	 * @param dest The destination address.
	 * @param amType The message ID.
	 * @param data The data to send.
	 * @return true if successful, false otherwise.
	 */
	public synchronized boolean send(int dest, short amType, byte[] data) {
		DataMsg dm = new DataMsg(dest, amType, data);
		int startAddr = 0;
		int numMsgs = dm.numBytes() / RMsg.numElements_data();	// calculate the number of RMsgs that need to be sent
		if (numMsgs * RMsg.numElements_data() < dm.numBytes()) numMsgs++;
		for (int i=0; i<numMsgs; i++) {
			int numBytes = RMsg.numElements_data();
			if (startAddr + numBytes > dm.numBytes()) numBytes = dm.numBytes() - startAddr;
			Vector<Short> dataFrag = new Vector<Short>();
			for (int j = 0; j < numBytes; j++) {
				dataFrag.add(dm.data(j+startAddr));
			}
			RMsgJ currMsg = new RMsgJ(TOS_UART_ADDRESS, startAddr, data.length, (short)dm.am(), dataFrag);
			Debugger.log("ReliableSender.send", "Sending msg " + currMsg);
			if (!sendMsg(currMsg)) {
				Debugger.err("ReliableSender.send:", "ERROR: Failed to send RMsg " + (i+1) + " of " + numMsgs + ".");
				return false;
			}
			startAddr += RMsg.numElements_data();
		}
		return true;
	}
	
	/**
	 * Sends a reliableMsg.
	 * 
	 * @param msg The message to send.
	 * @return true if the reliable msg was sent.
	 */
	private boolean sendMsg(RMsgJ msg) {
		int retryCounter = 0;
		while(retryCounter++ < ServillaProps.getProps().RS_MAX_RXMIT_COUNT()) {
			ackMsg = null;
			try {
				sni.send(TOS_BCAST_ADDRESS, msg);
			} catch(java.io.IOException e) {
				e.printStackTrace();
			}
			timeoutTimer = new TimeoutTimer(lock, RS_SNDR_RXMIT_TIMER_PERIOD);
			logSendMsg("Awaiting ReliableAckMsg message from " + msg.replyAddr() + "...");
			
			boolean done = false;
			while (!done) {
				waitForAck();
				if (ackMsg == null) {
					done = true;  // break out of inner while loop
					logSendMsg("Timeout "+ retryCounter +"/"+ServillaProps.getProps().RS_MAX_RXMIT_COUNT()+" waiting for ack from " + msg.replyAddr() + "...");
				} else {
					RAckMsgJ ack = (RAckMsgJ)ackMsg;
					if (ack.startAddr() == msg.startAddr()) {
						if (ack.result() == SERVILLA_ACCEPT || ack.result() == SERVILLA_DUPLICATE) {
							logSendMsg("Receiver ACCEPTED message with start address "+msg.startAddr()+", proceeding to next one...");
							return true;
						} else {
							logSendMsg("Receiver REJECTED message with start address "+msg.startAddr()+", aborting...");
							return false;
						}
					} else {
						logSendMsg("Received ACK was for the wrong RMsg, got " + ack.startAddr() + " expected " + msg.startAddr());
					}
				}
			}
		} // while...
		logSendMsg("ERROR: Exceeded maximum number of timeouts when sending state.");
		return false;
	}
	
	/**
	 * Wait for an ack message if it has not been received.
	 */
	private void waitForAck() {
		synchronized(lock) {
			if (ackMsg == null) {
				try {
					lock.wait();
				} catch(Exception e) {
					e.printStackTrace();
				}
			} else { // an ack message was already received
				if (timeoutTimer != null) timeoutTimer.kill();
			}
		}
	}
	
	/**
	 * This is called when a message is received.
	 */
	public void messageReceived(int to, MessageJ m) {
		if (m.getType() == RAckMsg.AM_TYPE) {
			synchronized(lock) {
				if (timeoutTimer != null) {
					timeoutTimer.kill();
					this.ackMsg = (RAckMsgJ)m;
					try { lock.notifyAll(); } catch (Exception e) { e.printStackTrace();}
				}
			}
		}
	}
	
	private void logSendMsg(String msg) { Debugger.log("ReliableSender.sendMsg:", msg); }
}
