package servilla.network;

import servilla.ServillaConstants;
import servilla.msgs.*;
import servilla.util.Debugger;
import net.tinyos.message.*;
import java.util.*;

/**
 * The sensor network interface wraps the TinyOS messages produced by MoteIF
 * into higher level message objects.
 *
 * @author Chien-Liang Fok
 */
public class SNI implements ServillaConstants, MessageListener {
	
	/**
	 * The interface to the WSN provided by TinyOS.
	 */
	private MoteIF moteIF = null;
	
	/**
	 * Maps AM number -> vector of listeners.
	 */
	private Hashtable<Integer,Vector<MessageListenerJ>> lTable = new Hashtable<Integer,Vector<MessageListenerJ>>();
	
	/**
	 * Whether the SNInterface is promiscuous.
	 */
	private boolean promiscuous;
	
	/**
	 * These are the messages that the SNInterface listens for.
	 */
	private Vector<MessageJ> msgs;
	
	/**
	 * Creates a SNInterface that is not actually connected to the sensor network.
	 *
	 * @param moteIF The MoteIF created by the TinyOS java layer.
	 */
	public SNI(boolean promiscuous) {
		this.promiscuous = promiscuous;
		msgs = new Vector<MessageJ>();
		msgs.add(new RMsgJ());
		msgs.add(new RAckMsgJ());
		msgs.add(new ResetMsgJ());
		msgs.add(new BeaconMsgJ());
		msgs.add(new GWMsgJ());
	}
	
	/**
	 * Creates a non-promiscuous SNInterface.
	 *
	 * @param moteIF The MoteIF created by the TinyOS java layer.
	 */
	public SNI(MoteIF moteIF) {this(moteIF, false);}
	
	/**
	 * Creates a SNInterface that can be promiscuous or non-promiscuous.
	 * A promiscuous SNInterface prints everything received to the screen.
	 *
	 * @param moteIF The MoteIF created by the TinyOS java layer.
	 * @param promiscuous Whether the SNInterface is promiscuous.
	 */
	public SNI(MoteIF moteIF, boolean promiscuous) {
		this(promiscuous);
		connect(moteIF);
	}
	
	/**
	 * Connects the SNInterface to the MoteIF.
	 * 
	 * @param moteIF The interface to the WSN provided by TinyOS.
	 */
	public void connect(MoteIF moteIF) {
		this.moteIF = moteIF;
		for (int i = 0; i < msgs.size(); i++) { moteIF.registerListener(msgs.get(i).toTOSMsg(), this); }
	}
	
	/**
	 * Disconnects the SNInterface from the MoteIF.
	 */
	public void disconnect() {
		if (moteIF == null) return;
		for (int i = 0; i < msgs.size(); i++) { moteIF.deregisterListener(msgs.get(i).toTOSMsg(), this); }
		moteIF = null;
	}
	
	/**
	 * Sets a new MoteIF.  This is used when servilla.Main disconnects from
	 * one sensor network and reconnects to another.
	 * 
	 * @param moteIF The new MoteIF interface.
	 */
	public void setMoteIF(MoteIF moteIF) {
		disconnect();
		connect(moteIF);
	}
	
	/**
	 * Registers a message listener.
	 * 
	 * @param msg The type of message to listen for.
	 * @param listener The listener.
	 */
	public void registerListener(MessageJ msg, MessageListenerJ listener) {
		Integer key = new Integer(msg.getType());
		if (!lTable.containsKey(key)) { lTable.put(key, new Vector<MessageListenerJ>()); }
		Vector<MessageListenerJ> list = (Vector<MessageListenerJ>)lTable.get(key);
		if (!list.contains(listener)) list.add(listener);
	}
	
	/**
	 * Deregisters a message listener.
	 * 
	 * @param msg The type of message the listener was listening for.
	 * @param listener The listener.
	 */
	public void deregisterListener(MessageJ msg, MessageListenerJ listener) {
		Integer key = new Integer(msg.getType());
		Vector list = (Vector)lTable.get(key);
		if (list != null) list.remove(listener);
	}

	/**
	 *  Sends a message to a specific address in the WSN.  Since TinyOS does not support
	 *  multi-hop routing, the address can only be the address of the mote attached to the
	 *  base station.
	 *  
	 *  @param msg The message to send.
	 *  @param address The destination address.  This is usually the address of the
	 *  mote attached to the base station.
	 */
	public void send(int address, MessageJ msg) throws java.io.IOException { 
		if (moteIF != null) { moteIF.send(address, msg.toTOSMsg()); }
	}
	
	/**
	 *  Broadcasts a message into the WSN.
	 *  
	 *  @param msg The message to send.
	 */
	public void send(MessageJ msg) throws java.io.IOException { send(TOS_BCAST_ADDRESS, msg); }
	
	/**
	 * This is called when a message is received from the underlying network 
	 *
	 * @param    to                  the destination address.
	 * @param    m                   a  Message
	 *
	 */
	public void messageReceived(int to, Message m) {
		MessageJ msgJ = null;
		for (int i = 0; i < msgs.size(); i++) { 
			MessageJ currMsg = msgs.get(i);
			if (currMsg.getType() == m.amType()) msgJ = currMsg.clone(m);
		}
		if (msgJ == null) {
			if (Debugger.debug)
				Debugger.log("SNInterface", "ERROR: Received Unknown Message (type: " + m.amType() + ") " + m);
			return;
		}
		
		if (promiscuous) {
			if (msgJ.getType() != BeaconMsg.AM_TYPE) { // don't print beacons
				String dest = "";
				if (to == TOS_BCAST_ADDRESS)
					dest = "TOS_BCAST_ADDR";
				else if (to == TOS_UART_ADDRESS)
					dest = "TOS_UART_ADDRESS";
				else
					dest = ""+to;
				
				Debugger.dbg("SNInterface", "Message dest=" + dest + ":\n" + msgJ.toString() + "\n");
			}
		}
		
		if (to == TOS_LOCAL_ADDRESS || promiscuous) {
			Vector<MessageListenerJ> listeners = lTable.get(m.amType());
			if (listeners != null) {
				for (int i = 0; i < listeners.size(); i++) {
					MessageListenerJ mlj = listeners.get(i);
					mlj.messageReceived(to, msgJ);
				}
			}
		}
	}
	
}
