package servilla.service;

import java.io.*;

import servilla.service.tree.*;
import servilla.util.ArrayToHexString;

/**
 * The service specification compiler.
 *
 * @author Chien-Liang Fok
 */
public class ServiceSpecC {

	public ServiceSpecC() {}
	
	/**
	 * Tests the scanner generated by JLex.
	 * Prints each token produced by the scanner.
	 * 
	 * @param programReader
	 * @throws IOException
	 */
	public void scan(String fileName) throws IOException {
		Reader programReader = new StringReader(servilla.ProgramReader.readProgram(fileName));
		ServiceSpecYylex scanner = new ServiceSpecYylex(programReader);
		java_cup.runtime.Symbol nextToken = scanner.next_token();
		do {
			if (nextToken.value instanceof Token) {				
				System.out.println("Type: " + name(nextToken.sym) + ",\tToken: " + nextToken.value);
			} else
				System.out.println("Type: " + name(nextToken.sym) + ",\t" + nextToken);
			nextToken = scanner.next_token();
		} while (nextToken.sym != sym.EOF);	
	}

	/**
	 * Compiles a service specification.
	 * 
	 * @param fileName The name of the file that contains the service specification.
	 * @param programReader Reads in the service specification.
	 * @throws IOException
	 * @throws SemanticException
	 * @throws NoFreeVariableException
	 * @throws Exception
	 */
	public ServiceSpec compile(String fileName) throws Exception {
		String programText = servilla.ProgramReader.readProgram(fileName);
		Reader programReader = new StringReader(programText);
		ServiceSpecParser parser = new ServiceSpecParser(fileName, new ServiceSpecYylex(programReader));
		try {
			parser.parse();
		} catch (Exception e) {
			throw e;
		}
		ServiceSpec ss = parser.serviceSpec();
		return ss;
	}
	
	/**
	 * Uses reflection to find the name of a constant defined in sym.java.
	 * This is used when testing the scanner.  It translate the sym (symbol) number
	 * into its name.
	 * 
	 * @param type the integer constant
	 * @return The name of the constant as defined in sym.java.
	 */
	private String name(int type) {
		try {
			Class c = Class.forName("servilla.service.sym");
			java.lang.reflect.Field[] fields = c.getDeclaredFields();
			for (int i = 0; i < fields.length; i++) {
				if (fields[i].getInt(c) == type)
					return fields[i].getName();
			}
		} catch(Exception e) {
			e.printStackTrace();
		}
		return "UNKNOWN";
	}
	
	private void log(String msg) {
		System.out.println(msg);
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		boolean scan = false;
		boolean test = false;
		for (int i = 0; i < args.length; i++) {
			if(args[i].equals("-scan"))
				scan = true;
			else if (args[i].equals("-test"))
				test = true;
		}
		if (test) {
			// debug code here
			System.out.println(""+PropertyType.valueOf("TYPE"));
		} else {
			try {
				String fileName = args[args.length-1];
				ServiceSpecC compiler = new ServiceSpecC();
				if (scan)
					compiler.scan(fileName);
				else {
					ServiceSpec ss = compiler.compile(fileName);
					String s = ArrayToHexString.toString(ss.toByteArray());
					System.out.println(s);
				}
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
	}
}
