package servilla.service.tree;

import java.util.*;
import servilla.*;
import servilla.util.*;

/**
 * Defines the types in Servilla.  The type is stored as a byte, meaning there can be
 * at most 256 unique types in Servilla.
 * 
 * @author Chien-Liang Fok
 */
public enum DataType implements ServillaConstants {
	VOID, 
	INT,
	FLOAT,
	STRING,
	TYPE,
	ANY,
	;
	
	static byte counter = 0;
	static byte nextCount() { return (byte)counter++;}
	
	private byte value;
	
	/**
	 * Creates a data type with a unique value.
	 */
	DataType() { this.value = DataType.nextCount(); }
	
	// This was automatically generated by servilla.service.tree.DataType
	// Copy this into servilla.service.tree.DataType
	public static DataType getType(int type) throws InvalidTypeException {
		if(type == VOID.value()) { return VOID;}
		if(type == INT.value()) { return INT;}
		if(type == FLOAT.value()) { return FLOAT;}
		if(type == STRING.value()) { return STRING;}
		if(type == TYPE.value()) { return TYPE;}
		if(type == ANY.value()) { return ANY;}
		else { throw new InvalidTypeException("Unknown type: " + type); }
	}
	
	/**
	 * Returns the size in bytes of the data type or -1 if the size
	 * is not known.  INT and FLOAT are both 4 bytes, while the size of
	 * a STRING is unknown.
	 * 
	 * @param type The data type whose size is being returned.
	 * @return The size of the data type, or -1 if not known.
	 */
	public int size() {
		if (value == INT.value() || value == FLOAT.value())
			return 4;
		else {
			Debugger.warn("DataType", "Attempted to get size of non int or float type.");
			return -1;
		}
	}

	/**
	 * Determines whether this type is an integer.
	 * 
	 * @return whether this type is an integer.
	 */
	public boolean isInt() { return toString().contains("INT"); }
	
	/**
	 * Determines whether this type is a float.
	 * 
	 * @return whether this type is a float.
	 */
	public boolean isFloat() { return toString().contains("FLOAT"); }
	
	/**
	 * Determines whether this type is a string.
	 * 
	 * @return whether this type is a float.
	 */
	public boolean isString() { return toString().contains("STRING"); }
	
	/**
	 * Returns the byte value of this data type.
	 * @return the byte value of this data type.
	 */
	public byte value() { return value; }
	
	/**
	 * Returns the int value of this data type.
	 * @return the int value of this data type.
	 */
	public int intValue() { return (int)value & 0xff;}
	
	public static void main(String[] args) {
		DataType[] v = DataType.values();
		System.out.println("// Copy this into nesc/servilla/types/Servilla.h");
		System.out.println("// This was generated by servilla.service.tree.DataType");
		System.out.println("// Total # data types: " + v.length);
		System.out.println("typedef enum {");
		for (int i = 0; i < v.length; i++) {
			System.out.println("\t" + v[i] + " = "+ v[i].value() + ",");
		}
		System.out.println("} DataType_t;\n");
		
		System.out.println("// Copy this into components/VarUtil/VarUtilM.nc");
		System.out.println("// This was generated by servilla.service.tree.DataType");
		System.out.println("command char* VarUtil.type2Str(uint8_t dataType) {");
		System.out.println("\tswitch(dataType) {");
		for (int i = 0; i < v.length; i++) {
			System.out.println("\t\tcase " + v[i] + ": return \""+ v[i] + "\"; break;");
		}
		System.out.println("\t\tdefault: return \"UNKNOWN_DATA_TYPE\";");
		System.out.println("\t}");
		System.out.println("}\n");
		
		System.out.println("\t// Copy this into servilla.service.tree.DataType");
		System.out.println("\t// This was automatically generated by servilla.service.tree.DataType");
		System.out.println("\tpublic static DataType getType(int type) throws InvalidTypeException {");
		for (int i = 0; i < v.length; i++) {
			System.out.println("\t\tif(type == " + v[i] + ".value()) { return "+ v[i] + ";}");
		}
		System.out.println("\t\telse { throw new InvalidTypeException(\"Unknown type: \" + type); }");
		System.out.println("\t}\n");
		
		// This defines the valid data types that are used in service specifications
		// For now, the tuple describing the input and output interfaces can only be
		// matched exactly.  There are no no wild cards or VOID types allowed in a spec.
		Vector<DataType> specDataTypes = new Vector<DataType>();
		for (int i = 0; i < v.length; i++) {
			if (!v[i].toString().endsWith("A") && !v[i].toString().startsWith("VOID")
					&& !v[i].toString().startsWith("TYPE") && !v[i].toString().startsWith("ANY")) {
				specDataTypes.add(v[i]);
			}
		}
		
		StringBuffer terminalDecl = new StringBuffer();
		StringBuffer nonterminalDecl = new StringBuffer();
		terminalDecl.append("terminal Field ");
		nonterminalDecl.append("nonterminal Field ");
		for (int i = 0; i < specDataTypes.size(); i++) {
			terminalDecl.append(specDataTypes.get(i) + "_T");
			nonterminalDecl.append(specDataTypes.get(i).toString().toLowerCase());
			if (i < specDataTypes.size()-1) {
				terminalDecl.append(", ");
				nonterminalDecl.append(", ");
			}
		}
		System.out.println("// Copy this into the terminal/non-terminal section of servilla.service.ServiceSpec.cup");
		System.out.println("// This was generated by servilla.service.tree.DataType");
		System.out.println(terminalDecl + ";\n" + nonterminalDecl + ";\n");
		
		StringBuffer fieldDecl = new StringBuffer();
		fieldDecl.append("field ::= ");
		for (int i = 0; i < specDataTypes.size(); i++) {
			DataType dt = specDataTypes.get(i);
			String lc = dt.toString().toLowerCase();
			fieldDecl.append(lc + ":" + lc + "v {: RESULT=" + lc + "v; :}");
			if (i < specDataTypes.size()-1) fieldDecl.append(" | ");
		}
		System.out.println("// Copy this into the production section of servilla.service.ServiceSpec.cup");
		System.out.println("// This was generated by servilla.service.tree.DataType");
		System.out.println(fieldDecl + ";");
		for (int i = 0; i < specDataTypes.size(); i++) {
			DataType dt = specDataTypes.get(i);
			String lc = dt.toString().toLowerCase();
			System.out.println(lc + "\t::= " + dt.toString() + "_T {: RESULT=new Field(DataType." + dt 
					+ ", 0); :} | " + dt + "_T array:a {: RESULT=new Field(DataType." + dt + ", a.intValue()); :};");
		}
		System.out.println("");
		
		System.out.println("// Copy this into the regexp servilla.service.ServiceSpec.lex");
		for (int i = 0; i < specDataTypes.size(); i++) {
			DataType dt = specDataTypes.get(i);
			String toPrint = dt + "_T=\"" + dt.toString().toLowerCase() + "_t\"";
			if (i == 0)
				toPrint += "// This was generated by servilla.service.tree.DataType";
			System.out.println(toPrint);
		}
		System.out.println("");
		
		System.out.println("// Copy this into the production setion of servilla.service.ServiceSpec.lex");
		System.out.println("// This was generated by servilla.service.tree.DataType");
		for (int i = 0; i < specDataTypes.size(); i++) {
			DataType dt = specDataTypes.get(i);
			System.out.println("<YYINITIAL> {" + dt + "_T} { return new Symbol(sym." + dt + "_T, new StringToken(yytext(), yyline, yychar)); }");
		}
		
	}
}
