package servilla.service.tree;

import servilla.compiler.tree.Type;

/**
 * Defines a field within a tuple.  A field consists of a data type and a number indicating
 * the number of array dimensions.  If the data type is not an array, the number of dimensions
 * is zero.
 * 
 * @author Chien-Liang Fok
 *
 */
public class Field {
	private DataType type;
	private int numDim;
	
	public Field(DataType type, int numDim) {
		this.type = type;
		this.numDim = numDim;
	}
	
	/**
	 * Return the field's primitive data type.
	 * @return the field's primitive data type.
	 */
	public DataType type() { return type; }
	
	/**
	 * Returns the number of array dimensions of this field.  If the field is
	 * not an array, 0 is returned.
	 * @return the number of array dimensions of this field.
	 */
	public int numDim() { return numDim; }
	
	/**
	 * Returns whether this field is an array.
	 * @return whether this field is an array.
	 */
	public boolean isArray() { return numDim != 0; }
	
	/**
	 * Determines whether this field matches the specified type.  They match if they
	 * have the same primtive type and array dimensions.  Note that the field does not
	 * specify the size of each dimension in the array.
	 * 
	 * @param t  The type to compare with.
	 * @return Whether this field matches the specified type.
	 */
	public boolean matches(Type t) {
		if (t.type() != type) return false;  // primitive types do not match
		if (t.isArray() != isArray()) return false; // one is an array and the other is not
		if (t.isArray() && t.numDim() != numDim()) return false; // number of dimensions do not match
		return true;
	}
	
	public String toString() {
		String result = type.toString();
		for (int i = 0; i < numDim; i++) {
			result += "[]";
		}
		return result;
	}
}
