package servilla.service.tree;

import java.util.*;

import servilla.util.ArrayToHexString;

/**
 * A relation/value pair in which the value is an integer.
 *
 * @author Chien-Liang Fok <liangfok@wustl.edu>
 */
public class IntValue extends Value {
	long value;
	
	public IntValue(long value) { this.value = value; }
	
	public long value() { return value; }
	
	/**
	 * Returns the type of the value.
	 * 
	 * @return the type of the value.
	 */
	public DataType type() {
		if (value < 0) {
			if ((value >> 16) > 0) {
				return DataType.INT32;
			} else {
				return DataType.INT16;
			}
		} else {
			if ((value >> 16) > 0) {
				return DataType.UINT32;
			} else {
				return DataType.UINT16;
			}
		}
	}
	
	/**
	 * Convert the IntValue into a vector of bytes.
	 * Note that the int value is stored in little-endian format.
	 * This makes it compatible with the the MSP430 processor.
	 * 
	 * Byte format: [data type (1B)] [int value (2B or 4B)]
	 * 
	 * @see service.Value#toByteArray()
	 * @return the byte representation of this class.
	 */
	public Vector<Byte> toByteVector() {
		DataType type = type();
		Vector<Byte> bv = new Vector<Byte>();
		bv.add(type.value());
		if (type == DataType.INT32 || type == DataType.UINT32) {
			for (int j = 0; j < 4; j++) {
				bv.add((byte)(value >> j*8));
			}
		} else {
			for (int j = 0; j < 2; j++) {
				bv.add((byte)(value >> j*8));
			}
		}
		return bv;
	}
	
	public String toString() {
		return "" + value;
	}
	
	public static void main(String[] args) throws Exception {
		System.out.println("IntValue: 6");
		IntValue rfv1 = new IntValue(6);
		System.out.println(ArrayToHexString.toString(rfv1.toByteArray()) + "\n");
		
		System.out.println("IntValue: -1");
		rfv1 = new IntValue(-1);
		System.out.println(ArrayToHexString.toString(rfv1.toByteArray()) + "\n");	
		
		System.out.println("IntValue: -32768");
		rfv1 = new IntValue(-32768);
		System.out.println(ArrayToHexString.toString(rfv1.toByteArray()) + "\n");	
		
		System.out.println("IntValue: 32767");
		rfv1 = new IntValue(32767);
		System.out.println(ArrayToHexString.toString(rfv1.toByteArray()) + "\n");
		
		System.out.println("IntValue: 131071");
		rfv1 = new IntValue(131071);
		System.out.println(ArrayToHexString.toString(rfv1.toByteArray()) + "\n");
	}
	
}