package servilla.service.tree;

import servilla.*;
import java.util.*;
import org.apache.commons.lang.ArrayUtils;

/**
 * Encapsulates a service specification.  A service specification consists of a list of properties.
 * 
 * @author Chien-Liang Fok
 */
public class ServiceSpec implements ServillaConstants {
	String fileName; 	// the file name of the service. This is only used in the toString
	int handle = -1;		// the handle assigned to the service
	Vector<Property> properties = new Vector<Property>();

	/**
	 * Creates a service spec.
	 * 
	 * @param fileName The name of the file that contains the spec.
	 */
	public ServiceSpec(String fileName) { this.fileName = fileName; }
	
	/**
	 * Determines whether there is a property of the specified
	 * type already in the spec.
	 * 
	 * @param type The type of service spec.
	 * @return true if a property of the specified type is already in the spec.
	 */
	private boolean containsType(PropertyType type) {
		for (int i = 0; i < properties.size(); i++) {
			if (properties.get(i).type() == type)
				return true;
		}
		return false;
	}
	
	/**
	 * Returns the number of attribute properties in the spec.
	 * 
	 * @return The number of attribute properties in the spec.
	 */
	public int numAttributes() {
		int result = 0;
		for (int i = 0; i < properties.size(); i++) {
			if (properties.get(i).type() == PropertyType.ATTRIBUTE) result++;
		}
		return result;
	}
	
	/**
	 * Returns the property of a specified type.  If there are more than one
	 * such properties, the count specifies which property to return.  Count
	 * starts at 1 (the first occurance).
	 * 
	 * @return The property of the specified type.
	 * @throws SpecCompilerException when the property does not exist.
	 */
	private Property getProperty(PropertyType pt, int count) throws SpecCompilerException {
		int origCount = count;
		for (int i = 0; i < properties.size(); i++) {
			Property cp = properties.get(i);
			if (cp.type() == pt && count-- == 0)
				return cp;
		}
		
		String error = null;
		if (origCount == 0) 
			error = "Spec " + fileName + " does not have a property of type " + pt;
		else
			error = "Spec " + fileName + " does not have " + origCount + " properties of type " + pt;
		throw new SpecCompilerException(error);
	}
	
	/**
	 * Returns the name property of the spec, or null if the spec does
	 * not have such a property.
	 * 
	 * @return the name property of the spec.
	 * @throws SpecCompilerException when the property does not exist.
	 */
	public NameProperty getNameProperty() throws SpecCompilerException {
		return (NameProperty)getProperty(PropertyType.NAME, 0);
	}
	
	/**
	 * Returns the input property of the spec, or null if the spec does
	 * not have such a property.
	 * 
	 * @return the input property of the spec.
	 * @throws SpecCompilerException when the property does not exist.
	 */
	public InputProperty getInputProperty() throws SpecCompilerException {
		return (InputProperty)getProperty(PropertyType.INPUT, 0);
	}
	
	/**
	 * Returns the output property of the spec, or null if the spec does
	 * not have such a property.
	 * 
	 * @return the output property of the spec.
	 * @throws SpecCompilerException when the property does not exist.
	 */
	public OutputProperty getOutputProperty() throws SpecCompilerException {
		return (OutputProperty)getProperty(PropertyType.OUTPUT, 0);
	}
	
	/**
	 * Returns the N'th attribute property of the spec, or null if the spec does
	 * not have such a property.
	 * 
	 * @param count The index of the attribute property.  This value must be 
	 * >= 0 and < numAttributes();
	 * @return the output property of the spec.
	 * @throws SpecCompilerException when the property does not exist.
	 */
	public AttributeProperty getAttributeProperty(int count) throws SpecCompilerException {
		return (AttributeProperty)getProperty(PropertyType.ATTRIBUTE, count);
	}
	
	/**
	 * Adds a property to the spec.
	 * @param prop The property to add to the spec.
	 */
	public void addProperty(Property prop) throws SpecCompilerException { 
		if (prop.type() != PropertyType.ATTRIBUTE && containsType(prop.type()))
			throw new SpecCompilerException("Error: Attempted to add two " + prop.type() + " properties.");
		properties.add(prop); 
	}
	
	/**
	 * Sets the service spec's handle.  Only specs provided by scripts
	 * have handles.  A handle is an uin16_t.
	 * 
	 * @param handle The script's handle
	 */
	public void setHandle(int handle) { this.handle = handle; }
	
	/**
	 * Returns whether this spec has a handle.
	 * @return whether this spec has a handle.
	 */
	public boolean hasHandle() { return handle != -1; }
	
	/**
	 * Returns the handle of this spec.
	 * @return the handle of this spec.
	 */
	public int handle() { return handle; }
	
	/**
	 * Convert the ServiceSpec into a vector of bytes.
	 * 
	 * Byte format for specification: 
	 *   [header if spec belongs to a script (servilla.ServillaConstants.SST_SPEC_OFFSET Bytes)] 
	 *   [Name property (n Bytes)]
	 *   [Input property (n Bytes)]
	 *   [Output property (n Bytes)]
	 *   [Number of attribute properties (1 byte)]
	 *   [Property 1 (n Bytes)]
	 *   ...
	 * 
	 * A spec provided by the service does NOT have a handle.
	 * 
	 * @param handle The service's handle, or NULL if this is a service description.
	 * @return the byte representation of this class.
	 */
	public Vector<Byte> toByteVector() throws SpecCompilerException {
		Vector<Byte> bv = new Vector<Byte>();
		// if a spec has a handle, that means it belongs to a script and must include a SST_SPEC_OFFSET-byte header
		if (hasHandle()) { 
			bv.add((byte)(handle & 0xff));
			bv.add((byte)((handle >> 8) & 0xff));
			for (int i = 0; i < SST_SPEC_OFFSET-2; i++) { bv.add((byte)0); } // for bookeeping purposes
		}
		bv.addAll(getNameProperty().toByteVector());
		bv.addAll(getInputProperty().toByteVector());
		bv.addAll(getOutputProperty().toByteVector());
		int numAttr = numAttributes();;
		if (numAttr > 0xff) 
			throw new SpecCompilerException("ERROR: " + fileName + " has too many attributes (" + numAttr + ")");
		bv.add((byte)numAttr);
		for (int i = 0; i < numAttr; i++) {
			bv.addAll(getAttributeProperty(i).toByteVector());
		}
		return bv;
	}
	
	/**
	 * Convert the ServiceSpec into an array of bytes.
	 * 
	 * Byte format for specification: 
	 *   [handle if available (2 Bytes)] 
	 *   [Name property (n Bytes)]
	 *   [Input property (n Bytes)]
	 *   [Output property (n Bytes)]
	 *   [Number of attribute properties (1 byte)]
	 *   [Property 1 (n Bytes)]
	 *   ...
	 * 
	 * A spec provided by the service does NOT have a handle.
	 * 
	 * @param handle The service's handle, or NULL if this is a service description.
	 * @return the byte representation of this class.
	 */
	public byte[] toByteArray() throws SpecCompilerException {
		return ArrayUtils.toPrimitive(toByteVector().toArray(new Byte[]{}));
	}

	public String toString() {
		String result = fileName;
		if (hasHandle()) {
			result += " (handle=" + handle + ")";
		} else {
			result += " (No Handle)";
		}
		for (int i = 0; i < properties.size(); i++) {
			result += "\n\t" + properties.get(i);
		}
		return result;
	}
}
