package servilla.service.tree;

import java.util.*;

/**
 * Defines the input or output interface of a service.
 * An interface is defined by an ordered set of data types.
 * 
 * @author Chien-Liang Fok
 */
public class Tuple {
	private Vector<Field> tuple = new Vector<Field>();

	/**
	 * Create an empty tuple.
	 */
	public Tuple() {}

	/**
	 * Add a field to the tuple.
	 * @param f The field to add.
	 */
	public void add(Field f) { tuple.add(f); }
	
	/**
	 * Add all fields to the tuple.
	 * @param fields The fields to add.
	 */
	public void addAll(Vector<Field> fields) { tuple.addAll(fields); }
	
	/**
	 * Returns the size of the tuple.  Since a tuple consists of a sequence
	 * of data types and each data type is one byte, the number of fields
	 * of the tuple equals the number of bytes.
	 * 
	 * @return the size of the tuple.
	 */
	public int size() { return tuple.size(); }
	
	/**
	 * Returns the specified field of the tuple.  
	 * @param index The index of the field.  Must be between 0 and size()-1.
	 * @return The field at the specified index.
	 */
	public Field getField(int index) { return tuple.get(index); }
	
	/**
	 * Removes a field from the tuple.
	 * 
	 * @param indx The index of the field to remove.  Fields
	 * are indexed from 0.  indx must be between 0 and size-1.
	 */
	public void remove(int indx) {tuple.remove(indx); }
	
	/**
	 * Converts the value into its byte representation.
	 * 
	 * The byte representation is:
	 * 	[number of fields (1B)] [fields (num. fields * 1B)]
	 * 
	 * @return The byte representation of the Value.
	 */
	public Vector<Byte> toByteVector() {
		Vector<Byte> bv = new Vector<Byte>();
		bv.add((byte)size());
		for (int i = 0; i < size(); i++) {
			bv.add(tuple.get(i).type().value());
			bv.add((byte)tuple.get(i).numDim());
		}
		return bv;
	}
	
	/**
	 * Returns the string representation of this class.
	 */
	public String toString() {
		String result = "{";
		for (int i = 0; i < size(); i++) {
			result += tuple.get(i);
			if (i < size()-1) result += ", ";
		}
		return result+"}";
	}
}
