package servilla.util;

import java.util.*;
import org.apache.commons.lang.ArrayUtils;

/**
 * Translates a value into its byte representation and saves the
 * bytes in the specified vector.  The values are stored in 
 * little endian format.
 * 
 * @author Chien-Liang Fok
 *
 */
public class ByteUtils {
	
	private static final void saveBytes(Vector<Byte> bv, long val, int numBytes) {
		for (int i = 0; i < numBytes; i++) {
			bv.add((byte)(val >> (8*i) & 0xff));
		}
	}
	
	public static final void saveBytes1(Vector<Byte> bv, short val) { saveBytes(bv, (long)val, 1); }
	public static final void saveBytes2(Vector<Byte> bv, int val) { saveBytes(bv, (long)val, 2); }
	public static final void saveBytes4(Vector<Byte> bv, long val) { saveBytes(bv, val, 4); }
	public static final void saveBytes4(Vector<Byte> bv, float val) { saveBytes(bv, (long)Float.floatToRawIntBits(val), 4); }
	
	/**
	 * Translates a string into an array of bytes and adds it to the end
	 * of the byte vector.  A null character (0) is added to the end of 
	 * the string.
	 * 
	 * @param bv The byte vector in which to store the array of bytes.
	 * @param val The string to store.
	 */
	public static final void saveBytes(Vector<Byte> bv, String val) { 
		bv.addAll(Arrays.asList(ArrayUtils.toObject(val.getBytes()))); 
		bv.add(new Byte((byte)0)); // null terminate the string.
	}
	
	/**
	 * Takes a vector of bytes and returns its string representation.
	 * The bytes are assumed to be in little endian format.  Thus, if
	 * the byte vector contains 0x01 0x02 0x03, the return value will be
	 * "0x030201".
	 * 
	 * @param bv The byte vector in little-endian format.
	 * @return The string represenatation of the bytes.
	 */
	public static String toString(Vector<Byte> bv) {
		StringBuffer sb = new StringBuffer();
		sb.append("0x");
		for (int i = bv.size()-1; i >= 0; i--) {
			sb.append(ArrayToHexString.byteToHexString(bv.get(i)).substring(2));
		}
		return sb.toString();
	}
	
	
	/**
	 * Converts an array of bytes in little-endian format into an uint16.
	 * 
	 * @param data The array of bytes in little-endiam format
	 * @param indx The index within the array in which the bytes start
	 * @return The uint16.
	 */
	public static final int getUint16(byte[] data, int indx) {
		int i = 0;
	    i |= data[indx+1] & 0xFF;
	    i <<= 8;
	    i |= data[indx] & 0xFF;
	    return i;
	}
	
	/**
	 * Converts an array of bytes in little-endian format into an int16.
	 * 
	 * @param data The array of bytes in little-endiam format
	 * @param indx The index within the array in which the bytes start
	 * @return The int16.
	 */
	public static final short getInt16(byte[] data, int indx) {
		short s = 0;
		s |= data[indx+1] & 0xFF;
	    s <<= 8;
	    s |= data[indx] & 0xFF;
	    return s;
	}
	
	/**
	 * Converts an array of bytes in little-endian format into an uint32.
	 * 
	 * @param data The array of bytes in little-endiam format
	 * @param indx The index within the array in which the bytes start
	 * @return The uint32.
	 */
	public static final long getUint32(byte[] data, int indx) {
	    long l = 0;
	    l |= data[indx+3] & 0xFF;
	    l <<= 8;
	    l |= data[indx+2] & 0xFF;
	    l <<= 8;
	    l |= data[indx+1] & 0xFF;
	    l <<= 8;
	    l |= data[indx+0] & 0xFF;
	    return l;
	}
	
	/**
	 * Converts an array of bytes in little-endian format into an uint32.
	 * 
	 * @param data The array of bytes in little-endiam format
	 * @param indx The index within the array in which the bytes start
	 * @return The uint32.
	 */
	public static final int getInt32(byte[] data, int indx) {
		int l = 0;
	    l |= data[indx+3] & 0xFF;
	    l <<= 8;
	    l |= data[indx+2] & 0xFF;
	    l <<= 8;
	    l |= data[indx+1] & 0xFF;
	    l <<= 8;
	    l |= data[indx+0] & 0xFF;
	    return l;
	}
	
	/**
	 * Converts an array of bytes in little-endian format into an uint32.
	 * 
	 * @param data The array of bytes in little-endiam format
	 * @param indx The index within the array in which the bytes start
	 * @return The uint32.
	 */
	public static final float getFloat32(byte[] data, int indx) {
		int i = getInt32(data, indx);
		return Float.intBitsToFloat(i);
	}
	/**
	 * Returns the string beginning at the indx.  A string is null terminated.
	 * 
	 * @param data The bytes.
	 * @param indx The starting address of the string.
	 * @return The string.
	 */
	public static final String getString(byte[] data, int indx) {
		Vector<Byte> v = new Vector<Byte>();
		do {
			v.add(data[indx]);
		} while (data[indx] != 0);
		byte[] bv = ArrayUtils.toPrimitive(v.toArray(new Byte[]{}));
		return new String(bv);
	}
	
	public static void main(String[] args) {
		byte[] test = new byte[] {0x01, (byte)0x80, 0x00, 0x00};
		System.out.println("int16 = "+getInt16(test, 0));
		System.out.println("uint16 = "+getUint16(test, 0));
	}
}
